
program BORN
	use sep
	use mem_mod
	use fbi_mod
	use initialize_born
	implicit none

	!===========================================================================
	!	Allocate and Initialize
	!---------------------------------------------------------------------------
	integer									:: isou,rec_dx,rec_ox,sou_dx,sou_ox,nz,nx,nt,sou_nx
	integer                                 :: rec_nx,rec_z,nh,start,finish,startA,finishA,np
	real 									:: dz,dx,dt,oz,ox,ot,cr
	logical									:: adjoint
	real, dimension(:,:,:), allocatable		:: data
	real, dimension(:,:), allocatable		:: vel_in, data_shot, image

	!===========================================================================
	real, dimension(:,:,:), allocatable	    :: insrc
	real, dimension(:,:), allocatable		:: insrc_in, coor
	!===========================================================================

	write(0,*) "======================================================"
	call system_clock(count_rate=cr)
	call system_clock(startA)

	write(0,*) "--------- Inside the BORN program"
	call init_rtm(nz,nx,nt,nh,dz,dx,dt,oz,ox,ot,sou_nx,sou_dx,sou_ox,rec_nx,rec_dx,rec_ox,rec_z,np)
	write(0,*) "--------- Done with BORN initialization routine"

	call mem_alloc3d(data, nt, rec_nx, sou_nx)
	call mem_alloc2d(data_shot, nt, rec_nx)
	call mem_alloc2d(image, nz, nx)
	call mem_alloc2d(vel_in, nz, nx)
	call mem_alloc2d(coor, 2, np)

	call sep_init()
	call sep_read(vel_in,"velmod")
	call model_update(vel_in)
	call from_param("isou", isou)
	call from_param("adjoint", adjoint)

!===========================================================================
	call system_clock(start)
	if (.not.exist_file('insrc' )) write(0,*) "NEED TO HAVE A COMPRESSED SOURCE WAVEFIELD INPUT"
	if (.not.exist_file('coor_src'  )) write(0,*) "NEED TO HAVE A WAVEFIELD COORDINATES INPUT"

	if (exist_file('insrc' )) then
		write(0,*) "READING IN THE SAVED SOURCE WAVEFIELD"
		call mem_alloc2d(insrc_in, np, nt)
		call mem_alloc3d(insrc, np, nt, 1)
		call sep_read(insrc_in, 'insrc' )
		insrc=0.0
		insrc(:,:,1)=insrc_in
		call sep_read(coor, 'coor_src' )
		! write(0,*) " sum(coor) = ", sum(coor)
		call init_srcwave_fbi(insrc,coor)
		write(0,*) "FINISHED READING IN THE SAVED SOURCE WAVEFIELD"
	end if
	call system_clock(finish)
	write(0,*) "TIME FOR LOADING SAVED SOURC WAVEFIELD = (seconds) ", ((finish-start)/cr)
!===========================================================================


	if (adjoint) then
	write(0,*) " ADJOINT"
		call to_history("n1",nz, 'model')
		call to_history("d1",dz, 'model')
		call to_history("o1",oz, 'model')
		call to_history("n2",nx, 'model')
		! call to_history("n3",(2*nh+1), 'model')
		call to_history("d2",dx, 'model')
		call to_history("o2",ox, 'model')
		call to_history("label1",'z [km]', 'model')
		call to_history("label2",'x [km]', 'model')
	else
		write(0,*) "FORWARD"
   		call to_history("n1",nt, 'data')
		call to_history("d1",dt, 'data')
		call to_history("o1",ot, 'data')
   		call to_history("n2",rec_nx, 'data')
		call to_history("d2",rec_dx, 'data')
		call to_history("o2",rec_ox, 'data')
		call to_history("label1",'Time [s]', 'data')
		call to_history("label2",'Receiver k', 'data')
	end if


	call system_clock(finishA)
	write(0,*) "TIME BEFORE BORN OPERATOR = (seconds) ", ((finishA-startA)/cr)
	!---------------------------------------------------------------------------
	! Linear operator: full nonlinear data => model perturbation
	if (adjoint) then
		write(0,*) " Adjoint: data perturbation => model perturbation"
		image=0.0
		call sep_read(data,"data")
		data_shot = data(:,:,isou)
write(0,*) "sum(data_shot) =", sum(data_shot)
		call L_lop3(.true., .false., image, data_shot, isou)
write(0,*) "sum(image) =", sum(image)
		call sep_write(image,'model')
	!---------------------------------------------------------------------------
	! Linear operator: model perturbation => full nonlinear data 
	else
		write(0,*) " Forward: model perturbation => full nonlinear data "
		data=0.0
		data_shot=0.0
		call sep_read(image,'model')
write(0,*) "sum(image) =", sum(image)
		call L_lop3(.false., .false., image, data_shot, isou)
		call sep_write(data_shot,'data')
write(0,*) "sum(data_shot) =", sum(data_shot)
	end if
	!---------------------------------------------------------------------------


	call sep_close()

	call system_clock(finishA)
	write(0,*) "TIME FOR FULL BORN PROGRAM = (seconds) ", ((finishA-startA)/cr)
end program



