
program CG_STEPPER
	use sep
	use mem_mod
	! use fbi_mod
	use initialize1
	implicit none

	!===========================================================================
	!	Allocate and Initialize
	!---------------------------------------------------------------------------
	real, dimension(:), allocatable		:: Hv, p, r, update
	real								:: alpha, alpha_old, rsold, rsnew, dz, dx, oz, ox, beta
	integer								:: nz, nx
	!===========================================================================
	write(0,*) "------- INSIDE CG_STEPPER PROGRAM-----------"
	call sep_init()
	call from_param("n1", nz)
	call from_param("n2", nx)
	call from_param("d1", dz)
	call from_param("d2", dx)
	call from_param("o1", oz)
	call from_param("o2", ox)


    call from_aux("Hv_in","n1",nz)
    call from_aux("Hv_in","n2",nx)

	call mem_alloc1d(Hv, nz*nx)
	call mem_alloc1d(p, nz*nx)
	call mem_alloc1d(update, nz*nx)
	call mem_alloc1d(r, nz*nx)


	call sep_read(Hv,"Hv_in")
	call sep_read(p,"p_in")
	call sep_read(r,"r_in")

	OPEN(UNIT=1,FILE="rsold.txt",FORM="FORMATTED",STATUS="OLD",ACTION="READ")
	READ(UNIT=1, FMT=*) rsold
	CLOSE(UNIT=1)

	OPEN(UNIT=1,FILE="alpha_old.txt",FORM="FORMATTED",STATUS="OLD",ACTION="READ")
	READ(UNIT=1, FMT=*) alpha_old
	CLOSE(UNIT=1)

!---------------------------------------------------------
	beta = DOT_PRODUCT(p,Hv)
	alpha=rsold/beta

! !--- Prevent negative curvature instances ----------------
! 	if (alpha<0.0) then
! 		alpha=alpha_old
! 	end if

!--- Create update ----------------------------------------
	! update=alpha*p
	update=p
	write(0,*) " sum(Hv) =", sum(Hv)
	write(0,*) " sum(p) =", sum(p)
	write(0,*) " DOT_PRODUCT(p,Hv) =", DOT_PRODUCT(p,Hv)

	write(0,*) " rsold =", rsold
	write(0,*) " alpha =", alpha

!--- Update direction and residual vectors  ---------------
	r=r-alpha*Hv
	rsnew=DOT_PRODUCT(r,r)
	p=r+(rsnew/rsold)*p
	rsold=rsnew
	write(0,*) " rsnew =", rsnew

!---------------------------------------------------------


	if (exist_file('update')) then
		call to_history("n1",nz, 'update')
		call to_history("n2",nx, 'update')
		! call to_history("n3",300, 'update')
		call to_history("d1",dz, 'update')
		call to_history("d2",dx, 'update')
		call to_history("o1",oz, 'update')
		call to_history("o2",ox, 'update')
		call to_history("label1",'z [km]', 'update')
		call to_history("label2",'x [km]', 'update')
		! call to_history("label3",'iteration', 'update')
		call sep_write(update,"update")
	end if


	if (exist_file('p_out')) then
		call to_history("n1",nz, 'p_out')
		call to_history("n2",nx, 'p_out')
		! call to_history("n3",300, 'p_out')
		call to_history("d1",dz, 'p_out')
		call to_history("d2",dx, 'p_out')
		call to_history("o1",oz, 'p_out')
		call to_history("o2",ox, 'p_out')
		call to_history("label1",'z [km]', 'p_out')
		call to_history("label2",'x [km]', 'p_out')
		! call to_history("label3",'iteration', 'p_out')
		call sep_write(p,"p_out")
	end if



	if (exist_file('r_out')) then
		call to_history("n1",nz, 'r_out')
		call to_history("n2",nx, 'r_out')
		! call to_history("n3",300, 'r_out')
		call to_history("d1",dz, 'r_out')
		call to_history("d2",dx, 'r_out')
		call to_history("o1",oz, 'r_out')
		call to_history("o2",ox, 'r_out')
		call to_history("label1",'z [km]', 'r_out')
		call to_history("label2",'x [km]', 'r_out')
		! call to_history("label3",'iteration', 'r_out')
		call sep_write(r,"r_out")
	end if



	if (exist_file('alpha_cg')) then
		call to_history("n1",1					, 'alpha_cg')
		call to_history("d1",1.0				, 'alpha_cg')
		call to_history("o1",0.0				, 'alpha_cg')
		call to_history("label1",'iteration'	, 'alpha_cg')
		call srite( 'alpha_cg', alpha ,4*1)
	end if




	if (exist_file('rsold')) then
		call to_history("n1",1					, 'rsold')
		call to_history("d1",1.0				, 'rsold')
		call to_history("o1",0.0				, 'rsold')
		call to_history("label1",'iteration'	, 'rsold')
		call srite( 'rsold', rsold ,4*1)
	end if




	if (exist_file('beta_cg')) then
		call to_history("n1",1					, 'beta_cg')
		call to_history("d1",1.0				, 'beta_cg')
		call to_history("o1",0.0				, 'beta_cg')
		call to_history("label1",'iteration'	, 'beta_cg')
		call srite( 'beta_cg', beta ,4*1)
	end if




	OPEN(UNIT=1,FILE="alpha_old.txt",FORM="FORMATTED",STATUS="REPLACE",ACTION="WRITE")
	WRITE(UNIT=1, FMT=*) alpha
	CLOSE(UNIT=1)


	OPEN(UNIT=1,FILE="rsold.txt",FORM="FORMATTED",STATUS="REPLACE",ACTION="WRITE")
	WRITE(UNIT=1, FMT=*) rsold
	CLOSE(UNIT=1)
	write(0,*) "-------  CG_STEPPER PROGRAM COMPLETE -----------"
end program









