
program DIFF3D_2

!===========================================================================
!		This program does a residual difference between a real and synthetic
!		dataset that may or may not have the same number of recievers

!		The real data input must have the ds and dr values in [m] units, not 
!		sampling units. 
!===========================================================================


	use sep
	use mem_mod
	use fbi_mod
	use initialize2
	implicit none

	!===========================================================================
	!	Allocate and Initialize
	!---------------------------------------------------------------------------
	integer									:: i,j,k,nt,ni,sou_nx,rec_nx,off_nx,is, fudge
	integer									:: ir,ir2,ii,dtt,zone,lower,upper, t1,t2, it, fade, tt
	real, dimension(:,:,:), allocatable		:: residual, realdata, syndata, weight
	real, dimension(:), allocatable			:: aa
	real 									:: dt,doff,ds,ot,ooff,os,objfunc,dr,or
	real 									:: pi, dx, offset, time,rr,ss,v1,v2, sval, ww
	!===========================================================================

	call sep_init()

    call from_aux("realdata","n1",nt)
    call from_aux("realdata","n2",off_nx)
    call from_aux("realdata","n3",sou_nx)
    call from_aux("realdata","d1",dt)
    call from_aux("realdata","d2",doff)
    call from_aux("realdata","d3",ds)
    call from_aux("realdata","o1",ot)
    call from_aux("realdata","o2",ooff)
    call from_aux("realdata","o3",os)
    call from_param("dx",dx)



	! (/ 1, 2 /)
    ! call from_aux("syndata","n1",nt)
    call from_aux("syndata","n2",rec_nx)
    ! call from_aux("syndata","n3",sou_nx)
    ! call from_aux("syndata","d1",dt)
    call from_aux("syndata","d2",dr)
    ! call from_aux("syndata","d3",ds)
    ! call from_aux("syndata","o1",ot)
    call from_aux("syndata","o2",or)
    ! call from_aux("syndata","o3",os)

	call mem_alloc3d(weight, nt, off_nx, sou_nx)
	call mem_alloc3d(realdata, nt, off_nx, sou_nx)
	call mem_alloc3d(syndata, nt, rec_nx, sou_nx)
	call mem_alloc3d(residual, nt, rec_nx, sou_nx)
	call mem_alloc1d(aa, nt*rec_nx*sou_nx)

	call sep_read(realdata,"realdata")	! Synthetic / modeled data
	call sep_read(syndata,"syndata")	! Clean (observed) data

	rr = MAXVAL(ABS(realdata))
	ss = MAXVAL(ABS(syndata))

	weight = 1.0
	fudge = 10
	zone=15
	v1=1700.0 !1496.21
	v2=1400.0
	pi=3.14159265358979323846264338327950288419


!  Normalize
	realdata = realdata/rr
	syndata  = syndata/ss

!	Calculate the residual
	residual=0.0
	do is=1,sou_nx
		do ir=1, off_nx
			ir2=(((is-1)*ds + (ir-1)*doff)/(dr*dx))+1
			if (ir2<=rec_nx) then
				residual(:,ir2,is) = -1*(realdata(:,ir,is) - syndata(:,ir2,is))  ! For adjoint Born
				! residual(:,ir2,is) = realdata(:,ir,is)	! For migration
				! write(0,*) "ir = ", ir
				! write(0,*) "ir2= ", ir2
			end if
		end do
	end do

	write(0,*) "nt = ", nt
	write(0,*) "doff = ", doff
	write(0,*) "dr = ", dr
	write(0,*) "os = ", os
	write(0,*) "or = ", or
	write(0,*) "sou_nx = ", sou_nx
	write(0,*) "off_nx = ", off_nx
	write(0,*) "ooff = ", ooff
	write(0,*) "dx = ", dx





!  Remove direct arrivals
	do is=1,sou_nx
		do ir=1, off_nx
			offset = ooff + (ir-1)*doff
			t1  = FLOOR(offset/(v1*dt))! Time index of mute window line #1 at reciever 'ir'
			t2  = FLOOR(offset/(v2*dt))+fudge ! Time index of mute window line #2 at reciever 'ir'

			! Pick reciever index
			ir2=(((is-1)*ds + (ir-1)*doff)/(dr*dx))+1

			! Make fading weights
			fade=abs(t2-t1)+(2*zone)
			do it=1,fade
				sval = (2*pi/fade)*(it-1)
				ww = (cos(sval)+1)/2

				! Check if outside array boundaries
				tt = t1-zone-1+it
				if (((tt)>0) .AND. ((tt)<(nt+1))) then
					weight(tt,ir2,is) = ww
					residual(tt,ir2,is) = residual(tt,ir2,is)*ww
				end if

			end do
		end do
	end do


! Send to output
    call to_history("n1",nt,'diff')
    call to_history("n2",rec_nx,'diff')
    call to_history("n3",sou_nx,'diff')
	call to_history("d1",dt,'diff')
 	call to_history("d2",dr,'diff')
	call to_history("d3",ds,'diff')
	call to_history("o1",ot,'diff')
 	call to_history("o2",or,'diff')
	call to_history("o3",os,'diff')
	call sep_write(residual,"diff")


!  Un-Normalize
	residual = residual*ss/rr


!----------  Calculate the objective function value ---------------
	aa=RESHAPE(residual, (/nt*rec_nx*sou_nx/))
	objfunc = DOT_PRODUCT(aa,aa)


! Output the weight mask of the mute
	if (exist_file('weights')) then
		call sep_write(weight,'weights')
 		call to_history("n1", nt,'weights')
 		call to_history("o1", ot,'weights')
 		call to_history("d1", dt,'weights')
 		call to_history("n2", off_nx,'weights')
 		call to_history("o2", ooff,'weights')
 		call to_history("d2", doff,'weights')
 		call to_history("n3", sou_nx,'weights')
 		call to_history("label1", 'time','weights')
 		call to_history("label2", 'rec_x','weights')
	end if


	if (exist_file('objfunc')) then
		call to_history("n1",1					, 'objfunc')
		call to_history("d1",1.0				, 'objfunc')
		call to_history("o1",0.0				, 'objfunc')
		call to_history("label1",'iteration'	, 'objfunc')
		call srite( 'objfunc', objfunc ,4*1)
	end if


	OPEN(UNIT=1,FILE="objfunc.txt",FORM="FORMATTED",STATUS="REPLACE",ACTION="WRITE")
	WRITE(UNIT=1, FMT=*) objfunc
	CLOSE(UNIT=1)


end program



