
program GNHESSIAN
	use sep
	use mem_mod
	use fbi_mod
	use initialize_born
	implicit none

	!===========================================================================
	!	Allocate and Initialize
	!---------------------------------------------------------------------------
	integer									:: isou,rec_dx,rec_ox,sou_dx,sou_ox,nz,nx,nt,sou_nx
	integer                                 :: rec_nx,rec_z,nh,start,finish,startA,finishA,np
	real 									:: dz,dx,dt,oz,ox,ot,cr
	logical									:: adjoint
	real, dimension(:,:,:), allocatable		:: imageA, imageB
	real, dimension(:,:), allocatable		:: vel_in, coor

	!===========================================================================
	real, dimension(:,:,:), allocatable	    :: insrc
	real, dimension(:,:), allocatable		:: insrc_in
	!===========================================================================

	write(0,*) "======================================================"
	call system_clock(count_rate=cr)
	call system_clock(startA)

	write(0,*) "--------- Inside the GNHESSIAN program"
	call init_rtm(nz,nx,nt,nh,dz,dx,dt,oz,ox,ot,sou_nx,sou_dx,sou_ox,rec_nx,rec_dx,rec_ox,rec_z,np)
	write(0,*) "--------- Done with GNHESSIAN initialization routine"

	call mem_alloc3d(imageA, nz, nx,(2*nh+1))
	call mem_alloc3d(imageB, nz, nx,(2*nh+1))
	call mem_alloc2d(vel_in, nz, nx)
	call mem_alloc2d(coor, 2, np)

	call sep_init()
	call sep_read(vel_in,"velmod")
	call model_update(vel_in)
	call from_param("isou", isou)
	call from_param("adjoint", adjoint)


!===========================================================================
	call system_clock(start)
	if (.not.exist_file('insrc' )) write(0,*) "NEED TO HAVE A COMPRESSED SOURCE WAVEFIELD INPUT"
	if (.not.exist_file('coor'  )) write(0,*) "NEED TO HAVE A WAVEFIELD COORDINATES INPUT"

	if (exist_file('insrc' )) then
		write(0,*) "READING IN THE SAVED SOURCE WAVEFIELD"
		call mem_alloc2d(insrc_in, np, nt)
		call mem_alloc3d(insrc, np, nt, 1)
		call sep_read(insrc_in, 'insrc' )
		insrc=0.0
		insrc(:,:,1)=insrc_in
		call sep_read(coor, 'coor' )
		call init_srcwave_fbi(insrc,coor)
		write(0,*) "FINISHED READING IN THE SAVED SOURCE WAVEFIELD"
	end if
	call system_clock(finish)
	write(0,*) "TIME FOR LOADING SAVED SOURC WAVEFIELD = (seconds) ", ((finish-start)/cr)
!===========================================================================


	if (adjoint) then
		write(0,*) " ADJOINT"
		call to_history("n1",nz, 'imageA')
		call to_history("d1",dz, 'imageA')
		call to_history("o1",oz, 'imageA')
		call to_history("n2",nx, 'imageA')
		call to_history("d2",dx, 'imageA')
		call to_history("o2",ox, 'imageA')
		call to_history("label1",'z [km]', 'imageA')
		call to_history("label2",'x [km]', 'imageA')
	else
		write(0,*) "FORWARD"
		call to_history("n1",nz, 'imageB')
		call to_history("d1",dz, 'imageB')
		call to_history("o1",oz, 'imageB')
		call to_history("n2",nx, 'imageB')
		call to_history("n3",(2*nh+1), 'imageB')
		call to_history("d2",dx, 'imageB')
		call to_history("o2",ox, 'imageB')
		call to_history("label1",'z [km]', 'imageB')
		call to_history("label2",'x [km]', 'imageB')
	end if


	!---------------------------------------------------------------------------
	! Linear operator: model perturbation2 => model perturbation1
	if (adjoint) then
		write(0,*) " Adjoint: data perturbation => model perturbation"
		imageA=0.0
		call sep_read(imageB,'imageB')
write(0,*) "sum(imageB) =", sum(imageB)
		call GN_hess(.true., .false., imageA, imageB, isou)
write(0,*) "sum(imageA) =", sum(imageA)
		call sep_write(imageA,'imageA')
	!---------------------------------------------------------------------------
	! Linear operator: model perturbation1 => model perturbation2
	else
		write(0,*) " Forward: model perturbation => full nonlinear data "
		imageB=0.0
		call sep_read(imageA,'imageA')
write(0,*) "sum(imageA) =", sum(imageA)
		call GN_hess(.false., .false., imageA, imageB, isou)
write(0,*) "sum(imageB) =", sum(imageB)
		call sep_write(imageB,'imageB')
	end if
	!---------------------------------------------------------------------------

	call sep_close()

	call system_clock(finishA)
	write(0,*) "TIME FOR FULL GNHESSIAN PROGRAM = (seconds) ", ((finishA-startA)/cr)
end program



