
program GNHESS_INVERSION
	use sep
	use mem_mod
	use fbi_mod
	use initialize2
	use cgstep_mod
	use solver_mod
	implicit none

	!===========================================================================
	!	Allocate and Initialize
	!---------------------------------------------------------------------------
	integer									:: isou,rec_dx,rec_ox,sou_dx,sou_ox,nz,nx
	integer                                 :: nt,sou_nx,rec_nx,rec_z,nh,sou_z
	real 									:: dz,dx,dt,oz,ox,ot
	logical									:: adjoint
	real, dimension(:,:,:), allocatable		:: data
	real, dimension(:,:), allocatable		:: vel_in, deltamodel, RTMgrad
	real, dimension(:), allocatable			:: grad_in, model_out

	!===========================================================================
	real, dimension(:,:,:,:), allocatable	:: insrc
	real, dimension(:,:,:), allocatable		:: insrc_in
	!===========================================================================

	write(0,*) "======================================================"
	write(0,*) "--------- Inside the GN HESS program"
	call init_rtm(nz,nx,nt,nh,dz,dx,dt,oz,ox,ot,sou_nx,sou_dx,sou_ox,rec_nx,rec_dx,rec_ox,rec_z)

	write(0,*) "--------- Done with GN HESS initialization routine"

	call mem_alloc2d(RTMgrad, nz, nx)
	call mem_alloc2d(deltamodel, nz, nx)
	call mem_alloc2d(vel_in, nz, nx)
	call mem_alloc1d(grad_in, nz*nx)
	call mem_alloc1d(model_out, nz*nx)


	call sep_init()
	call sep_read(vel_in,"velmod")
	write(0,*) "PROGRAM:  sum(vel_in) = ", sum(vel_in)
	call model_update(vel_in)
	call from_param("isou", isou)


	call to_history("n1",nz, 'deltamodel')
	call to_history("d1",dz, 'deltamodel')
	call to_history("o1",oz, 'deltamodel')
	call to_history("n2",nx, 'deltamodel')
	call to_history("d2",dx, 'deltamodel')
	call to_history("o2",ox, 'deltamodel')
	call to_history("label1",'z [m]', 'deltamodel')
	call to_history("label2",'x [m]', 'deltamodel')
	
!===========================================================================
	if (exist_file('insrc' )) then
		write(0,*) "READING IN THE SAVED SOURCE WAVEFIELD"
		call mem_alloc3d(insrc_in, nz, nx, nt)
		call mem_alloc4d(insrc, nz, nx, nt, 1)
		call sep_read(insrc_in, 'insrc' )
		insrc=0.0
		insrc(:,:,:,1)=insrc_in
		call init_srcwave_fbi(insrc)
		write(0,*) "FINISHED READING IN THE SAVED SOURCE WAVEFIELD"
	end if
!===========================================================================


!------ DO CG INVERSION FOR DELTAMODEL ------------
	deltamodel=0.0
	call sep_read(RTMgrad,"gradient")
	write(0,*) "sum(gradient) =", sum(RTMgrad)
	grad_in = RESHAPE(RTMgrad, (/ nz*nx /))
	call solver_tiny( model_out, grad_in, GN_Hess_func, cgstep, 15)
	deltamodel = RESHAPE(model_out, (/ nz,nx /))
	write(0,*) "sum(deltamodel) =", sum(deltamodel)
	call sep_write(deltamodel,'deltamodel')


	call sep_close()

end program



