
program GN_inversionCG
  use sep
  use mem_mod
  use fbi_mod
  ! use initialize2
  use cgstep_mod
  use solver_mod
  use Hess_GN_mod_TJ2
  use dot_mod
  implicit none

  !===========================================================================
  ! Allocate and Initialize
  !---------------------------------------------------------------------------
  integer                                 :: isou,rec_dx,rec_ox,sou_dx,sou_ox,nz,nx
  integer                                 :: nt,sou_nx,rec_nx,rec_z,nh,sou_z,niter,rc
  real                                    :: dz,dx,dt,oz,ox,ot,threshold
  logical                                 :: adjoint
  real, dimension(:,:,:), allocatable     :: data
  real, dimension(:,:), allocatable       :: vel_in, deltamodel, RTMgrad
  real, dimension(:), allocatable         :: grad_in, model_out
  character*10240                         :: PBS_inputs, PBS_inputs_path, line_buffer
  integer                                 :: open_stat, read_stat
! =======================================================

  call sep_init()
  call from_history("n1",nz)
  call from_history("n2",nx)

! !===========================================================================
  write(0,*) "======================================================"
  write(0,*) "--------- Inside the GN HESS INVERSION program"
!   INITIALIZE THE PBS INPUTS
    PBS_inputs_path = ' '
    call from_param('PBS_inputs',PBS_inputs_path)
    open(file=PBS_inputs_path,unit=22,form='formatted', status='old', iostat=open_stat)
    if (open_stat .ne. 0 ) then
        write(0,*) "Cannot read the input file for PBS_inputs"
        stop
    endif
    read_stat=0
    read(unit=22,fmt='(A)',iostat=read_stat) PBS_inputs


    write(0,*)"Initialize routines"
    !Gauss-Newton Hessian computation routines
    call Hess_GN_init2(nz, nx, 1, PBS_inputs)

! !===========================================================================

  call mem_alloc2d(RTMgrad, nz, nx)
  call mem_alloc2d(deltamodel, nz, nx)
  call mem_alloc1d(grad_in, nz*nx)
  call mem_alloc1d(model_out, nz*nx)

  call to_history("n1",nz, 'deltamodel')
  call to_history("d1",dz, 'deltamodel')
  call to_history("o1",oz, 'deltamodel')
  call to_history("n2",nx, 'deltamodel')
  call to_history("d2",dx, 'deltamodel')
  call to_history("o2",ox, 'deltamodel')
  call to_history("label1",'z [m]', 'deltamodel')
  call to_history("label2",'x [m]', 'deltamodel')

! Read input parameters
  call from_param("niter",niter,15)
  call from_param("threshold",threshold,10.0)
  call solver_init(threshold)
  if (.not.exist_file('hess_in_path')) write(0,*) "You're missing the 'hess_in_path' input parameter."
  if (.not.exist_file('hess_out_path')) write(0,*) "You're missing the 'hess_out_path' input parameter."


!------ DO CG INVERSION FOR DELTAMODEL ------------
  deltamodel=0.0
  call sep_read(RTMgrad)
  grad_in = RESHAPE(RTMgrad, (/ nz*nx /))
  write(0,*) "sum(grad_in) =", sum(grad_in)
  call solver_tiny( model_out, grad_in, Hess_GN_call2, cgstep, niter)
  write(0,*) "sum(model_out) =", sum(model_out)
  deltamodel = RESHAPE(model_out, (/ nz,nx /))
  call sep_write(deltamodel)


  ! Closing routine
  call Hess_GN_close()
  deallocate(RTMgrad,model_out,grad_in)
  call sep_close()
  call exit (0)

end program



