!Program for inverting the GN FWI Hessian matrix with MKL RCI CG library
program GN_inversion

	use sep
! =======================================================
	use Hess_GN_mod_TJ2
! =======================================================
	use dot_mod
	implicit none
	include 'mkl_rci.fi'
	
	integer                             :: nz, nx, ny, rc, sreed, wc, srite
	real                                :: dz, dx, dy, oz, ox, oy, rnorm, threshold, lambda, scale
	double precision                    :: normalize
	real, dimension(:),allocatable      :: model_in, model_out, gradient
	!MKL RCI CG variables
	!We want to invert Hx = b
	integer                             :: n 			! model size
	double precision, allocatable       :: x(:) 		! approximate solution
	double precision, allocatable       :: b(:)			! right-hand side of equation 
	integer                             :: RCI_request 	! variable for handling RCs
	integer                             :: ipar(128) 	! Integer parameters
	double precision                    :: dpar(128) 	! Double precision parameters
	double precision, allocatable       :: tmp(:,:)		! temporary array for computation
	integer                             :: itercount, niter
    logical                             :: lberg
! =======================================================
    character*10240                     :: PBS_inputs, PBS_inputs_path, line_buffer
    integer                             :: open_stat, read_stat
! =======================================================


  	call sep_init()
	
  	call from_history("n1", nz)
  	call from_history("o1", oz)
   	call from_history("d1", dz)
  	call from_history("n2", nx)
  	call from_history("o2", ox)
  	call from_history("d2", dx)
  	call from_history("n3", ny)
  	call from_history("o3", oy)
    call from_history("d3", dy)
    call from_par("niter",niter)
! =======================================================
    call from_par("lberg",lberg,.FALSE.)
    call from_par("lambda",lambda,0.001)
    call from_par("threshold",threshold,10.0)
! =======================================================

    write(0,*)"Gauss-Newton Hessian inversion with MKL RCI CG routine"
  	
	!Initialize arrays for inversion
  	n = nz*nx*ny ! model size
  	allocate(x(n),b(n)); x = 0.; b = 0.
  	allocate(tmp(n,3)); tmp = 0.
  	
  	
	!allocate memory for Hessian application
  	allocate(model_in(nz*nx*ny),model_out(nz*nx*ny),gradient(nz*nx*ny))
	!reading the gradient
  	rc = sreed("in",gradient,4*nz*nx*ny)
  	b = -gradient !Inverting the system H m = -g
	
    !writing history for residual and norm(r)
    if (exist_file('residual')) then
        call to_history("n1", nz,"residual")
        call to_history("o1", oz,"residual")
        call to_history("d1", dz,"residual")
        call to_history("n2", nx,"residual")
        call to_history("o2", ox,"residual")
        call to_history("d2", dx,"residual")
        call to_history("n3", ny,"residual")
        call to_history("o3", oy,"residual")
        call to_history("d3", dy,"residual")
        call to_history("o4", 1., "residual")
        call to_history("d4", 1., "residual")
    end if
  	
    if (exist_file('rnorm')) then
        call to_history("n2", 1, "rnorm")
        call to_history("o1", 1., "rnorm")
        call to_history("d1", 1., "rnorm")
    end if

    if (exist_file('intermediate')) then
        call to_history("n1", nz,"intermediate")
        call to_history("o1", oz,"intermediate")
        call to_history("d1", dz,"intermediate")
        call to_history("n2", nx,"intermediate")
        call to_history("o2", ox,"intermediate")
        call to_history("d2", dx,"intermediate")
        call to_history("n3", niter,"intermediate")
    end if

  	!History for output
  	call to_history("n1", nz)
  	call to_history("o1", oz)
   	call to_history("d1", dz)
  	call to_history("n2", nx)
  	call to_history("o2", ox)
  	call to_history("d2", dx)
  	call to_history("n3", ny)
  	call to_history("o3", oy)
  	call to_history("d3", dy)
  	
	
  	
! =======================================================
!   INITIALIZE THE PBS INPUTS
    PBS_inputs_path = ' '
    call from_param('PBS_inputs',PBS_inputs_path)
    open(file=PBS_inputs_path,unit=22,form='formatted', status='old', iostat=open_stat)
    if (open_stat .ne. 0 ) then
        write(0,*) "Cannot read the input file for PBS_inputs"
        stop
    endif
    read_stat=0
    read(unit=22,fmt='(A)',iostat=read_stat) PBS_inputs


    write(0,*)"Initialize routines"
    !Gauss-Newton Hessian computation routines
    call Hess_GN_init2(nz, nx, 1, PBS_inputs)
! =======================================================

  	!Initialize MKL inversion routine
  	call dcg_init(n, x, b, RCI_request, ipar, dpar, tmp)
  	!Change parameters to your inversion flavor
  	
  	!Check variables for inversion
  	call dcg_check(n, x, b, RCI_request, ipar, dpar, tmp)
  	if (RCI_request .ne. 0 ) call erexit("Not passing dcg_check")
  	!If the starting model is equal zero
  	itercount = -1

  	
  	!Running the solver

  	do 
  		call dcg(n, x, b, RCI_request, ipar, dpar, tmp)
  		if (RCI_request.eq.0) then !Apply the operator
  			write(0,*)"Solution found"
  			exit	
  		elseif(RCI_request.eq.1)then
			if(itercount .ge. 0)then
  				write(0,*)"Applying the Hessian"
  				model_in = tmp(1:n,1)
write(0,*)"sum(model_in) = ", sum(model_in)
            call Hess_GN_op2(.false., .false., model_in, model_out)
! write(0,*)" Finished Hess_GN_call "
            ! ==========================================
            if (lberg) then
                write(0,*)"Doing Levenberg Hessian stabilization. Lambda value = ", lambda
                scale=maxval(model_out)/maxval(model_in)
                model_out = model_out + model_in*lambda*scale
            end if
          ! ===========================================
write(0,*)"sum(model_out) = ", sum(model_out)
  				tmp(1:n,2) = model_out
			endif
  		elseif(RCI_request.eq.2) then !Stop test
  			call dcg_get(n, x, b, RCI_request, ipar, dpar, tmp, itercount)
  			!writing residual and norm of the residual
  			if(itercount .ne. 0) then
if (exist_file('residual')) call to_history("n4", itercount,"residual")
if (exist_file('rnorm')) call to_history("n1", itercount,"rnorm")
  				gradient = tmp(:,3) !converting to single precision
  				if(itercount .eq. 1) normalize = sum(tmp(:,3)*tmp(:,3))
  				rnorm = sum(tmp(:,3)*tmp(:,3))/normalize*100.0
  				write(0,*)"Residual norm percentage:",rnorm
if (exist_file('residual')) call sep_write(gradient,"residual")
                if (exist_file('rnorm')) then 
                    wc = srite("rnorm",rnorm,4)
                end if
        
                !writing approximate solution
                gradient = x; !converting to single precision
if (exist_file('intermediate')) call sep_write(gradient,"intermediate")

                if (rnorm<threshold) then
                    write(0,*)"-------------Stopped inversion. Residual is below ", threshold
                    exit
                end if
  			endif

  			write(0,*)"Checking for stop, at iter:", itercount
  			if(itercount .ge. niter) then
  				write(0,*)"Max number of iterations reached: Stopping solver"
  				exit
  			endif

  		elseif(RCI_request.eq.3) then ! preconditioner
  			write(0,*)"Applying preconditioner (if present)"
  			cycle
  		elseif(RCI_request.lt.0) then
  			write(0,*)"Error in running the inversion: negative flag from solver",RCI_request
  			exit
  		else
  			write(0,*)"Unknown flag: stopping"
  			exit
  		endif
  	enddo
  
	
    ! Make the same sign as the input image/gradient
    gradient = -1.0*gradient
    wc = srite("out",gradient,4*nz*nx*ny)


  	!Close Hessian computation
  	call Hess_GN_close()
	
  		
  	deallocate(gradient,model_out,model_in)
  	deallocate(tmp)
  	deallocate(b,x)
  	call sep_close()
  	call exit (0)

end program
