!Module for applying Gauss-Newton FWI Hessian (i.e., (Born*)(Born))
!Takes a non-extended model, runs Born followed by RTM, and return a non-extended model
module Hess_GN_mod_TJ2
	use sep
	implicit none

	!Module variables
	character*55,private						  :: wavelet=' ',sources=' ',recs=' ',wavefields=' ',wantwavefield=' ',readwavefield=' ',wavefield=' ',dt=' ',threads=' ',background_vel=' ',padded=' '
	character*1024,private						  :: command
	character*10240,private						  :: pbs_inputs
	integer,private								  :: nz, nx, ny
	real, dimension(:,:,:), allocatable, private  :: model_tmp, model_tmp2
	real, private 								  :: oz, ox, oy, dz, dx, dy
	contains
	

	
!===============================================================================	
!===============================================================================


	!  Initialization routine
	subroutine Hess_GN_init2(nz_in, nx_in, ny_in, pbs_inputs_in)
		implicit none
		character*10240			:: pbs_inputs_in
		integer					:: nz_in,nx_in,ny_in
		pbs_inputs = trim(pbs_inputs_in)
		!Initialize all the variables
		nz = nz_in
		nx = nx_in
		ny = ny_in
		allocate(model_tmp(nz,nx,ny)); model_tmp = 0.
		allocate(model_tmp2(nz,nx,ny)); model_tmp2 = 0.
		call sep_init()
		return
	end subroutine





  	function Hess_GN_call2(adj,add,mm,dd) result(stat)
    	logical,intent(in)   	:: adj,add
   		real,dimension(:)  		:: mm
   		real,dimension(:)    	:: dd
   		integer                 :: stat
    	call Hess_GN_op2(adj,add,mm,dd)
    	stat=0
  	end function




	
	!Operator routine
	subroutine Hess_GN_op2(adj, add, mm1, mm2)
	
		!Input-Output variables
		logical, intent(in)								:: adj, add
		real, dimension(nz,nx,ny), intent(inout)		:: mm1 
		real, dimension(nz,nx,ny), intent(inout)		:: mm2 
		!Internal variables
		integer											:: wc, srite, rc, sreed, sk, sseek
		character(len=2048) 							:: command
		character(len=500) 								:: cmd1
	    character(len=2048) 							:: format_string
        format_string = "(A50,A1524)"
		
    	if(adj) then !Adjoint=Forward
			! write(0,*)">>>>>>>>>>>>>>>ADJOINT HESSIAN"
			if(.not. add) mm1 = 0. !Zero-out the data	
		 	call to_history("n1", nz,"hess_in_path")
		  	call to_history("o1", 0,"hess_in_path")
		  	call to_history("d1", 1,"hess_in_path")
		  	call to_history("n2", nx,"hess_in_path")
		  	call to_history("o2", 0,"hess_in_path")
		  	call to_history("d2", 1,"hess_in_path")
			wc=srite("hess_in_path",mm2,nz*nx*ny*4)
	        write (command,format_string) "hessian.py  ", pbs_inputs
			! write(0,*)" command = ", command
			call system(command)
			rc = sreed('hess_out_path',model_tmp,nz*nx*ny*4)
			mm1 = mm1 + model_tmp
			sk = sseek("hess_in_path",0,0)
			sk = sseek("hess_out_path",0,0)
		else	     !Forward
			! write(0,*)">>>>>>>>>>>>>>> FORWARD HESSIAN"
			if(.not. add) mm2 = 0. !Zero-out the data	
		 	call to_history("n1", nz,"hess_in_path")
		  	call to_history("o1", 0,"hess_in_path")
		  	call to_history("d1", 1,"hess_in_path")
		  	call to_history("n2", nx,"hess_in_path")
		  	call to_history("o2", 0,"hess_in_path")
		  	call to_history("d2", 1,"hess_in_path")
			wc=srite("hess_in_path",mm1,nz*nx*ny*4)
	        write (command,format_string) "hessian.py  ", pbs_inputs
			! write(0,*)" command = ", command
			call system(command)
			rc = sreed('hess_out_path',model_tmp,nz*nx*ny*4)
			mm2 = mm2 + model_tmp
			sk = sseek("hess_in_path",0,0)
			sk = sseek("hess_out_path",0,0)
		endif

		return
	end subroutine
	


	subroutine Hess_GN_close()
		if(allocated(model_tmp)) deallocate(model_tmp)
	end subroutine

end module


