!Module for applying Gauss-Newton FWI Hessian (i.e., (Born*)(Born))
!Takes a non-extended model, runs Born followed by RTM, and return a non-extended model
module Hess_full_mod_1

  use sep

  implicit none
  !Module variables
  character*512,private					:: command
  integer,private						:: nz, nx, ny
  real, dimension(:,:,:), allocatable	:: model_tmp
  
  
  contains
	
	!Initialization routine
	subroutine Hess_full_init(nz_in,nx_in,ny_in)
		
		implicit none
		integer					:: nz_in,nx_in,ny_in

		!Initialize all the variables
		call sep_init()
		nz = nz_in
		nx = nx_in
		ny = ny_in
		allocate(model_tmp(nz,nx,ny)); model_tmp = 0.

		return
	end subroutine Hess_full_init

  	subroutine Hess_full_call(adj,add,mm,dd)
    	logical,intent(in)   	:: adj,add
   		real,dimension(:)  		:: mm
   		real,dimension(:)    	:: dd
    		call Hess_full_op(adj,add,mm,dd)
  	end subroutine Hess_full_call

	
	!Operator routine
	subroutine Hess_full_op(adj, add, mm1, mm2)
	
		!Input-Output variables
		logical, intent(in)								:: adj, add
		real, dimension(nz,nx,ny), intent(inout)		:: mm1 
		real, dimension(nz,nx,ny), intent(inout)		:: mm2 
		!Internal variables
		integer								:: wc, srite, rc, sreed, sk, sseek
		
    	if(adj) then !Adjoint=Forward
        	if (.not. add) mm1 = 0. !Zero-out the model
        	wc=srite("tmp/model_in.H",mm2,nz*nx*ny*4)
			sk = sseek("tmp/model_in.H",0,0)
			
			!GAUSS-NEWTON HESSIAN COMPONENT
			command = 'bin/Bornfast_fw.x <background.H wavelet=wavelet.H deltavel=tmp/model_in.H sources=Par/sources.p recs=Par/receivers.p wavefields=wavefield_fw_background.H readwavefield=1 wavefield=tmp/pertwavefield_full_hess.H wantwavefield=0 dt=1e-3 padded=1 threads=16 >tmp/born_data_tmp.H'
			call system(command)
			command = 'bin/Bornfast_adj.x <background.H wavelet=wavelet.H deltadata=tmp/born_data_tmp.H sources=Par/sources.p recs=Par/receivers.p  wavefields=wavefield_fw_background.H readwavefield=0 dt=1e-3 padded=1 threads=16 >tmp/born_model_tmp.H'
			call system(command)
			call system('Transp < tmp/born_model_tmp.H reshape=2,4 | Stack | Window n1=1 >tmp/model_tmp.H  out=tmp/model_tmp.H@')
			rc = sreed('tmp/model_tmp.H',model_tmp,nz*nx*ny*4)
			sk = sseek("tmp/model_tmp.H",-4*nz*nx*ny,1)
			mm1 = mm1 + model_tmp
			
			!WEMVA-like HESSIAN COMPONENT
			command = 'bin/Hessian_wemva.x <background.H deltadata=true_data.H image=tmp/model_in.H padded=1 wavelet=wavelet.H sources=Par/sources.p recs=Par/receivers.p readwavefield=0 wavefields=wavefield_fw_background.H wavefieldr=wavefieldr_background.H scatfields=tmp/pertwavefield_full_hess.H dt=1e-3 threads=16 >tmp/wemva_model.H'
			call system(command)
			call system('Transp < tmp/wemva_model.H reshape=2,4 | Stack | Window n1=1 >tmp/model_wemva_tmp.H out=tmp/model_wemva_tmp.H@')
			rc = sreed('tmp/model_wemva_tmp.H',model_tmp,nz*nx*ny*4)
			sk = sseek("tmp/model_wemva_tmp.H",-4*nz*nx*ny,1)
			mm1 = mm1 + model_tmp
			

		else	     !Forward
			if(.not. add) mm2 = 0. !Zero-out the data	
			wc=srite("tmp/model_in.H",mm1,nz*nx*ny*4)
			sk = sseek("tmp/model_in.H",0,0)
			
			!GAUSS-NEWTON HESSIAN COMPONENT
			command = 'bin/Bornfast_fw.x <background.H wavelet=wavelet.H deltavel=tmp/model_in.H sources=Par/sources.p recs=Par/receivers.p wavefields=wavefield_fw_background.H readwavefield=0 wavefield=tmp/pertwavefield_full_hess.H wantwavefield=1 dt=1e-3 padded=1 threads=16 >tmp/born_data_tmp.H verbose=1'
			call system(command)
			command = 'bin/Bornfast_adj.x <background.H wavelet=wavelet.H deltadata=tmp/born_data_tmp.H sources=Par/sources.p recs=Par/receivers.p  wavefields=wavefield_fw_background.H readwavefield=0 dt=1e-3 padded=1 threads=16 >tmp/born_model_tmp.H verbose=1'
			call system(command)
			call system('Transp < tmp/born_model_tmp.H reshape=2,4 | Stack | Window n1=1 >tmp/model_tmp.H  out=tmp/model_tmp.H@')
			rc = sreed('tmp/model_tmp.H',model_tmp,nz*nx*ny*4)
			sk = sseek("tmp/model_tmp.H",-4*nz*nx*ny,1)
			mm2 = mm2 + model_tmp
			
			!WEMVA-like HESSIAN COMPONENT
			command = 'bin/Hessian_wemva.x <background.H deltadata=true_data.H image=tmp/model_in.H padded=1 wavelet=wavelet.H sources=Par/sources.p recs=Par/receivers.p readwavefield=0 wavefields=wavefield_fw_background.H wavefieldr=wavefieldr.H scatfields=tmp/pertwavefield_full_hess.H dt=1e-3 threads=16 >tmp/wemva_model.H verbose=1'
			call system(command)
			call system('Transp < tmp/wemva_model.H reshape=2,4 | Stack | Window n1=1 >tmp/model_wemva_tmp.H out=tmp/model_wemva_tmp.H@')
			rc = sreed('tmp/model_wemva_tmp.H',model_tmp,nz*nx*ny*4)
			sk = sseek("tmp/model_wemva_tmp.H",-4*nz*nx*ny,1)
			mm2 = mm2 + model_tmp
			
				
				
		endif
		
		return
	end subroutine Hess_full_op
	
	subroutine Hess_full_close()
		if(allocated(model_tmp)) deallocate(model_tmp)
	end subroutine Hess_full_close
	

end module Hess_full_mod_1