program MOD_VEL2
	use sep
	use mem_mod
	use calc_mod
	use initialize1
	implicit none

	real                               	:: Vsalt, p1, p2, dz, dx, oz, ox, alpha
	integer								:: nz, nx
	logical								:: buildonly,drlseonly
	real, dimension(:,:), allocatable	:: phi1, phi2, phi_grad1, phi_grad2
	real, dimension(:,:), allocatable	:: vel_back, vel_full, tomo_grad, drlse_term1, drlse_term2

	call sep_init()

	call from_param("q1", Vsalt, 4500.0)
	call from_param("n1", nz)
	call from_param("n2", nx)
	call from_param("d1", dz)
	call from_param("d2", dx)
	call from_param("o1", oz)
	call from_param("o2", ox)
	call from_param("p1", p1)
	call from_param("p2", p2)
	call from_param("alpha", alpha)
	call from_param("buildonly", buildonly)
	call from_param("drlseonly", drlseonly)

!---------------- Allocate -------------------------------
	call mem_alloc2d(phi1, nz, nx)
	call mem_alloc2d(phi2, nz, nx)
	call mem_alloc2d(phi_grad1, nz, nx)
	call mem_alloc2d(phi_grad2, nz, nx)
	call mem_alloc2d(vel_back, nz, nx)
	call mem_alloc2d(vel_full, nz, nx)
	call mem_alloc2d(tomo_grad, nz, nx)
	call mem_alloc2d(drlse_term1, nz, nx)
	call mem_alloc2d(drlse_term2, nz, nx)


!------------ Read in values ------------------------------
	call sep_read(phi1, "phi1")
	call sep_read(vel_back, "vel_back")


!------------ If only DRLSE ------------------------------
	if (drlseonly) then
		write(0,*) " -------------------------------------------"
		write(0,*) " DRLSE ONLY; NO GRADIENT"
		phi_grad1	= 0.0
		phi_grad2	= 0.0
		tomo_grad	= 0.0
		call sep_read(drlse_term1, "drlse_term1")
		if (exist_file('drlse_term2')) call sep_read(drlse_term2, "drlse_term2")
	else
!------------ If only assembling ------------------------------
		if (buildonly) then
			write(0,*) " -------------------------------------------"
			write(0,*) " BUILD MODEL ONLY; NO GRADIENT"
			phi_grad1	= 0.0
			drlse_term1	= 0.0
			phi_grad2	= 0.0
			drlse_term2	= 0.0
			tomo_grad	= 0.0
		else	
			if (exist_file('phi_grad1'  )) then
				call sep_read(phi_grad1, "phi_grad1")
			else
				phi_grad1=0.0
			end if
			if (exist_file('tomo_grad'  )) then
				call sep_read(tomo_grad, "tomo_grad")
			else
				tomo_grad=0.0
			end if
			if (exist_file('drlse_term1')) then
				call sep_read(drlse_term1, "drlse_term1")
			else
				drlse_term1=0.0
			end if
			!------------ Only use a second levelset if it exists ----------------
			if (exist_file('phi_grad2'  )) call sep_read(phi_grad2, "phi_grad2")
			if (exist_file('drlse_term2')) call sep_read(drlse_term2, "drlse_term2")
		end if
	end if


!------------ Normalize the phi_grad ------------------------------
	if (maxval(phi_grad1)>0.0) then
		phi_grad1 = phi_grad1/(maxval(phi_grad1))
	else
		phi_grad1 = 0.0
	end if


	! drlse_term1 =  drlse_term1*4444.0
	write(0,*) "---------------------------------------------------------"
	write(0,*) "maxval(phi1) BEFORE = ", maxval(phi1)
	write(0,*) "sum(drlse_term1) = ", sum(drlse_term1)
	write(0,*) "sum(phi_grad1) = ", sum(phi_grad1)
	write(0,*) "p1 = ", p1

!------------ Calculate implicit surfaces ------------------------------
	phi1 = phi1 + phi_grad1*p1 + drlse_term1

	write(0,*) "sum(phi_grad1*p1) = ", sum(phi_grad1*p1)
	write(0,*) "p1 = ", p1


	vel_back = vel_back + tomo_grad*alpha

	if (exist_file('phi2'   )) then
		call sep_read(phi2, "phi2")
		phi2 = phi2 + phi_grad2*p2 + drlse_term2
	else
		phi2 = phi1
	end if

!-----------  Calculate full-velocity model  ----------------------------
	call calc_vel_init( nz, nx)
	call calc_velocity(phi1, phi2, vel_back, vel_full, Vsalt)

! !-----------  Output L1 norm of full-velocity model  ----------------------------
! 	write(0,*) "sum(vel_full) AFTER = ", sum(vel_full)
! 	OPEN(UNIT=1,FILE="MODVELnorm.txt",FORM="FORMATTED",STATUS="REPLACE",ACTION="WRITE")
! 	WRITE(UNIT=1, FMT=*) sum(vel_full)
! 	CLOSE(UNIT=1)

! 	!-----------  Output L1 norm of full-velocity model  ----------------------------
	write(0,*) "maxval(phi1) AFTER = ", maxval(phi1)
! 	OPEN(UNIT=1,FILE="phi1norm.txt",FORM="FORMATTED",STATUS="REPLACE",ACTION="WRITE")
! 	WRITE(UNIT=1, FMT=*) sum(phi1)
! 	CLOSE(UNIT=1)

!-----------  Write outputs ------------------------------------------------
	if (exist_file('vel_full_test')) then
   		call to_history("n1",nz, 'vel_full_test')
   		call to_history("n2",nx, 'vel_full_test')
		! call to_history("n3",niter, 'vel_full_test')
		call to_history("d1",dz, 'vel_full_test')
		call to_history("d2",dx, 'vel_full_test')
		call to_history("o1",oz, 'vel_full_test')
		call to_history("o2",ox, 'vel_full_test')
		call to_history("label1",'z [km]', 'vel_full_test')
		call to_history("label2",'x [km]', 'vel_full_test')
		! call to_history("label3",'iteration', 'vel_full_test')
		call sep_write(vel_full,	'vel_full_test'		)
	end if


	if (exist_file('phi1_out')) then
   		call to_history("n1",nz, 'phi1_out')
   		call to_history("n2",nx, 'phi1_out')
		call to_history("d1",dz, 'phi1_out')
		call to_history("d2",dx, 'phi1_out')
		call to_history("o1",oz, 'phi1_out')
		call to_history("o2",ox, 'phi1_out')
		call to_history("label1",'z [km]', 'phi1_out')
		call to_history("label2",'x [km]', 'phi1_out')
		call sep_write(phi1,	 'phi1_out')
	end if

	if (exist_file('phi2_out')) then
   		call to_history("n1",nz, 'phi2_out')
   		call to_history("n2",nx, 'phi2_out')
		call to_history("d1",dz, 'phi2_out')
		call to_history("d2",dx, 'phi2_out')
		call to_history("o1",oz, 'phi2_out')
		call to_history("o2",ox, 'phi2_out')
		call to_history("label1",'z [km]', 'phi2_out')
		call to_history("label2",'x [km]', 'phi2_out')
		call sep_write(phi2,	 'phi2_out')
	end if

	if (exist_file('vel_back_out')) then
   		call to_history("n1",nz, 'vel_back_out')
   		call to_history("n2",nx, 'vel_back_out')
		call to_history("d1",dz, 'vel_back_out')
		call to_history("d2",dx, 'vel_back_out')
		call to_history("o1",oz, 'vel_back_out')
		call to_history("o2",ox, 'vel_back_out')
		call to_history("label1",'z [km]', 'vel_back_out')
		call to_history("label2",'x [km]', 'vel_back_out')
		call sep_write(vel_back, 'vel_back_out')
	end if


write(0,*) "MOD_VEL program complete"
end program