
program MUTE_DA

!===========================================================================
!		This program does a residual difference between a real and synthetic
!		dataset that may or may not have the same number of recievers

!		The real data input must have the ds and dr values in [m] units, not 
!		sampling units. 
!===========================================================================


	use sep
	use mem_mod
	use fbi_mod
	use initialize2
	implicit none

	!===========================================================================
	!	Allocate and Initialize
	!---------------------------------------------------------------------------
	integer									:: i,j,k,nt,ni,sou_nx,off_nx,is, fudge
	integer									:: ir,ir2,ii,dtt,fade,lower,upper, t1,t2, it, tt
	real, dimension(:,:,:), allocatable		:: residual, syndata, weight
	real, dimension(:), allocatable			:: aa
	real 									:: dt,doff,ds,ot,ooff,os,objfunc,dr,or,shift1, shift2
	real 									:: pi, dx, offset, time,rr,ss,v1,v2, sval, ww
	logical									:: windowed, reverse
	!===========================================================================

	call sep_init()

    call from_history("n1",nt)
    call from_history("n2",off_nx)
    call from_history("n3",sou_nx)
    call from_history("d1",dt)
    call from_history("d2",doff) ! Needs to be in meters
    call from_history("d3",ds)
    call from_history("o1",ot)
    call from_history("o2",ooff) ! Needs to be in meters
    call from_history("o3",os)

    call from_param("windowed",windowed,.TRUE.)
    call from_param("reverse",reverse,.FALSE.)


    call from_param("fudge",fudge,10)
    call from_param("fade",fade,15)
    call from_param("v1",v1,160000.0)
    call from_param("v2",v2,1400.0)
    call from_param("shift1",shift1,0.0)
    call from_param("shift2",shift2,shift1)

	call mem_alloc3d(weight, nt, off_nx, sou_nx)
	call mem_alloc3d(residual, nt, off_nx, sou_nx)
	call mem_alloc1d(aa, nt*off_nx*sou_nx)

	call sep_read(residual)	! Synthetic / modeled data


	pi=3.14159265358979323846264338327950288419
	write(0,*) "nt = ", nt
	write(0,*) "doff = ", doff
	write(0,*) "dr = ", dr
	write(0,*) "os = ", os
	write(0,*) "or = ", or
	write(0,*) "sou_nx = ", sou_nx
	write(0,*) "off_nx = ", off_nx
	write(0,*) "ooff = ", ooff
	write(0,*) "dx = ", dx




	weight=1.0
!  Remove direct arrivals
	do is=1,sou_nx
		do ir=1, off_nx
			offset = ooff + (ir-1)*doff
			t1  = shift1 + FLOOR(offset/(v1*dt))! Time index of mute window line #1 at reciever 'ir'
			t2  = shift2 + FLOOR(offset/(v2*dt))+fudge ! Time index of mute window line #2 at reciever 'ir'

			! Top fade out
			if (windowed) then
			! write(0,*) "Windowed"
				do it=1,fade+1
					tt=t1-fade-1+it
					if (((tt)>0) .AND. ((tt)<(nt+1))) then
						ww = 1-(REAL(it-1)/REAL(fade+1))
						weight(tt,ir,is) = ww
						! residual(tt,ir,is) = residual(tt,ir,is)*ww
					end if
				end do
			else
			! write(0,*) "Not windowed"
				do it=1,nt
					if (it<(t1+1)) then
						ww = 0
						weight(it,ir,is) = ww
						! residual(it,ir,is) = residual(it,ir,is)*ww
					end if
				end do
			end if


			! Bottom fade out
			do it=1,fade+1
				tt=t2+fade+1-it
				if (((tt)>0) .AND. ((tt)<(nt+1))) then
					ww = 1-(REAL(it-1)/REAL(fade+1))
					weight(tt,ir,is) = ww
					! residual(tt,ir,is) = residual(tt,ir,is)*ww
				end if
			end do

			! Middle area zeroing
			do it=1,nt
				if ((it<t2) .AND. (it>t1)) then
					ww=0.0
					weight(it,ir,is) = ww
					! residual(it,ir,is) = residual(it,ir,is)*ww
				end if
			end do

		end do
	end do

! Reverse the weighting if necessary
	if (reverse) then
		weight = weight*(-1.0) + 1
	end if

! Apply the weighting
	residual = residual*weight


! Send to output
if (exist_file('muted')) then
    call to_history("n1",nt,"muted")
    call to_history("n2",off_nx,"muted")
    call to_history("n3",sou_nx,"muted")
    call to_history("d1",dt,"muted")
    call to_history("d2",doff,"muted")
    call to_history("d3",ds,"muted")
    call to_history("o1",ot,"muted")
    call to_history("o2",ooff,"muted")
    call to_history("o3",os,"muted")
	call sep_write(residual,"muted")
end if


! Output the weight mask of the mute
if (exist_file('weights')) then
	call sep_write(weight,'weights')
	call to_history("n1", nt,'weights')
	call to_history("o1", ot,'weights')
	call to_history("d1", dt,'weights')
	call to_history("n2", off_nx,'weights')
	call to_history("o2", ooff,'weights')
	call to_history("d2", doff,'weights')
	call to_history("n3", sou_nx,'weights')
	call to_history("label1", 'time','weights')
	call to_history("label2", 'rec_x','weights')
end if


end program



