
program REMOVE_SALT
use sep
implicit none

!===============================================================
real, dimension(:,:), allocatable	:: mask, smoothmodel, aftermask, vel_back_in
integer								:: nz, nx, iz, ix, count, fudge
integer								:: ii, numpass, k, l, j, i, top, bottom, left, right, radius, countr
real								:: cutvel, dz, dx, oz, ox, val


! Allocate the internal arrays / dimensions
call sep_init()
call from_param("cutvel",cutvel,4000.0)  ! Salt velocity
call from_param("stenc_rad",radius,75)
call from_param("smpass",numpass,20)
call from_param("fudge",fudge,3)

call from_aux("vel_in","n1",nz)
call from_aux("vel_in","n2",nx)
call from_aux("vel_in","d1",dz)
call from_aux("vel_in","d2",dx)
call from_aux("vel_in","o1",oz)
call from_aux("vel_in","o2",ox)

allocate(mask(nz,nx), smoothmodel(nz,nx), vel_back_in(nz,nx), aftermask(nz,nx))


! Check for background velocity model input
if (exist_file('vel_in')) then
	call sep_read(vel_back_in,'vel_in')
else
	write(0,*) "You need to provide a background velocity input"
end if



! Build mask and zero out the background salt
mask=0.0
write(0,*) "salt cutting vel =", cutvel
do iz=1,nz
	do ix=1,nx
		if (vel_back_in(iz,ix)>cutvel) then

			top    = iz - fudge
			bottom = iz + fudge
			left   = ix - fudge
			right  = ix + fudge
			if (top    < 1) top  = 1
			if (left   < 1) left = 1
			if (bottom > size(vel_back_in,1)) bottom = size(vel_back_in,1)
			if (right  > size(vel_back_in,2)) right  = size(vel_back_in,2)

			do i=top,bottom
				do j=left,right
					vel_back_in(i,j)=0.0
					mask(i,j)=1.0
				end do
			end do
		end if
	end do
end do
aftermask = vel_back_in


do ii=1,numpass
	write(0,*) "SMOOTHING PASS  ",ii
	do j=1, size(vel_back_in,2)        ! x axis
		do i=1, size(vel_back_in,1)    ! z axis

			if (mask(i,j)==1.0) then

				top    = i - radius
				bottom = i + radius
				left   = j - radius
				right  = j + radius

				if (top    < 1) top  = 1
				if (left   < 1) left = 1

				if (bottom > size(vel_back_in,1)) bottom = size(vel_back_in,1)
				if (right  > size(vel_back_in,2)) right  = size(vel_back_in,2)

				countr = 0
				val    = 0.0
				do k=top, bottom
					do l=left, right
						if (vel_back_in(k,l)>0.0) then
							countr = countr + 1
							val = val + vel_back_in(k,l)
						end if
					end do
				end do
				smoothmodel(i,j) = real(val)/real(countr)
			else
				smoothmodel(i,j) = vel_back_in(i,j)
			end if
			end do
	end do
	vel_back_in=smoothmodel
end do



! Output the mask of what salt was removed
	if (exist_file('aftermask')) then
		call sep_write(aftermask,'aftermask')
 		call to_history("n1", nz,'aftermask')
 		call to_history("o1", oz,'aftermask')
 		call to_history("d1", dz,'aftermask')
 		call to_history("n2", nx,'aftermask')
 		call to_history("o2", ox,'aftermask')
 		call to_history("d2", dx,'aftermask')
 		call to_history("label1", 'z','aftermask')
 		call to_history("label2", 'x','aftermask')
	end if




! Output the salt free background model
	if (exist_file('saltfree')) then
		call sep_write(smoothmodel,'saltfree')
 		call to_history("n1", nz,'saltfree')
 		call to_history("o1", oz,'saltfree')
 		call to_history("d1", dz,'saltfree')
 		call to_history("n2", nx,'saltfree')
 		call to_history("o2", ox,'saltfree')
 		call to_history("d2", dx,'saltfree')
 		call to_history("label1", 'z','saltfree')
 		call to_history("label2", 'x','saltfree')
	else
		write(0,*) "You need to provide a saltfree output called 'saltfree'"
	end if


end program
