program RES_MUTE
    use sep
    use calc_mod
    use mem_mod
    implicit none

    integer                             :: nt, rec_nx, nshots, tt1i, tt2i, ix, is, radius
    real                                :: x, dx, ox, v1, v2, tt1, tt2, ot, dt, t1, t2
    real,allocatable,dimension(:,:)     :: weights, smoothweights
    real,allocatable,dimension(:,:,:)   :: residual
    real, dimension(:), allocatable     :: aa


    call sep_init()
    call from_history("n1",nt)
    call from_history("o1",ot)
    call from_history("d1",dt)
    call from_history("n2",rec_nx)
    call from_history("o2",ox)
    call from_history("d2",dx)
    call from_history("n3",nshots)

    call mem_alloc3d(residual, nt,rec_nx,nshots)
    call mem_alloc2d(weights, nt,rec_nx)
    call mem_alloc2d(smoothweights, nt,rec_nx)
    call mem_alloc1d(aa, nt*rec_nx*nshots)

    call sep_read(residual)


    ! Get input hyperbola parameters
    call from_param("v1",v1)
    call from_param("t1",t1)    
    call from_param("radius",radius)

    if (exist_file('v2' )) call from_param("v2",v2)
    if (exist_file('t2' )) call from_param("t2",t2)

    weights=0.0

    do ix=1,rec_nx
        x = ox + (ix-1)*dx
        tt1 = sqrt(t1**2 + (x**2)/(v1**2))

        ! Do bottom event curve if aske for
        if (exist_file('t2' )) then
            tt2 = sqrt(t2**2 + 4*(x**2)/(v2**2))
        else
            tt2=(nt-1)*dt+ot
        end if

        tt1i = (tt1 - ot)/dt + 1
        tt2i = (tt2 - ot)/dt + 1


! write(0,*) "============================================="
! write(0,*) "v1 = ", v1
! write(0,*) "x = ", x
! write(0,*) "t1 = ", t1

! write(0,*) "tt1 = ", tt1
! write(0,*) "tt1i = ", tt1i
! write(0,*) "tt2i = ", tt2i
! write(0,*) "nt = ", nt


        weights(tt1i:tt2i,ix) = 1.0
    end do



!----- Apply the smoothed weighting ----------------------
    call smoother(weights, smoothweights, radius)
    do is=1,nshots
        residual(:,:,is) = residual(:,:,is)*smoothweights
    end do
    call sep_write(residual)


   ! Output the weight mask of the mute
    if (exist_file('weights')) then
        call to_history("n1", nt,'weights')
        call to_history("o1", ot,'weights')
        call to_history("d1", dt,'weights')
        call to_history("n2", rec_nx,'weights')
        call to_history("o2", ox,'weights')
        call to_history("d2", dx,'weights')
        call to_history("n3", nshots,'weights')
        call to_history("label1", 'time','weights')
        call to_history("label2", 'rec_x','weights')
        call sep_write(smoothweights,'weights')
    end if 



    ! Exit this
    call sep_close()
    deallocate(residual,weights)
    call exit(0)

end program
