
program RTM_INVERSION
	use sep
	use mem_mod
	use fbi_mod
	use initialize2
	use cgstep_mod
	use solver_mod
	implicit none

	!===========================================================================
	!	Allocate and Initialize
	!---------------------------------------------------------------------------
	integer									:: isou,rec_dx,rec_ox,sou_dx,sou_ox,nz,nx,nt,bc1,bc2,bc3,bc4,wfld_j,sou_nx,rec_nx,rec_z,niter,nh
	real 									:: dz,dx,dt,oz,ox,ot,tstep,vstep
	logical									:: adjoint
	real, dimension(:,:,:), allocatable		:: data, image
	real, dimension(:,:), allocatable		:: vel_in, model
	real, dimension(:), allocatable         :: model1, data_shot


	!===========================================================================
	real, dimension(:,:,:,:), allocatable	:: insrc
	real, dimension(:,:,:), allocatable		:: insrc_in
	!===========================================================================

	write(0,*) "======================================================"
	write(0,*) "--------- Inside the BORN program"
	call init_rtm(nz,nx,nt,nh,dz,dx,dt,oz,ox,ot,bc1,bc2,bc3,bc4,wfld_j,sou_nx,sou_dx,sou_ox,rec_nx,rec_dx,rec_ox,rec_z,tstep,vstep,niter)
	write(0,*) "--------- Done with BORN initialization routine"

	call mem_alloc3d(data, nt, rec_nx, sou_nx)
	call mem_alloc1d(data_shot, nt*rec_nx)
	call mem_alloc3d(image, nz, nx,(2*nh+1))
	call mem_alloc2d(vel_in, nz, nx)
	call mem_alloc2d(model, nz, nx)
	call mem_alloc1d(model1, nz*nx)

	call sep_init()
	call sep_read(vel_in,"velmod")
	call model_update(vel_in)
	call from_param("isou", isou)
	! call from_param("adjoint", adjoint)


!===========================================================================
	if (exist_file('insrc' )) then
		write(0,*) "READING IN THE SAVED SOURCE WAVEFIELD"
		call mem_alloc3d(insrc_in, nz, nx, nt)
		call mem_alloc4d(insrc, nz, nx, nt, 1)
		call sep_read(insrc_in, 'insrc' )
		insrc=0.0
		insrc(:,:,:,1)=insrc_in
		call init_srcwave_fbi(insrc)
		write(0,*) "FINISHED READING IN THE SAVED SOURCE WAVEFIELD"
	end if
!===========================================================================



	call to_history("n1",nz, 'model')
	call to_history("d1",dz, 'model')
	call to_history("o1",oz, 'model')
	call to_history("n2",nx, 'model')
	call to_history("n3",(2*nh+1), 'model')
	call to_history("d2",dx, 'model')
	call to_history("o2",ox, 'model')
	call to_history("label1",'z [km]', 'model')
	call to_history("label2",'x [km]', 'model')




	model1=0.0
	call sep_read(data,"data")
	data_shot = RESHAPE(data(:,:,isou), (/ nt*rec_nx /))
	write(0,*) "sum(data_shot) =", sum(data_shot)
	call solver_tiny( model1, data_shot, L_lop2_func, cgstep, 50)
	model = RESHAPE(model1, (/ nz, nx /))
	write(0,*) "sum(model) =", sum(model)
	call sep_write(model,'model')




! 	!---------------------------------------------------------------------------
! 	! Linear operator: full nonlinear data => model perturbation
! 	if (adjoint) then
! 		write(0,*) " Adjoint: data perturbation => model perturbation"
! 		source=0.0
! 		call sep_read(data,"data")
! 		data_shot = data(:,:,isou)
! write(0,*) "sum(data_shot) =", sum(data_shot)
! 		call L_source(.true., .false., source, data_shot, isou)
! write(0,*) "sum(source) =", sum(source)
! 		call sep_write(source,'source')
! 	!---------------------------------------------------------------------------
! 	! Linear operator: model perturbation => full nonlinear data 
! 	else
! 		write(0,*) " Forward: model perturbation => full nonlinear data "
! 		data=0.0
! 		data_shot=0.0
! 		call sep_read(source,'source')
! write(0,*) "sum(source) =", sum(source)
! 		call L_source(.false., .false., source, data_shot, isou)
! 		call sep_write(data_shot,'data')
! write(0,*) "sum(data_shot) =", sum(data_shot)
! 	end if





	call sep_close()

end program



