program SERIAL_1
	use sep
	use mem_mod
	use fbi_mod
	use calc_mod

	use drlse_mod
	use initialize1
	implicit none

	real                               	:: Vsalt, mu, objfuncval, sou_oz, sou_ox, sou_dx, oz, ox, dz, dx, js
	integer								:: nz, nx, nt, rec_nx, sou_nx
	real, dimension(:,:), allocatable	:: phi1, phi2, junk2, scaling, drlse_term1, drlse_term2, s1, s2, vel_back
	real, dimension(:,:,:), allocatable	:: residual, residual2
	logical								:: singlewell

	call sep_init()
	call from_param("q1", Vsalt)
	call from_param("n1", nz)
	call from_param("n2", nx)
	call from_param("nt", nt)
	call from_param("nr", rec_nx)
	call from_param("ns", sou_nx)
	call from_param("fs", sou_ox)
	call from_param("js", sou_dx)
	call from_param("fs_z", sou_oz)
	call from_param("o1", oz)
	call from_param("o2", ox)
	call from_param("d1", dz)
	call from_param("d2", dx)
	call from_param("singlewell", singlewell)


!------------ Initialize ----------------------------------
	call serial1_calc_init(nz,nx,vsalt,sou_nx, sou_oz, sou_ox, sou_dx, oz, ox, dz, dx)


!---------------- Allocate -------------------------------
	call mem_alloc2d(phi1, nz, nx)
	call mem_alloc2d(junk2, nz, nx)
	call mem_alloc2d(scaling, nz, nx)
	call mem_alloc2d(drlse_term1, nz, nx)
	call mem_alloc2d(s1, nz, nx)
	call mem_alloc2d(vel_back, nz, nx)



!------------ Read in values ------------------------------
	call sep_read(phi1, "phi1")
	call sep_read(vel_back, "vel_back")


	if (exist_file('phi2'))	then
		call mem_alloc2d(phi2, nz, nx)
		call mem_alloc2d(drlse_term2, nz, nx)
		call mem_alloc2d(s2, nz, nx)
		call sep_read(phi2,'phi2')
!-----------  Calculate DRLSE term 2 ------------------------------------------------
		call drlse_init(nz, nx, 1.0, 1.0)
		call calc_drlse_term( phi2, drlse_term2, s2 ,singlewell)
		drlse_term2 = (mu*drlse_term2)
	end if



!-----------  Create the velocity scaling coefficients -----------------------------
	junk2   = Vsalt
	scaling = junk2 - vel_back  ! overridden


!-----------  Calculate DRLSE term 1 ------------------------------------------------
	!mu < (h^2)/(4*gamma)
	! mu =  0.9*(1.0**2.0)/4.0
	mu =  0.95*(1.0**2.0)/4.0
	call drlse_init(nz, nx, 1.0, 1.0)
	call calc_drlse_term( phi1, drlse_term1, s1, singlewell )
	drlse_term1 = (mu*drlse_term1)


!-----------  Write outputs ------------------------------------------------
	call outfiles_serial1()
	if (exist_file('scaling'    ))	call sep_write(scaling,  	'scaling'		)
	if (exist_file('s1'			)) 	call sep_write(s1,  		's1'			)
	if (exist_file('s2'			)) 	call sep_write(s2,  		's2'			)
	if (exist_file('drlse_term1')) 	call sep_write(drlse_term1,	'drlse_term1'	)
	if (exist_file('drlse_term2')) 	call sep_write(drlse_term2,	'drlse_term2'	)


call sep_close()
write(0,*)   " reached the end of the serial1 program  ============================"


end program