program TAPER_WAVELET

 use sep
 implicit none

 real, dimension(:), allocatable				:: data_in, taper
 real								:: dt, ot, tdelay, val, sigma, taper_offset, fund
 integer							:: it, nt, t1, t2, t3, t4

 call sep_init()
 call from_history("n1",nt)
 call from_history("d1",dt)
 call from_history("o1",ot)
 call from_param("tdelay",tdelay)
 call from_param("sigma",sigma)
 call from_param("fund",fund)

 ! Where to begin taper relative to center of wavelet 
! This is a rough scaling approximation based on observation only
 taper_offset=(0.1*15.0/real(fund)) 


 t1 = (tdelay-taper_offset-sigma)/dt + 0.5 ! 0.5 for rounding
 t2 = (tdelay-taper_offset)/dt + 0.5
 t3 = (tdelay+taper_offset)/dt + 0.5 
 t4 = (tdelay+taper_offset+sigma)/dt + 0.5

 allocate(data_in(nt))
 allocate(taper(nt))
 taper = 1.0
 call sep_read(data_in)

write(0,*) "t1 = ", t1*dt
write(0,*) "t2 = ", t2*dt
write(0,*) "t3 = ", t3*dt
write(0,*) "t4 = ", t4*dt

 do it=1, nt
	! Before the taper make sure wavelet is = 0.0
	if (it<t1) then
		taper(it) = 0.0
		data_in(it) = 0.0
	end if
	! Taper the left side of the wavelet
	if (it>=t1 .and. it<t2) then
		val = cos( ((sigma/dt)-it+t1)*3.14159265358/(2.0*sigma/dt) )**2
		taper(it) = val
		data_in(it) = data_in(it)*val
	end if
	! Taper the right side of the wavelet
	if (it>t3 .and. it<=t4) then
		val = cos( (it-t3)*3.14159265358/(2.0*sigma/dt) )**2
		taper(it) = val
		data_in(it) = data_in(it)*val
	end if
	! After the taper make sure wavelet is = 0.0
	if (it>t4) then
		taper(it) = 0.0
		data_in(it) = 0.0
	end if
 end do	
	
 call sep_write(data_in)
 if (exist_file('taper' )) then
	 call to_history("n1",nt,"taper")
	 call to_history("d1",dt,"taper")
	 call to_history("o1",ot,"taper")
	 call sep_write(taper, 'taper' )
 end if

 call sep_close()

end program
