
program WAVE_MODELING
	use sep
	use mem_mod
	use fbi_mod
	use initialize2
	implicit none

	!===========================================================================
	!	Allocate and Initialize
	!---------------------------------------------------------------------------
	integer									:: isou,rec_dx,rec_ox,sou_dx,sou_ox,nz,nx,nt,sou_nx,rec_nx,rec_z,nh, sou_z
	real 									:: dz,dx,dt,oz,ox,ot
	logical									:: adjoint
	real, dimension(:,:,:), allocatable		:: data
	real, dimension(:,:), allocatable		:: vel_in, data_shot
	real, dimension(:), allocatable			:: source

	!===========================================================================
	real, dimension(:,:,:,:), allocatable	:: insrc
	real, dimension(:,:,:), allocatable		:: insrc_in
	!===========================================================================

	write(0,*) "======================================================"
	write(0,*) "--------- Inside the WAVE MODELING program"
	call init_wavemodeling(nz,nx,nt,nh,dz,dx,dt,oz,ox,ot,sou_nx,sou_dx,sou_ox,rec_nx,rec_dx,rec_ox,rec_z, sou_z)
	write(0,*) "--------- Done with WAVE MODELING initialization routine"

	call mem_alloc3d(data, nt, rec_nx, sou_nx)
	call mem_alloc2d(data_shot, nt, rec_nx)
	call mem_alloc1d(source, nt)
	call mem_alloc2d(vel_in, nz, nx)

	call sep_init()
	call sep_read(vel_in,"velmod")
	write(0,*) "PROGRAM:  sum(vel_in) = ", sum(vel_in)
	call model_update(vel_in)
	call from_param("isou", isou)
	call from_param("adjoint", adjoint)

! !===========================================================================
! 	if (exist_file('insrc' )) then
! 		call mem_alloc3d(insrc_in, nz, nx, nt)
! 		call mem_alloc4d(insrc, nz, nx, nt, 1)
! 		call sep_read(insrc_in, 'insrc' )
! 		insrc=0.0
! 		insrc(:,:,:,1)=insrc_in
! 		call init_srcwave_fbi(insrc)
! 	end if
! !===========================================================================

	if (adjoint) then
	write(0,*) " ADJOINT"
   		call to_history("n1",nt, 'source')
		call to_history("d1",dt, 'source')
		call to_history("o1",ot, 'source')
		! call to_history("n2",sou_nx, 'source')
		call to_history("label1",'Time [s]', 'source')
		! call to_history("label2",'Source k', 'source')
	else
		write(0,*) "FORWARD"
   		call to_history("n1",nt, 'data')
		call to_history("d1",dt, 'data')
		call to_history("o1",ot, 'data')
   		call to_history("n2",rec_nx, 'data')
		call to_history("d2",rec_dx, 'data')
		call to_history("o2",rec_ox, 'data')
		! call to_history("n3",sou_nx, 'data')
		call to_history("label1",'Time [s]', 'data')
		call to_history("label2",'Receiver k', 'data')
		! call to_history("label3",'Source k', 'data')
	end if


	!---------------------------------------------------------------------------
	! Linear operator: full nonlinear data => model perturbation
	if (adjoint) then
		write(0,*) " Adjoint: data perturbation => model perturbation"
		source=0.0
		call sep_read(data,"data")
		data_shot = data(:,:,isou)
write(0,*) "sum(data_shot) =", sum(data_shot)
		call L_source(.true., .false., source, data_shot, isou)
write(0,*) "sum(source) =", sum(source)
		call sep_write(source,'source')
	!---------------------------------------------------------------------------
	! Linear operator: model perturbation => full nonlinear data 
	else
		write(0,*) " Forward: model perturbation => full nonlinear data "
		data=0.0
		data_shot=0.0
		call sep_read(source,'source')
write(0,*) "sum(source) =", sum(source)
		call L_source(.false., .false., source, data_shot, isou)
		call sep_write(data_shot,'data')
write(0,*) "sum(data_shot) =", sum(data_shot)
	end if



	! call to_history("n1",nz, 'model')
	! call to_history("d1",dz, 'model')
	! call to_history("o1",oz, 'model')
	! call to_history("n2",nx, 'model')
	! call to_history("d2",dx, 'model')
	! call to_history("o2",ox, 'model')
	! call to_history("label1",'z [km]', 'model')
	! call to_history("label2",'x [km]', 'model')
	! call Illumination(source)
	! call sep_write(source,'model')

	call sep_close()

end program



