
program WEMVA_test
	use sep
	use mem_mod
	use wemva_mod
	use initialize_wemva_hess
	implicit none

	!===========================================================================
	!	Allocate and Initialize
	!---------------------------------------------------------------------------
	integer									:: isou,rec_dx,rec_ox,sou_dx,sou_ox,nz,nx,nt,bc1,bc2,bc3,bc4,wfld_j,sou_nx,rec_nx,rec_z,niter,nh,np
	real 									:: dz,dx,dt,oz,ox,ot,tstep,vstep
	logical									:: adjoint
	real, dimension(:,:,:), allocatable	    :: insrc, rec_wave
	! real, dimension(:,:,:), allocatable		:: scat_src_wave
	real, dimension(:,:), allocatable		:: vel_in, image_in, output, wemva1out, wemva2out, coor_rec, coor_src
	!===========================================================================
	write(0,*) "======================================================"
	write(0,*) "Inside the WEMVA_test program"

	call init_wemva_hessian(nz,nx,nt,nh,dz,dx,dt,oz,ox,ot,bc1,bc2,bc3,bc4,wfld_j,sou_nx,sou_dx,sou_ox,rec_nx,rec_dx,rec_ox,rec_z,tstep,vstep,niter,np)

	call mem_alloc2d(wemva1out, nz, nx)
	call mem_alloc2d(wemva2out, nz, nx)
	call mem_alloc2d(image_in, nz, nx)
	call mem_alloc2d(output, nz, nx)
	call mem_alloc2d(vel_in, nz, nx)

	call sep_init()
	call sep_read(vel_in,"velmod")
	call model_updateWH(vel_in)
	call sep_read(image_in,"modpert")

!===========================================================================
	call mem_alloc3d(insrc, np, nt, 1)
	call mem_alloc3d(rec_wave, np, nt, 1)

	call mem_alloc2d(coor_src, 2, np)
	call mem_alloc2d(coor_rec, 2, np)


write(0,*) "start reading"
	call sep_read(insrc,"src_wave")
	call sep_read(rec_wave,"rec_wave")

	call sep_read(coor_src,"coor_src")
	call sep_read(coor_rec,"coor_rec")

	call init_saved_wavefields(insrc,insrc)
	call init_wave_coor(coor_src, coor_rec, np)

write(0,*) "wavefields initialized"
!===========================================================================

	call from_param("isou", isou)
	call from_param("adjoint", adjoint)

	if (adjoint) then
	write(0,*) " ADJOINT"
		call to_history("n1",nz, 'model')
		call to_history("d1",dz, 'model')
		call to_history("o1",oz, 'model')
		call to_history("n2",nx, 'model')
		call to_history("n3",(2*nh+1), 'model')
		call to_history("d2",dx, 'model')
		call to_history("o2",ox, 'model')
		call to_history("label1",'z [km]', 'model')
		call to_history("label2",'x [km]', 'model')
	else
		! write(0,*) "FORWARD"
  !  		call to_history("n1",nt, 'data')
		! call to_history("d1",dt, 'data')
		! call to_history("o1",ot, 'data')
  !  		call to_history("n2",rec_nx, 'data')
		! call to_history("d2",rec_dx, 'data')
		! call to_history("o2",rec_ox, 'data')
		! call to_history("label1",'Time [s]', 'data')
		! call to_history("label2",'Receiver k', 'data')
	end if


	!---------------------------------------------------------------------------
	! Linear operator: full nonlinear data => model perturbation
	output = 0.0
	if (adjoint) then
		write(0,*) " ---------------- WEMVA 1 --------------"
		call wemva1(.true., .false., image_in, wemva1out, rec_wave, isou)
		output = output + wemva1out
		write(0,*) " ---------------- WEMVA 2 --------------"
		call wemva2(.true., .false., image_in, wemva2out, rec_wave, isou)
		output = output + wemva2out
		call sep_write(output,'model')
		call sep_write(wemva1out,'wemva1')
		call sep_write(wemva2out,'wemva2')
	!---------------------------------------------------------------------------
	! Linear operator: model perturbation => full nonlinear data 
	else
		! write(0,*) " ---------------- WEMVA 2 --------------"
		! call wemva2(.false., .false., image_in, wemva1out, rec_wave, isou)
		! output = output + wemva1out
		! write(0,*) " ---------------- WEMVA 1 --------------"
		! call wemva1(.false., .false., image_in, wemva2out, rec_wave, isou)
		! output = output + wemva2out
	end if


	call sep_close()

end program



