module calc_mod
	use fbi_mod
	use secder_mod
	use drlse_mod
	implicit none
	integer, private								:: nz, nx, nt, rec_nx, sou_nx, isou
	real, dimension(:,:,:), allocatable, private	:: rec_data, syndata
	real, dimension(:,:), pointer, private 			:: phi_grad, vel_full, phi2

	real, dimension(:,:), allocatable, private		:: residual, temp2d, temp2dinv, scaling_temp, velfull_sm, distmap
	real, dimension(:,:), allocatable, private  	:: phi_a, phi_x, phi_y, phi_in, laplac_mat, curvature, Nyy, Nxx, Vgx, Vgy, Vg, Vgx2, Vgy2, Vg2, tmap, full_vel2
	real, private									:: small_number, tdelay, vc, vsalt, sou_ox, sou_dx, rec_ox, rec_dx, dt, dx, dz, pi, oz, ox, sou_oz
	contains




!/////////////////////////////////////////////////////////////////////////////////////
	subroutine calc_init( nz_in, nx_in, nt_in, rec_nx_in, sou_nx_in, tdelay_in, vc_in, vsalt_in, sou_ox_in, sou_dx_in, rec_ox_in, rec_dx_in, dt_in, dx_in, dz_in, oz_in, ox_in, sou_oz_in)
		integer		:: nz_in, nx_in, nt_in, rec_nx_in, sou_nx_in
		real		:: tdelay_in, vc_in, vsalt_in, sou_ox_in, sou_dx_in, rec_ox_in, rec_dx_in, dt_in, dx_in, dz_in, oz_in, ox_in, sou_oz_in
		nz = nz_in
		nx = nx_in
		nt = nt_in
		rec_nx = rec_nx_in
		sou_nx = sou_nx_in

		tdelay = tdelay_in
		vc = vc_in
		vsalt = vsalt_in
		sou_ox = sou_ox_in
		sou_oz = sou_oz_in
		sou_dx = sou_dx_in
		rec_ox = rec_ox_in
		rec_dx = rec_dx_in
		dt = dt_in
		dx = dx_in
		dz = dz_in

		oz = oz_in
		ox = ox_in

		small_number = 0.00000000000001
		pi = 3.14159265359

!----------------- allocate --------------------------
		call mem_alloc3d(rec_data,  nt, rec_nx, sou_nx)
		call mem_alloc2d(residual,  nt, rec_nx)
		call mem_alloc3d(syndata,  nt, rec_nx, sou_nx)

		call mem_alloc2d(phi_a, nz, nx)
		call mem_alloc2d(temp2d, nz, nx)
		call mem_alloc2d(temp2dinv, nz, nx)
		call mem_alloc2d(phi_x, nz, nx)
		call mem_alloc2d(phi_y, nz, nx)
		call mem_alloc2d(phi_in, nz, nx)
		call mem_alloc2d(curvature, nz, nx)
		call mem_alloc2d(laplac_mat, nz, nx)
		call mem_alloc2d(Nxx, nz, nx)
		call mem_alloc2d(Nyy, nz, nx)
		call mem_alloc2d(Vgx, nz, nx)
		call mem_alloc2d(Vgy, nz, nx)
		call mem_alloc2d(Vg, nz, nx)

		call mem_alloc2d(Vgx2, nz, nx)
		call mem_alloc2d(Vgy2, nz, nx)
		call mem_alloc2d(Vg2, nz, nx)
		call mem_alloc2d(tmap, nz, nx)
		call mem_alloc2d(full_vel2, nz, nx)
		call mem_alloc2d(scaling_temp, nz, nx)
		call mem_alloc2d(velfull_sm, nz, nx)

		call mem_alloc2d(distmap, nz, nx)

!----------------- read in rec_data ------------------
		!call sep_read(rec_data, "rec_data")
	end subroutine







!/////////////////////////////////////////////////////////////////////////////////////
	subroutine serial1_calc_init( nz_in, nx_in, vsalt_in, sou_nx_in, sou_oz_in, sou_ox_in, sou_dx_in, oz_in, ox_in, dz_in, dx_in)
		integer		:: nz_in, nx_in, sou_nx_in
		real 		:: vsalt_in, sou_oz_in, sou_ox_in, sou_dx_in, oz_in, ox_in, dz_in, dx_in

		nz = nz_in
		nx = nx_in
		vsalt = vsalt_in

		sou_oz = sou_oz_in
		sou_ox = sou_ox_in
		sou_dx = sou_dx_in
		sou_nx = sou_nx_in

		oz = oz_in
		ox = ox_in
		dz = dz_in
		dx = dx_in

		call mem_alloc2d(full_vel2, nz, nx)
		call mem_alloc2d(velfull_sm, nz, nx)
		call mem_alloc2d(Vgx2, nz, nx)
		call mem_alloc2d(Vgy2, nz, nx)
		call mem_alloc2d(Vgx, nz, nx)
		call mem_alloc2d(Vgy, nz, nx)
		call mem_alloc2d(distmap, nz, nx)
		call mem_alloc2d(temp2d, nz, nx)

	end subroutine




!/////////////////////////////////////////////////////////////////////////////////////
	subroutine serial2_calc_init( nz_in, nx_in, vel_full_in, phi2_in)
		integer							:: nz_in, nx_in
		real, dimension(:,:), target	:: vel_full_in, phi2_in

		nz = nz_in
		nx = nx_in
		vel_full => vel_full_in
		phi2=> phi2_in

	end subroutine






!/////////////////////////////////////////////////////////////////////////////////////
	subroutine calc_vel_init( nz_in, nx_in)
		integer		:: nz_in, nx_in

		nz = nz_in
		nx = nx_in
		call mem_alloc2d(full_vel2, nz, nx)

	end subroutine






!/////////////////////////////////////////////////////////////////////////////////////
	subroutine smoother(model, smoothmodel, radius)
		real, dimension(:,:), target	:: model, smoothmodel
		integer							:: radius, left, right, top, bottom, countr, i, j, k, l
		real							:: val

		if (radius==0) then 
			smoothmodel =  model
		else
			do j=1, size(model,2)        ! x axis
				do i=1, size(model,1)    ! z axis

					top    = i - radius
					bottom = i + radius
					left   = j - radius
					right  = j + radius

					if (top    < 1) top  = 1
					if (left   < 1) left = 1

					if (bottom > size(model,1)) bottom = size(model,1)
					if (right  > size(model,2)) right  = size(model,2)

					countr = 0
					val    = 0.0
					do k=top, bottom
						do l=left, right
							countr = countr + 1
							val = val + model(k,l)
						end do
					end do
					smoothmodel(i,j) = real(val)/real(countr)
	 			end do
			end do
		end if
	end subroutine







!/////////////////////////////////////////////////////////////////////////////////////
	subroutine calc_velocity(phi_in1, phi_in2, vel_back, full_vel, vs)
		real, dimension(:,:), target	:: phi_in1, phi_in2, full_vel, vel_back
		integer							:: iz, ix, stencil, ip, np, it
		real							:: vs

		! Build the smoothing field
		full_vel = vel_back
		do iz=1,nz
			do ix=1,nx
				if (phi_in2(iz,ix)>=0.0) then
					full_vel(iz,ix) = 0.0
				end if
				if (phi_in1(iz,ix)>=0.0) then
					full_vel(iz,ix) = vs
				end if
			end do
		end do

		! Smoothing the field
		stencil = 1
		np=350
		do ip=1,np
			full_vel2 = full_vel
			do iz=1,nz
				do ix=1,nx
					if ((phi_in2(iz,ix)>=0.0) .and. (phi_in1(iz,ix)<0.0)) then ! If in transition zone
						full_vel2(iz,ix) = (full_vel(iz+stencil,ix) + full_vel(iz-stencil,ix) + full_vel(iz,ix+stencil) + full_vel(iz,ix-stencil)) / 4.0
					end if
				end do
			end do
			full_vel = full_vel2
		end do 
	end subroutine









!/////////////////////////////////////////////////////////////////////////////////////
	subroutine phi_split_weights(velfull_in, weightsTOP, weightsBOT, s, phi_in)
		real, dimension(:,:), target	:: velfull_in, weightsBOT, weightsTOP, s, phi_in
		integer							:: iz, ix, isou
		real							:: soux, souz, zz, xx, dist, minw, maxx
		
!---------  NORMAL GRADIENT OF VELOCITY MAP ----------------
!		call smoother_nonzero(velfull_in, velfull_sm, 2)
		velfull_sm = 0.0

		where (phi_in>0.0)  velfull_sm = vsalt
		call gradient(velfull_sm, Vgx2, Vgy2)

		do ix=1,nx
			do iz=1,nz
				maxx = maxval( (/ abs(Vgx2(iz,ix)), abs(Vgy2(iz,ix)) /) )
				if (maxx>0.0) then
					Vgx2(iz,ix) = Vgx2(iz,ix)/maxx
					Vgy2(iz,ix) = Vgy2(iz,ix)/maxx
				else
					Vgx2(iz,ix) = 0.0
					Vgy2(iz,ix) = 0.0
				end if				
			end do
		end do

		weightsBOT = 0.0
		souz = sou_oz + oz

		do isou=1,sou_nx
			distmap = 0.0
			temp2d  = 0.0
			!---------  MAKE DISTANCE MAP ----------------
			soux = ((isou-1)*sou_dx + sou_ox)*dx + ox

			do ix=1,nx
				do iz=1,nz
					zz = ((iz-1)*dz + oz) - souz
					xx = ((ix-1)*dx + ox) - soux
					dist = sqrt(xx**2 + zz**2)
					distmap(iz,ix) = distmap(iz,ix) + dist
				end do
			end do

			!---------  GRADIENT OF DISTANCE MAP ---------------
			Vgx = 0.0
			Vgy = 0.0
			call gradient(distmap, Vgx, Vgy)

			do ix=1,nx				
				do iz=1,nz
					maxx = maxval( (/ abs(Vgx(iz,ix)), abs(Vgy(iz,ix)) /) )
					if (maxx>0.0) then
						Vgx(iz,ix) = Vgx(iz,ix)/maxx
						Vgy(iz,ix) = Vgy(iz,ix)/maxx
					else
						Vgx(iz,ix) = 0.0
						Vgy(iz,ix) = 0.0
					end if				
				end do
			end do
			!---------  Dot product ---------------
			temp2d = Vgx*Vgx2 + Vgy*Vgy2	
			weightsTOP = weightsTOP + temp2d

		end do	

		!---------  Masking of MAP ----------------
		where ( s<=0.0 ) weightsTOP = 0.0

!---------  Inverse THE WEIGHTING MAP ----------------
		weightsBOT = weightsTOP*-1.0

!---------  THRESHOLD THE WEIGHTING MAP ----------------
		where(weightsBOT<0.0)  weightsBOT=0.0
		where(weightsBOT>0.0)  weightsBOT=1.0

		where(weightsTOP<0.0)  weightsTOP=0.0
		where(weightsTOP>0.0)  weightsTOP=1.0

		weightsBOT = weightsBOT/maxval(abs(weightsBOT))
		weightsTOP = weightsTOP/maxval(abs(weightsTOP))

	end subroutine    








!/////////////////////////////////////////////////////////////////////////////////////
	subroutine edge_smoother(model, smoothmodel, radius)
		real, dimension(:,:), target	:: model, smoothmodel
		integer							:: radius, left, right, top, bottom, countr, i, j, k, l
		real							:: val

		do j=1, size(model,2)        ! x axis
			do i=1, size(model,1)    ! z axis

				if (model(i,j)==0.0) then

					top    = i - radius
					bottom = i + radius
					left   = j - radius
					right  = j + radius

					if (top    < 1) top  = 1
					if (left   < 1) left = 1
	
					if (bottom > size(model,1)) bottom = size(model,1)
					if (right  > size(model,2)) right  = size(model,2)

					countr = 0
					val    = 0.0
					do k=top, bottom
						do l=left, right
							countr = countr + 1
							val = val + model(k,l)
						end do
					end do
					smoothmodel(i,j) = real(val)/real(countr)
				else
					smoothmodel(i,j) = model(i,j)
				end if
 			end do
		end do
	end subroutine






! !/////////////////////////////////////////////////////////////////////////////////////
! 	subroutine salt_masker1(tomo_grad)
! 		real, dimension(:,:), target	:: tomo_grad
! 		integer							:: i, j	
! 		real							:: fade, weight
! 		real, dimension(nz,nx)			:: weight_arr
		
! 		weight = 0.0
! 		do j=1,nx
! 			do i=1,nz
! 				if (phi2(i,j) > 0.0 ) then
! 					weight = 1.0
! 				end if
! 				weight_arr(i,j) = weight
! 			end do
! 		end do
! 		tomo_grad = tomo_grad*weight_arr
! 		if (exist_file('weight2'   )) call sep_write(weight_arr, 'weight2' )
! 	end subroutine





!/////////////////////////////////////////////////////////////////////////////////////
	subroutine salt_masker(tomo_grad)
		real, dimension(:,:), target	:: tomo_grad
		integer							:: i, j	
		real							:: fade, weight
		real, dimension(nz,nx)			:: weight_arr
		
		fade = 0.1
		do j=1,nx
			do i=1,nz
				if (phi2(i,j) > fade ) then
					weight = 0.0
				else if (phi2(i,j) > 0.0 ) then
					weight = (( cos( phi2(i,j)*pi/(fade) ) + 1.0 ) / 2.0)
				else
					weight = 1.0
				end if
				weight_arr(i,j) = weight
			end do
		end do
		tomo_grad = tomo_grad*weight_arr
		! if (exist_file('weight2'   )) call sep_write(weight_arr, 'weight2' )
	end subroutine






!/////////////////////////////////////////////////////////////////////////////////////
	subroutine water_masker(tomo_grad, val)
		real, dimension(:,:), target	:: tomo_grad
		integer							:: i, j	
		real							:: fade, weight, val
		real, dimension(nz,nx)			:: weight_arr
		
		do i=1,nz
			do j=1,nx
				if (vel_full(i,j) < val ) then	! USe 1550 m/s water vel for safety
					weight = 0.0
				else
					weight = 1.0
				end if
				weight_arr(i,j) = weight
			end do
		end do
		tomo_grad = tomo_grad*weight_arr
		if (exist_file('weight2'   )) call sep_write(weight_arr, 'weight2' )
	end subroutine




! !/////////////////////////////////////////////////////////////////////////////////////
! 	subroutine scaling_calc(vel_full, phi1, phi2, scaling, inner)
! 		real, dimension(:,:), target	:: vel_full, phi1, phi2, scaling
! 		integer							:: iz, ix, npass, nstencil
! 		real							:: avg
! 		logical							:: inner

! 		call gradient(vel_full, Vgx, Vgy)
! 		Vg = sqrt( Vgx**2 + Vgy**2) ! Nabla term?
! 		do ix=1,5
! 			Vgx = 0.0
! 			call smoother_nonzero(Vg, Vgx, 1)
! 			Vg = Vgx
! 		end do

! write(0,*) " ABORT ABORT ABORT NOW MOTHER PH**KER "

! 		tmap      = 1.0
! 		temp2d    = 0.0
! 		temp2dinv = 1.0
! 		! Build the smoothing field
! 		do ix=1,nx
! 			do iz=1,nz
! 				if (inner) then ! if finding scaling for inner phi
! 					if (phi1(iz,ix) <0.0) then ! if outside salt
! 						tmap(iz,ix) = 0.0
! 					end if
! 					if (phi2(iz,ix)<0.0) then ! if inside transition zone or salt
! 						temp2dinv(iz,ix) = 0.0
! 					end if
! 				else ! if finding scaling for outer phi
! 					if (phi2(iz,ix)>0.0) then ! if inside transition zone or salt
! 						tmap(iz,ix) = 0.0
! 					end if
! 					if (phi1(iz,ix)>0.0) then ! if inside transition zone or salt
! 						temp2dinv(iz,ix) = 0.0
! 					end if
! 				end if
! 			end do
! 		end do

! 		! Grab edge values / complete edge mask
! 		do iz=1,nz
! 			do ix=1,nx
! 				if ( (tmap(iz,ix)==1.0) .and. ( (tmap(iz,ix+1)==0.0) .or. (tmap(iz,ix-1)==0.0) .or. (tmap(iz+1,ix)==0.0) .or. (tmap(iz-1,ix)==0.0) ) ) then ! if edge value
! 					temp2d(iz,ix)    = Vg(iz,ix)
! 					temp2dinv(iz,ix) = 0.0
! 				end if
! 			end do
! 		end do

! if (exist_file('test'  )) call sep_write(temp2d, 'test' )

! 		npass=150
! 		nstencil = 1
! 		call transition_smoother_new(phi1, phi2, temp2dinv, temp2d, npass, nstencil, scaling)

! write(0,*) " Check this out"
! 	end subroutine




! subroutine transition_smoother_new(phi1, phi2, zeros, initvals, npass, nstencil, scaling)
! 		real, dimension(:,:), target	:: phi1, phi2, initvals, zeros, scaling
! 		integer							:: i, j, k, isx, isz, sx, sz, npass, nstencil, count
! 		real							:: sum, avg
		
! 		scaling = initvals
! 		scaling_temp = 0.0
! 		! Smooth transition zone
! 		do k=1,npass
! 			do i=1,nx
! 				sx = i-nstencil
! 				if ((sx<1.0).or.(sx+2.0*nstencil>nx)) cycle
! 				do j=1,nz
! 					sz = j-nstencil
! 					if ((sz<1.0).or.(sz+2.0*nstencil>nz)) cycle
! 					count=0
! 					sum=0.0					
! 					do isx=0,2*nstencil
! 						do isz=0,2*nstencil
! 							sum = sum + scaling(sz+isz,sx+isx)
! 							count = count + 1	
! 						end do
! 					end do
! 					avg = sum/real(count)
! 					scaling_temp(j,i) = avg
! 				end do
! 			end do
! 		scaling = scaling_temp*zeros + initvals
! 		end do
! 	end subroutine				
					
				



! 	subroutine transition_smoother(phi1, phi2, tmap, temp2d, npass, nstencil)
! 		real, dimension(:,:), target	:: phi1, phi2, tmap, temp2d
! 		integer							:: i, j, k, isx, isz, sx, sz, npass, nstencil, count
! 		real							:: sum, avg
		
! 		! Smooth transition zone
! 		do k=1,npass
! 		do i=1,nx
! 			sx = i-nstencil
! 			if ((sx<1.0).or.(sx+2.0*nstencil>nx)) cycle
! 			do j=1,nz
! 				sz = j-nstencil
! 				if ((sz<1.0).or.(sz+2.0*nstencil>nz)) cycle
! 				if ( (phi1(j,i)<0.0) .and. (phi2(j,i)>0.0) ) then !if inside transition zone 
! 					! Run stencil for an averaged point
! 					count=0
! 					sum=0.0					
! 					do isx=0,2*nstencil
! 						do isz=0,2*nstencil
! 							if ( (tmap(sz+isz,sx+isx)==1.0) .and. (temp2d(sz+isz,sx+isx)>0.0) ) then !if inside transition zone
! 								sum = sum + temp2d(sz+isz,sx+isx)
! 								count = count + 1	
! 							else if (tmap(sz+isz,sx+isx)==0.0) then
! 								sum = sum + temp2d(sz+isz,sx+isx)
! 								count = count + 1
! 							end if
! 						end do
! 					end do
! 					avg = sum/real(count)
! 					temp2d(j,i) = avg
! 		!write(0,*) " sum", sum
! 		!write(0,*) " count", count
! 				end if
! 			end do
! 		end do
! 		end do
! 	end subroutine				
					
				
				










! !/////////////////////////////////////////////////////////////////////////////////////
! 	subroutine t2gain(model, data, adj, add)
! 		real, dimension(:,:,:), target	:: model, data
! 		logical				:: adj, add
! 		real 				:: time, ot
! 		integer				:: it, ir, is
		
! 		ot=0.0
! 		if (adj) then 
! 			if (.not. add) then
! 				model=0.0
! 				do is=1,sou_nx
! 					do ir=1,rec_nx
! 						do it=1,nt
! 							time = ot + (it-1)*dt
! 							model(it,ir,is) = data(it,ir,is)*time**2
! 						end do
! 					end do
! 				end do
! 			end if
! 		else 
! 			if (.not. add) then
! 				data=0.0
! 				do is=1,sou_nx
! 					do ir=1,rec_nx
! 						do it=1,nt
! 							time = ot + (it-1)*dt
! 							data(it,ir,is) = model(it,ir,is)*time**2
! 						end do
! 					end do
! 				end do
! 			end if
! 		end if

! 	end subroutine    









! !/////////////////////////////////////////////////////////////////////////////////////
! 	subroutine rescale_pars( input, maxalpha, maxbeta, ratio, line )
! 		real, dimension(:), target	:: input
! 		real				:: maxalpha, maxbeta, ratio
! 		logical				:: line


! 		!-------------- Extend the alpha / beta coordinates to make a line  -------------------------
! 		if (line) then
! 			if (sum(abs(input))>0.0) then
! 				if (abs(ratio)>1.0) then
! 					input(1) = maxalpha*sign(1.0,input(1))
! 					input(2) = abs((1/ratio))*maxalpha*sign(1.0,input(2))
! 				else 
! 					input(2) = maxbeta*sign(1.0,input(2))
! 					input(1) = abs(ratio)*maxbeta*sign(1.0,input(1))
! 				end if
! 				input(1) = (input(1)/maxalpha)*0.5
! 				input(1) = 0.5 - input(1)
! 				input(2) = (input(2)/maxbeta)*0.5
! 				input(2) = input(2) + 0.5
! 			else
! 				input(1) = 0.5
! 				input(2) = 0.5
! 			end if
! 		else
! 		!-------------- Rescale the alpha / beta coordinates to make a point -------------------------
! 			input(1) = (input(1)/maxalpha)*0.5
! 			input(1) = 0.5 - input(1)
! 			input(2) = (input(2)/maxbeta)*0.5
! 			input(2) = input(2) + 0.5
! 		end if

! 	end subroutine











! !/////////////////////////////////////////////////////////////////////////////////////
! 	subroutine calc_drlse_term( phi_in, drlse_term, s )
! 		real, dimension(:,:), target	:: s, phi_in, drlse_term

! !----------- Create nabla term (gradient of phi with respect to x and y) ------------
! 		call neumann_bound_cond(phi_in, phi_a)
! 		call gradient(phi_a, phi_x, phi_y)
! 		s = sqrt( phi_x**2 + phi_y**2) ! Nabla term?

! !-----------  Calculate DRLSE term  -------------------------------------------------
! 		Nxx = phi_x/(s + small_number)
! 		Nyy = phi_y/(s + small_number)

! !-----------  Single-well DRLSE  -------------------------------------------------
! 		!call div(Nxx,Nyy,curvature)
! 		!call del2(phi_a,laplac_mat)
! 		!drlse_term = (4*laplac_mat - curvature)

! !-----------  Double-well DRLSE  -------------------------------------------------
! 		call dist_reg_p2(phi_a, drlse_term)

! 	end subroutine






! !/////////////////////////////////////////////////////////////////////////////////////





! !/////////////////////////////////////////////////////////////////////////////////////
! 	subroutine calc_residual(vel_in, residual2)
! 		real, dimension(:,:,:), target		:: residual2
! 		real, dimension(:,:), target		:: vel_in
! 		residual2 = 0.0
! 		syndata = 0.0
! 		call model_update(vel_in)
! 		call L_nlop2(.false., syndata)
! 		residual2 = syndata - rec_data
! !write(0,*) " inside the calc_residual.mod " 
! 	end subroutine






! !/////////////////////////////////////////////////////////////////////////////////////
! 	subroutine phi_grad_operator(model, data, t1, grad, scaling, adj)
! 		real, dimension(:,:,:), target		:: data
! 		real, dimension(:,:), target		:: model, t1, grad, scaling
! 		logical, intent(in)			:: adj

! 		t1 = 0.0
! 		if (adj) then	! residual ==> phi_grad
! 			model  = 0.0
! 			call L_lop2(.true. , .false., t1, data)
! 			model = grad*scaling*t1
! 		else		! phi_grad ==> residual
! 			data = 0.0
! 			t1 = scaling*grad*model
! 			call L_lop2(.false., .false., t1, data)
! 		end if
! 	end subroutine




! !/////////////////////////////////////////////////////////////////////////////////////
! 	subroutine smoother_nonzero(model, smoothmodel, radius)
! 		real, dimension(:,:), target	:: model, smoothmodel
! 		integer							:: radius, left, right, top, bottom, countr, i, j, k, l
! 		real							:: val

! 		do j=1, size(model,2)        ! x axis
! 			do i=1, size(model,1)    ! z axis

! 				top    = i - radius
! 				bottom = i + radius
! 				left   = j - radius
! 				right  = j + radius

! 				if (top    < 1) top  = 1
! 				if (left   < 1) left = 1

! 				if (bottom > size(model,1)) bottom = size(model,1)
! 				if (right  > size(model,2)) right  = size(model,2)

! 				countr = 0
! 				val    = 0.0
! 				do l=left, right
! 					do k=top, bottom					
! 						if (model(k,l)>0.0) then
! 							countr = countr + 1
! 							val = val + model(k,l)
! 						end if
! 					end do
! 				end do
! 				if (countr>0.0)then
! 					smoothmodel(i,j) = real(val)/real(countr)
! 				else
! 					smoothmodel(i,j) = 0.0
! 				end if
!  			end do
! 		end do
! 	end subroutine











! !/////////////////////////////////////////////////////////////////////////////////////
! 	subroutine dotproduct2(model, data, adj, add)
! 		real, dimension(:,:), target	:: model
! 		real				:: data
! 		logical				:: adj, add
! 		integer				:: iz, ix
	
! 		if (adj) then
! 			model = 0.0
! 			do iz=1, size(model,1)
! 				do ix=1, size(model,2)
! 					model(iz,ix) = model(iz,ix) + data*data
! 				end do
! 			end do
! 		else
! 			data = 0.0
! 			do iz=1, size(model,1)
! 				do ix=1, size(model,2)
! 					data = data + model(iz,ix)*model(iz,ix)
! 				end do
! 			end do
! 		end if
! 	end subroutine



! !/////////////////////////////////////////////////////////////////////////////////////
! 	subroutine dotproduct3(model, data, adj, add)
! 		real, dimension(:,:,:), target	:: model
! 		real				:: data
! 		logical				:: adj, add
! 		integer				:: iz, ix, it
	
! 		if (adj) then
! 			model = 0.0
! 			do iz=1, size(model,1)
! 				do ix=1, size(model,2)
! 					do it=1,size(model,3)
! 						model(iz,ix,it) = model(iz,ix,it) + data*data
! 					end do
! 				end do
! 			end do
! 		else
! 			data = 0.0
! 			do iz=1, size(model,1)
! 				do ix=1, size(model,2)
! 					do it=1,size(model,3)
! 						data = data + model(iz,ix,it)*model(iz,ix,it)
! 					end do
! 				end do
! 			end do
! 		end if
! 	end subroutine










! !/////////////////////////////////////////////////////////////////////////////////////
! 	subroutine	direct_arrival_mute(data, muted_data, mrange)
! 		real, dimension(:,:,:), target		:: data, muted_data
! 		real					:: mrange, sx, rx, dxx, tcent, ot, ww, pi, vc2, tdelay2
! 		integer					:: is, ir, it, itt, ntt

! 		real, dimension(:,:,:), allocatable	:: weight
! 		allocate(weight(nt, rec_nx, sou_nx))
! 		weight = 1.0

! 		ntt = mrange/dt
! 		pi = 3.14159265359
! 		muted_data = data

! 		! Tweak the mute slope to get avoid clipping the first reflection
! 		vc2 = vc*1.05
! 		tdelay2 = tdelay*1.35

! 		do is=1,sou_nx
! 			sx = (is-1)*sou_dx + sou_ox
! 			do ir=1,rec_nx
! 				rx = (ir-1)*rec_dx + rec_ox
! 				dxx = abs(sx-rx)
! 				tcent = dxx/vc2 + tdelay2 + (mrange/2.0)
! 				itt = tcent/dt + 1
! 				do it=1,nt
! 					if (it <= itt) then
! 						ww = 0.0
! 					else if ( (it > itt) .and. (it <= itt+ntt) ) then
! 						ww = 1 - (  abs(   cos( (0.5*pi/(ntt-1))*(it-itt-1) )  )  )**0.25
! 					else if (it > itt+ntt) then
! 						ww = 1.0
! 					end if
! 					weight(it,ir,is) = ww
! 				end do
! 			end do
! 		end do	
! 		muted_data = data*weight
! 		if (exist_file('weight'  )) call sep_write(weight,  'weight' )
		
! 	end subroutine








! !/////////////////////////////////////////////////////////////////////////////////////
! !***********************************************************************************************************************************
! !  M33INV  -  Compute the inverse of a 3x3 matrix.
! !
! !  A       = input 3x3 matrix to be inverted
! !  AINV    = output 3x3 inverse of matrix A
! !  OK_FLAG = (output) .TRUE. if the input matrix could be inverted, and .FALSE. if the input matrix is singular.
! !***********************************************************************************************************************************

!       subroutine M33INV (A, AINV, OK_FLAG)

!       !IMPLICIT NONE

!       DOUBLE PRECISION, DIMENSION(3,3), INTENT(IN)  :: A
!       DOUBLE PRECISION, DIMENSION(3,3), INTENT(OUT) :: AINV
!       LOGICAL, INTENT(OUT) :: OK_FLAG

!       DOUBLE PRECISION, PARAMETER :: EPS = 1.0D-10
!       DOUBLE PRECISION :: DET
!       DOUBLE PRECISION, DIMENSION(3,3) :: COFACTOR


!       DET =   A(1,1)*A(2,2)*A(3,3)  &
!             - A(1,1)*A(2,3)*A(3,2)  &
!             - A(1,2)*A(2,1)*A(3,3)  &
!             + A(1,2)*A(2,3)*A(3,1)  &
!             + A(1,3)*A(2,1)*A(3,2)  &
!             - A(1,3)*A(2,2)*A(3,1)

!       IF (ABS(DET) .LE. EPS) THEN
!          AINV = 0.0D0
!          OK_FLAG = .FALSE.
!          !RETURN
!       END IF

!       COFACTOR(1,1) = +(A(2,2)*A(3,3)-A(2,3)*A(3,2))
!       COFACTOR(1,2) = -(A(2,1)*A(3,3)-A(2,3)*A(3,1))
!       COFACTOR(1,3) = +(A(2,1)*A(3,2)-A(2,2)*A(3,1))
!       COFACTOR(2,1) = -(A(1,2)*A(3,3)-A(1,3)*A(3,2))
!       COFACTOR(2,2) = +(A(1,1)*A(3,3)-A(1,3)*A(3,1))
!       COFACTOR(2,3) = -(A(1,1)*A(3,2)-A(1,2)*A(3,1))
!       COFACTOR(3,1) = +(A(1,2)*A(2,3)-A(1,3)*A(2,2))
!       COFACTOR(3,2) = -(A(1,1)*A(2,3)-A(1,3)*A(2,1))
!       COFACTOR(3,3) = +(A(1,1)*A(2,2)-A(1,2)*A(2,1))

!       AINV = TRANSPOSE(COFACTOR) / DET

!       OK_FLAG = .TRUE.

!       !RETURN

!       end subroutine M33INV









! !/////////////////////////////////////////////////////////////////////////////////////
! 	subroutine velmodel_clip( velmodel, minvel, maxvel )
! 		real				:: minvel, maxvel
! 		real, dimension(:,:), target	:: velmodel
! 		integer				:: i, j	

! 		do j=1,nx
! 			do i=1,nz
! 				if (velmodel(i,j) > maxvel) then
! 					velmodel(i,j)=maxvel
! 				else if (velmodel(i,j) < minvel) then
! 					velmodel(i,j)=minvel
! 				end if
! 			end do
! 		end do
! 	end subroutine








end module
