



module comb_mod

  use laplace_mod
  use bcder_mod

  implicit none

  integer, private        :: n1, n2,nth
  real, dimension(:,:), pointer, private  :: veldt2, vwt, wt
  real, dimension(:,:,:), allocatable, private  :: temp

  contains



  subroutine comb_init(n1_in, n2_in, bc1_in, bc2_in, bc3_in, bc4_in, d1_in, d2_in, veldt2_in, vwt_in, wt_in)
    integer                      :: n1_in, n2_in, bc1_in, bc2_in, bc3_in, bc4_in
    real                         :: d1_in, d2_in
    real, dimension(:,:), target :: veldt2_in, vwt_in, wt_in
    integer, external :: omp_get_num_threads
    n1 = n1_in
    n2 = n2_in
    veldt2 => veldt2_in
    vwt => vwt_in
    wt => wt_in
    !$OMP PARALLEL
    nth=omp_get_num_threads()
    !$OMP END PARALLEL
    if(.not. allocated(temp)) allocate(temp(n1,n2,nth))
    call laplace_init(n1, n2, d1_in, d2_in,nth)
    call bcder_init(n1_in, n2_in, bc1_in, bc2_in, bc3_in, bc4_in)

  end subroutine




  function comb_op(adj, add, model, data) result(stat)
    logical,intent(in) :: adj, add
    real,dimension(:)  :: model, data
    integer            :: stat
    call comb_op2(adj, add, model, data,1)
    stat=0
  end function




  subroutine comb_op2(adj, add, model, data,ith) ! This operator is the C_5 operator (equation 19)
    logical,intent(in)       :: adj, add 
    real, dimension(n1,n2) :: model, data
    integer :: ith
    if(adj) then
      if(.not. add) model = 0.
      temp(:,:,ith) = veldt2*(1.-wt)*data
      call laplace_op2(adj, .true., model, temp(:,:,ith),ith)
      model = model + (2.-wt)*data
      temp(:,:,ith) = -vwt*wt*data
      call bcder_op2(adj, .true., model, temp(:,:,ith))
    else
      if(.not. add) data = 0.
      call laplace_op2(adj, .false., model, temp(:,:,ith),ith)
      data = data + (1.-wt)*(veldt2*temp(:,:,ith) + 2.*model)
      call bcder_op2(adj, .false., model, temp(:,:,ith))
      data = data + wt*(model - vwt*temp(:,:,ith))
    end if
  end subroutine

end module




!==================================================================================
!               DOUBLE PRECISION
!==================================================================================


! module comb_mod

!   use laplace_mod
!   use bcder_mod

!   implicit none

!   integer, private        :: n1, n2,nth
!   double precision, dimension(:,:), pointer, private  :: veldt2, vwt, wt
!   double precision, dimension(:,:,:), allocatable, private  :: temp

!   contains



!   subroutine comb_init(n1_in, n2_in, bc1_in, bc2_in, bc3_in, bc4_in, d1_in, d2_in, veldt2_in, vwt_in, wt_in)
!     integer                      :: n1_in, n2_in, bc1_in, bc2_in, bc3_in, bc4_in
!     real                         :: d1_in, d2_in
!     double precision, dimension(:,:), target :: veldt2_in, vwt_in, wt_in
!     integer, external :: omp_get_num_threads
!     n1 = n1_in
!     n2 = n2_in
!     veldt2 => veldt2_in
!     vwt => vwt_in
!     wt => wt_in
!     !$OMP PARALLEL
!     nth=omp_get_num_threads()
!     !$OMP END PARALLEL
!     if(.not. allocated(temp)) allocate(temp(n1,n2,nth))
!     call laplace_init(n1, n2, d1_in, d2_in,nth)
!     call bcder_init(n1_in, n2_in, bc1_in, bc2_in, bc3_in, bc4_in)

!   end subroutine




!   function comb_op(adj, add, model, data) result(stat)
!     logical,intent(in) :: adj, add
!     double precision,dimension(:)  :: model, data
!     integer            :: stat
!     call comb_op2(adj, add, model, data,1)
!     stat=0
!   end function




!   subroutine comb_op2(adj, add, model, data,ith) ! This operator is the C_5 operator (equation 19)
!     logical,intent(in)       :: adj, add 
!     double precision, dimension(n1,n2) :: model, data
!     integer :: ith
!     if(adj) then
!       if(.not. add) model = 0.
!       temp(:,:,ith) = veldt2*(1.-wt)*data
!       call laplace_op2(adj, .true., model, temp(:,:,ith),ith)
!       model = model + (2.-wt)*data
!       temp(:,:,ith) = -vwt*wt*data
!       call bcder_op2(adj, .true., model, temp(:,:,ith))
!     else
!       if(.not. add) data = 0.
!       call laplace_op2(adj, .false., model, temp(:,:,ith),ith)
!       data = data + (1.-wt)*(veldt2*temp(:,:,ith) + 2.*model)
!       call bcder_op2(adj, .false., model, temp(:,:,ith))
!       data = data + wt*(model - vwt*temp(:,:,ith))
!     end if
!   end subroutine

! end module
