


module fbi_mod

	use sep
	use mem_mod
	use tomo_mod
	use sscale_mod
	use lint_mod
	use step_mod
	use scats0_sso_mod
	use scatr0_mod
	use secder_mod
	use pad_mod

	implicit none
	integer, private                                     	:: nt, nz, nx, nt2, wfld_j, it, nth, nh, np
	integer, private                                     	:: sou_nx, sou_ox, sou_dx, sou_z, isou, sou_x
	integer, private                                     	:: rec_nx, rec_ox, rec_dx, rec_z, irec, rec_x, edge_par_shot, edge_par_rec
	integer, private                                     	:: bc1, bc2, bc3, bc4
	real, private                                        	:: dt, d1, o1, d2, o2, dtw
	real, dimension(:,:), allocatable, target, private   	:: veldt2, vwt, wt, modelbS !,vel, refl, !Not present in main: veldt2, vwt, wt. ! Not sure refl is necessary
	real,dimension(:,:,:,:),allocatable, target, private    :: temp1,temp2,temp1i,temp2i
	real, dimension(:,:,:), allocatable, target, private 	:: datab, wflds0, modelb
	real, dimension(:,:,:), pointer, private                :: insrc
	real, dimension(:,:), pointer, private                  :: coor
	real, dimension(:,:), allocatable, target, private   	:: datac !,vel, refl, !Not present in main: veldt2, vwt, wt. ! Not sure refl is necessary
	real, dimension(:), allocatable                         :: wavelet ! This is a copy of what is in main. Could use a pointer instead if you want.
	contains
!--------------------------------------------------------------------------------------
  


	subroutine init1(nz_in, o1_in, d1_in, nx_in, o2_in, d2_in, bc1_in, bc2_in, bc3_in, bc4_in, np_in) ! Copying variables 1
		integer     		:: nz_in, nx_in, bc1_in, bc2_in, bc3_in, bc4_in, np_in
		real        		:: d1_in, o1_in, d2_in, o2_in
		nz = nz_in
		nx = nx_in
		d1 = d1_in
		o1 = o1_in
		d2 = d2_in
		o2 = o2_in
		bc1 = bc1_in
		bc2 = bc2_in
		bc3 = bc3_in
		bc4 = bc4_in
		np = np_in
	end subroutine



	subroutine init2(sou_nx_in, sou_ox_in, sou_dx_in, sou_z_in, rec_nx_in, rec_ox_in, rec_dx_in, rec_z_in, edge_par_shot_in, edge_par_rec_in) ! Copying variables 2
		integer             :: sou_nx_in, sou_ox_in, sou_dx_in, sou_z_in
		integer             :: rec_nx_in, rec_ox_in, rec_dx_in, rec_z_in, edge_par_shot_in, edge_par_rec_in
		sou_nx = sou_nx_in
		sou_ox = sou_ox_in
		sou_dx = sou_dx_in
		sou_z = sou_z_in
		rec_nx = rec_nx_in
		rec_ox = rec_ox_in
		rec_dx = rec_dx_in
		rec_z = rec_z_in
		edge_par_shot = edge_par_shot_in
		edge_par_rec = edge_par_rec_in
	end subroutine



	subroutine init3(nt_in, dt_in, nh_in, wfld_j_in)
		integer       		:: nt_in, wfld_j_in, nh_in
		real          		:: dt_in
		nt = nt_in
		nh = nh_in
		wfld_j = wfld_j_in
		dt = dt_in
	end subroutine



	subroutine init_alloc() ! Allocations
		integer, external	:: omp_get_num_threads

		!$OMP PARALLEL
		nth=omp_get_num_threads()
		!$OMP END PARALLEL

		write(0,*) " ALLOCATED nth= ", nth
		write(0,*) " ALLOCATED nh= ", nh
		write(0,*) " INIT3 dt= ", dt
		write(0,*) " INIT3 wfld_j= ", wfld_j

		nt2 = (nt-1)*wfld_j+1
		dtw = dt / wfld_j

		call mem_alloc2d(veldt2, nz, nx)
		call mem_alloc2d(vwt, nz, nx)
		call mem_alloc2d(wt, nz, nx)
		call mem_alloc3d(wflds0, np, nt,nth)
		call mem_alloc4d(temp1, nz, nx, nt,nth)
		call mem_alloc4d(temp2, nz, nx, nt,nth)
		call mem_alloc4d(temp1i, nz, nx, nt2,nth)
		call mem_alloc4d(temp2i, nz, nx, nt2,nth)
		call mem_alloc3d(datab, nt, rec_nx, nth)
		call mem_alloc2d(datac, nt, rec_nx)
		call mem_alloc3d(modelb, nz, nx, nth)
		call mem_alloc2d(modelbS, nt, nth)

		call sscale_init(nz, nx, nt, dtw, veldt2, wt)
		call lint_init(nz, nx, nt, wfld_j)
		call step_init(nz, nx, nt2, bc1, bc2, bc3, bc4, d1, d2, veldt2, vwt, wt)
		call scats0_init(nz, nx, nt, nh, wflds0, coor,np)
		call secder_init(nz, nx, nt, dt)

	end subroutine



!=================================================================================
	subroutine init_srcwave_fbi(insrc_in, coor_in) ! Reading in the source wavefield
		real, dimension(:,:,:), target 		:: insrc_in
		real, dimension(:,:), target 		:: coor_in
		insrc=0.0
		insrc => insrc_in
		coor => coor_in
		call scats0_init(nz, nx, nt, nh, insrc, coor,np)
	end subroutine
!=================================================================================


	subroutine init_wavelet(wavelet_in) ! Allocations
		real, dimension(:)				:: wavelet_in 
		call mem_alloc1d(wavelet, nt)
		wavelet = wavelet_in 
	end subroutine



	subroutine model_update(vel_total) ! Copy velocity model and prepare boundary condition stuff
		real, dimension(:,:)			:: vel_total
		vwt = 0.
		wt = 0.
		call AbcWt(vel_total, vwt, wt, dtw, d1, d2, bc1, bc2, bc3, bc4)
		veldt2 = vel_total*vel_total*dtw*dtw
	end subroutine



!==============================================================
!		LINEAR BORN OPERATOR (MODULAR)
!==============================================================



	function L_lop3_func(adj, add, model, data) result(stat)
		logical,intent(in)             :: adj, add
		real,dimension(:)              :: model, data
		integer                        :: stat
		call L_lop3(adj, add, model, data,1)
		stat=0
	end function




	subroutine L_lop3(adj, add, model, data, isou) ! Born operator: perturbation <=> First order scattered data (i.e. data due to perturbation) (equation 22)
		logical, intent(in)                              :: adj, add
		real, dimension(nz, nx)                          :: model
		real, dimension(nt, rec_nx)                      :: data
		integer                                          :: i, ith, isou, ih, start, finish
		real                                             :: cr
		call system_clock(count_rate=cr)
		call system_clock(start)
		ith=1
		temp1(:,:,:,ith) = 0.

! 		Initializing source locaion based on shot number
		sou_x = (isou-1)*sou_dx + sou_ox
		call pad_init(sou_z, sou_x, rec_nx, rec_dx, sou_ox, rec_ox, rec_z, nz, nx, nt)

		!--------------------------------------------------------------------------
		if(adj) then ! Adjoint: First order scattered data => perturbation
				write(0,*) " >>>>>  ADJOINT BORN "

				if(.not. add) modelb(:,:,ith) = 0.
				if(.not. add) model = 0.

				call Rpad_op2(.false., .false., data, temp1(:,:,:,ith),ith) ! Kr operator
				call lint_op2(.false., .false., temp1(:,:,:,ith), temp2i(:,:,:,ith)) ! L operator !SOMETHING IS GOING WRONG HERE. Different depending on whether source is made or read in
				call step_op2(.true., .false., temp1i(:,:,:,ith), temp2i(:,:,:,ith),ith) ! G' operator 
				call lint_op2(.true., .false., temp1(:,:,:,ith), temp1i(:,:,:,ith)) ! L' operator
				call sscale_op2(.true., .false., temp2(:,:,:,ith), temp1(:,:,:,ith)) ! S' operator
				call secder_op2(.true., .false., temp1(:,:,:,ith), temp2(:,:,:,ith)) ! T' operator
				call scats0_op2(.true., .false., modelb(:,:,ith), temp1(:,:,:,ith),ith) ! (U^_s0)' operator

				! Stack over all threads
				do i=1,nth
					model=model + modelb(:,:,i)
				end do

				if (exist_file('rec_wave')) call sep_write(temp1(:,:,:,ith),'rec_wave')
		!--------------------------------------------------------------------------
	    else ! Forward: perturbation => First order scattered data
				write(0,*) " >>>>>  FORWARD BORN "

				if(.not. add) datab(:,:,ith) = 0.
				if(.not. add) data = 0.

				! operator starts here
				modelb(:,:,ith)=model
				call scats0_op2(.false., .false., modelb(:,:,ith), temp1(:,:,:,ith), ith) ! U^_s0 operator
				call secder_op2(.false., .false., temp1(:,:,:,ith), temp2(:,:,:,ith)) ! T operator
				call sscale_op2(.false., .false., temp2(:,:,:,ith), temp1(:,:,:,ith)) ! S operator
				call lint_op2(.false., .false., temp1(:,:,:,ith), temp1i(:,:,:,ith)) ! L operator
				call step_op2(.false., .false., temp1i(:,:,:,ith), temp2i(:,:,:,ith),ith) ! G operator
				call lint_op2(.true., .false., temp1(:,:,:,ith), temp2i(:,:,:,ith)) ! L' operator
				call Rpad_op2(.true., .false., datab, temp1(:,:,:,ith),ith) ! Kr operator

				! Stack over all threads
				do i=1,nth
					data=data+datab(:,:,i)
				end do
	    end if

		call system_clock(finish)
		write(0,*) "TIME FOR BORN L_lop3 = (seconds) ", ((finish-start)/cr)

	end subroutine  







!==============================================================
!		GN HESSIAN
!==============================================================


	function GN_hess_func(adj, add, model, data) result(stat)
		logical,intent(in)             :: adj, add
		real,dimension(:)              :: model, data
		integer                        :: stat
		call GN_hess(adj, add, model, data,1)
		stat=0
	end function


! DP test is pretty close: (9302726. ,9302716.) for single precision
	subroutine GN_hess(adj, add, model, model2, isou) ! Born operator: perturbation <=> First order scattered data (i.e. data due to perturbation) (equation 22)
		logical, intent(in)                              :: adj, add
		real, dimension(nz, nx)                          :: model, model2
		integer                                          :: i, isou, start, finish
		real                                             :: cr

		call system_clock(count_rate=cr)
		call system_clock(start)
		datac = 0.
		!	Either initialize the saved src wavefield, or build one for shot isou
		! call srcmaker(isou)

		if(adj) then ! Adjoint:
			if(.not. add) model = 0.
			call L_lop3(.false., .false., model2, datac, isou)
			call L_lop3(.true., .false., model, datac, isou)
		else
			if(.not. add) model2 = 0.
			call L_lop3(.false., .false., model, datac, isou)	
			call L_lop3(.true., .false., model2, datac, isou)
		end if	

		call system_clock(finish)
		write(0,*) "TIME FOR GN HESSIAN = (seconds) ", ((finish-start)/cr)

	end subroutine   
















! Dot product test works for this. DONT TOUCH!!!!!!
	! subroutine GN_hess(adj, add, model, model2, isou) ! Born operator: perturbation <=> First order scattered data (i.e. data due to perturbation) (equation 22)
	! 	logical, intent(in)                              :: adj, add
	! 	real, dimension(nz, nx, -nh:nh)                  :: model, model2
	! 	integer                                          :: i, isou, start, finish
	! 	real                                             :: cr

	! 	call system_clock(count_rate=cr)
	! 	call system_clock(start)
	! 	datac = 0.

	! 	if(adj) then ! Adjoint:
	! 		if(.not. add) model = 0.
	! 		call L_lop2(.false., .false., model2, datac, isou)
	! 		call L_lop2(.true., .false., model, datac, isou)
	! 	else
	! 		if(.not. add) model2 = 0.
	! 		call L_lop2(.false., .false., model, datac, isou)	
	! 		call L_lop2(.true., .false., model2, datac, isou)
	! 	end if	

	! 	call system_clock(finish)
	! 	write(0,*) "TIME FOR GN HESSIAN = (seconds) ", ((finish-start)/cr)

	! end subroutine   




! !==============================================================
! !		LINEAR BORN OPERATOR
! !==============================================================


! 	function L_lop2_func(adj, add, model, data) result(stat)
! 		logical,intent(in)             :: adj, add
! 		real,dimension(:)              :: model, data
! 		integer                        :: stat
! 		call L_lop2(adj, add, model, data,1)
! 		stat=0
! 	end function




! 	subroutine L_lop2(adj, add, model, data, isou) ! Born operator: perturbation <=> First order scattered data (i.e. data due to perturbation) (equation 22)
! 		logical, intent(in)                              :: adj, add
! 		real, dimension(nz, nx, -nh:nh)                  :: model
! 		real, dimension(nt, rec_nx)                      :: data
! 		integer                                          :: i, ith, isou, ih, start, finish
! 		real                                             :: cr

! 		call system_clock(count_rate=cr)
! 		call system_clock(start)

! 		ith=1
! 		temp1(:,:,:,ith) = 0.

! ! 		Initializing source locaion based on shot number
! 		sou_x = (isou-1)*sou_dx + sou_ox
! 		call pad_init(sou_z, sou_x, rec_nx, rec_dx, sou_ox, rec_ox, rec_z, nz, nx, nt)

! 		! Initialize the source wavefield
! 		if (exist_file('insrc' )) then  
! 			call scats0_init(nz, nx, nt, nh, insrc, coor, np)
! 			write(0,*) " >>>>>  Initialized the saved source wavefield rather than recompute one. "
! 		else ! Precompute the u_s0 wavefield using the wavelet
! 			write(0,*) " >>>>>  Making the source wavefield; couldn't find a precomputed one. "
! 			call Spad_op2(.false., .false., wavelet, temp1(:,:,:,ith),ith) ! Ks' operator
! 			call sscale_op2(.false., .false., temp1(:,:,:,ith), temp2(:,:,:,ith)) ! S operator
! 			call lint_op2(.false., .false., temp2(:,:,:,ith), temp1i(:,:,:,ith)) ! L operator
! 			call step_op2(.false., .false., temp1i(:,:,:,ith), temp2i(:,:,:,ith), ith) ! G operator
!     		call lint_op2(.true., .false., wflds0(:,:,:,ith), temp2i(:,:,:,ith)) ! L' operator
! 		end if


! 		!--------------------------------------------------------------------------
! 		if(adj) then ! Adjoint: First order scattered data => perturbation
! 				write(0,*) " >>>>>  ADJOINT BORN "

! 				if(.not. add) modelb(:,:,:,ith) = 0.
! 				if(.not. add) model = 0.

! 				call Rpad_op2(.false., .false., data, temp1(:,:,:,ith),ith) ! Kr operator
! 				call lint_op2(.false., .false., temp1(:,:,:,ith), temp2i(:,:,:,ith)) ! L operator !SOMETHING IS GOING WRONG HERE. Different depending on whether source is made or read in
! 				call step_op2(.true., .false., temp1i(:,:,:,ith), temp2i(:,:,:,ith),ith) ! G' operator 
! 				call lint_op2(.true., .false., temp1(:,:,:,ith), temp1i(:,:,:,ith)) ! L' operator
! 				call sscale_op2(.true., .false., temp2(:,:,:,ith), temp1(:,:,:,ith)) ! S' operator
! 				call secder_op2(.true., .false., temp1(:,:,:,ith), temp2(:,:,:,ith)) ! T' operator
! 				call scats0_op2(.true., .false., modelb(:,:,:,ith), temp1(:,:,:,ith),ith) ! (U^_s0)' operator

! 				! Stack over all threads
! 				do i=1,nth
! 					model=model + modelb(:,:,:,i)
! 				end do

! 		!--------------------------------------------------------------------------
! 	    else ! Forward: perturbation => First order scattered data
! 				write(0,*) " >>>>>  FORWARD BORN "

! 				if(.not. add) datab(:,:,ith) = 0.
! 				if(.not. add) data = 0.

! 				! operator starts here
! 				modelb(:,:,:,ith)=model
! 				call scats0_op2(.false., .false., modelb(:,:,:,ith), temp1(:,:,:,ith),ith) ! U^_s0 operator
! 				call secder_op2(.false., .false., temp1(:,:,:,ith), temp2(:,:,:,ith)) ! T operator
! 				call sscale_op2(.false., .false., temp2(:,:,:,ith), temp1(:,:,:,ith)) ! S operator
! 				call lint_op2(.false., .false., temp1(:,:,:,ith), temp1i(:,:,:,ith)) ! L operator
! 				call step_op2(.false., .false., temp1i(:,:,:,ith), temp2i(:,:,:,ith),ith) ! G operator
! 				call lint_op2(.true., .false., temp1(:,:,:,ith), temp2i(:,:,:,ith)) ! L' operator
! 				call Rpad_op2(.true., .false., datab, temp1(:,:,:,ith),ith) ! Kr operator

! 				! Stack over all threads
! 				do i=1,nth
! 					data=data+datab(:,:,i)
! 				end do

! 	    end if

! 		call system_clock(finish)
! 		write(0,*) "TIME FOR BORN L_lop2 = (seconds) ", ((finish-start)/cr)

! 	end subroutine   





! 	subroutine srcmaker(isou) ! Born operator: perturbation <=> First order scattered data (i.e. data due to perturbation) (equation 22)
! 		integer                          :: i, ith, isou, ih, start, finish
! 		real                             :: cr
! 		call system_clock(count_rate=cr)
! 		call system_clock(start)
! 		ith=1
! 		temp1(:,:,:,ith) = 0.

! ! 		Initializing source locaion based on shot number
! 		sou_x = (isou-1)*sou_dx + sou_ox
! 		call pad_init(sou_z, sou_x, rec_nx, rec_dx, sou_ox, rec_ox, rec_z, nz, nx, nt)

! 		! Initialize the source wavefield
! 		if (exist_file('insrc' )) then  
! 			call scats0_init(nz, nx, nt, nh, insrc, coor, np)
! 			write(0,*) " >>>>>  Initialized the saved source wavefield rather than recompute one. "
! 		else ! Precompute the u_s0 wavefield using the wavelet
! 			write(0,*) " >>>>>  NEED TO MAKE A SOURCE WAVEFIELD AND THEN COMPRESS IT SO IT CAN BE USED AS INPUT "
! 			! call Spad_op2(.false., .false., wavelet, temp1(:,:,:,ith),ith) ! Ks' operator
! 			! call sscale_op2(.false., .false., temp1(:,:,:,ith), temp2(:,:,:,ith)) ! S operator
! 			! call lint_op2(.false., .false., temp2(:,:,:,ith), temp1i(:,:,:,ith)) ! L operator
! 			! call step_op2(.false., .false., temp1i(:,:,:,ith), temp2i(:,:,:,ith), ith) ! G operator
!    !  		call lint_op2(.true., .false., wflds0(:,:,:,ith), temp2i(:,:,:,ith)) ! L' operator
! 		end if
! 		write(0,*) "sum(wflds0) = ", sum(wflds0)

! 		call system_clock(finish)
! 		write(0,*) "TIME FOR INITIALIZING THE SOURCE WAVEFIELD INTO FBI = (seconds) ", ((finish-start)/cr)

! 	end subroutine








end module 
