module fourier

  implicit none
  include "fftw3.f"

  complex, dimension(:), allocatable, private  :: in1, out1, in2, out2
  integer, private                             :: n1, n2
  integer*8, private                           :: plan1f, plan1b, plan2f, plan2b
  contains

!//////////////////////////////////////////////////////////////////////////

!	NEED TO CHANGE CDATA TO 1D FOR FFT1D.F90 TO WORK. SO THAT MEANS
!	NEW SUBROUTINES NEED TO BE ADDED FOR THAT. 

!//////////////////////////////////////////////////////////////////////////


  subroutine ft_init(n1_in, n2_in)
    integer               :: n1_in, n2_in
    n1 = n1_in
    n2 = n2_in
    allocate(in1(n1), out1(n1))
    allocate(in2(n2), out2(n2))
    in1 = 0.
    out1 = 0.
    in2 = 0.
    out2 = 0.
    call sfftw_plan_dft_1d(plan1f, n1, in1, out1, FFTW_FORWARD, FFTW_ESTIMATE)
    call sfftw_plan_dft_1d(plan1b, n1, out1, in1, FFTW_BACKWARD, FFTW_ESTIMATE)
    call sfftw_plan_dft_1d(plan2f, n2, in2, out2, FFTW_FORWARD, FFTW_ESTIMATE)
    call sfftw_plan_dft_1d(plan2b, n2, out2, in2, FFTW_BACKWARD, FFTW_ESTIMATE)
  end subroutine





  subroutine ft1axis(adj, sign1, cdata)
  ! Forward and Ajoint/Inverse Fourier Transform on axis 1
    logical, intent(in)		:: adj
    integer			:: i2
    complex, dimension(:,:) 	:: cdata
    real			:: sign1
    if(adj) then
      do i2=1,n2
        out1(1:(n1+1)/2) = cdata(n1-(n1+1)/2+1:n1,i2)
        out1((n1+1)/2+1:n1) = cdata(1:n1-(n1+1)/2,i2)
        call sfftw_execute_dft(plan1b, out1, in1)
        cdata(:,i2) = in1
      enddo
    else
      do i2=1,n2
        in1 = cdata(:,i2)
        call sfftw_execute_dft(plan1f, in1, out1)
        cdata(:,i2) = out1
        cdata(1:n1-(n1+1)/2,i2) = out1((n1+1)/2+1:n1) ! cdata(first_half)  = out1(second_half)
        cdata(n1-(n1+1)/2+1:n1,i2) = out1(1:(n1+1)/2) ! cdata(second_half) = out1(first_half)
      enddo
    endif
    cdata = sign1*cdata/sqrt(1.*n1)
  end subroutine





  subroutine ft2axis(adj, sign2, cdata)
  ! Forward and Ajoint/Inverse Fourier Transform on axis 2
    logical, intent(in)		:: adj
    integer			:: i1
    complex, dimension(:,:)	:: cdata
    real			:: sign2
    if(adj) then

   !#####
   ! Fix this so there is an adjoint part
	do i1=1,n1
       	  out2(1:(n2+1)/2) = cdata(i1,n2-(n2+1)/2+1:n2)
       	  out2((n2+1)/2+1:n2) = cdata(i1,1:n2-(n2+1)/2)
       	  call sfftw_execute_dft(plan2b, out2, in2)
       	  cdata(i1,:) = in2
      	enddo

  !#####
    else
      do i1=1,n1
        in2 = cdata(i1,:)
        call sfftw_execute_dft(plan2f, in2, out2)
        cdata(i1,1:n2-(n2+1)/2) = out2((n2+1)/2+1:n2) ! cdata(first_half)  = out1(second_half)
        cdata(i1,n2-(n2+1)/2+1:n2) = out2(1:(n2+1)/2) ! cdata(second_half) = out1(first_half)
      enddo

    endif
    cdata = sign2*cdata/sqrt(1.*n2)
  end subroutine





  subroutine ft1axis_1D(adj, sign1, cdata)
  ! Forward and Ajoint/Inverse Fourier Transform on axis 1
    logical, intent(in)		:: adj
    complex, dimension(:) 	:: cdata
    real			:: sign1
    if(adj) then
        out1(1:(n1+1)/2) = cdata(n1-(n1+1)/2+1:n1)
        out1((n1+1)/2+1:n1) = cdata(1:n1-(n1+1)/2)
        call sfftw_execute_dft(plan1b, out1, in1)
        cdata(:) = in1
    else
        in1 = cdata(:)
        call sfftw_execute_dft(plan1f, in1, out1)
        cdata(:) = out1
! First half of cdata becomes second half of out1
        cdata(1:n1-(n1+1)/2) = out1((n1+1)/2+1:n1) 
! Second half of cdata becomes first half of out1
        cdata(n1-(n1+1)/2+1:n1) = out1(1:(n1+1)/2)
    endif
    cdata = sign1*cdata/sqrt(1.*n1)
  end subroutine




end module
