module initialize_wemva_hess
	use sep
	use wemva_mod
	use calc_mod
	use mem_mod
	!use levelset_mod
	!use drlse_mod
	use secder_mod
	implicit none
	contains



	subroutine init_wemva_hessian(nz,nx,nt,nh,dz,dx,dt,oz,ox,ot,bc1,bc2,bc3,bc4,wfld_j,sou_nx,sou_dx,sou_ox,rec_nx,rec_dx,rec_ox,rec_z,tstep,vstep,niter,np)
		!----------------------------------------------------------------------------------------------
		integer								:: nt, ntw, nz, nx, wfld_j, niter, smooth_rad, nh, np
		integer								:: sou_nx, sou_ox, sou_dx, sou_z, sou_x, isou
		integer								:: rec_nx, rec_ox, rec_dx, rec_z, rec_x, irec, edge_par_shot, edge_par_rec
		integer								:: bc1, bc2, bc3, bc4
		integer								:: fs, fr, fs_z, fr_z, nltest, ltest, lmax, numgamma, nt2
		real								:: dt, dz, oz, dx, ox, ot, tstep, vstep, mu, v1, tdelay, vc, mrange, vel_max, vel_min

		real, dimension(:), allocatable     :: wavelet
		call sep_init()

		write(0,*) " Initializing WEVA_HESS (initialize2.f90)"
		!----------------------------------------------------------------------------------------------
		call from_param("dz", dz)
		call from_param("dx", dx)
		!----------------------------------------------------------------------------------------------
		call from_aux('velmod', "n1", nz)
		call from_aux('velmod', "o1", oz)
		call from_aux('velmod', "n2", nx)
		call from_aux('velmod', "o2", ox)
		call from_aux("waveletin", "d1", dt)
		call from_aux("waveletin", "n1", ntw)

	    call from_aux("src_wave","n1",np)

		!----------------------------------------------------------------------------------------------

		call from_param("q1", v1)
		call from_param("vc", vc)
		call from_param("ot", ot)
		call from_param("smooth_rad", smooth_rad)
		call from_param("mrange", mrange)
		call from_param("vel_min", vel_min)		
		call from_param("vel_max", vel_max)	

		call from_param("bc1", bc1)
		call from_param("bc2", bc2)
		call from_param("bc3", bc3)
		call from_param("bc4", bc4)
		call from_param("wfld_j", wfld_j)
		call from_param("niter", niter)
		call from_param("nt", nt)
		call from_param("nh", nh)
		call from_param("edge_par_shot", edge_par_shot)
		call from_param("edge_par_rec", edge_par_rec)
		call from_param("ns", sou_nx, nx      )
		call from_param("fs", fs    , nx/2    ); sou_ox=fs!+1
		call from_param("js", sou_dx, 1       )
		call from_param("nr", rec_nx, sou_nx  )
		call from_param("fr", fr    , nx/2    ); rec_ox=fr!+1
		call from_param("jr", rec_dx, 1       )
		call from_param("fs_z", fs_z); sou_z=fs_z!+1
		call from_param("fr_z", fr_z); rec_z=fr_z!+1

		nt2 = (nt-1)*wfld_j+1


		call from_param("numgamma", numgamma)	

		call from_param("nltest", nltest)	
		call from_param("ltest", ltest)	
		call from_param("lmax", lmax)			


		!----------------------------------------------------------------------------------------------
		! Check to make sure that the "nt" from the wavelet is longer than the nt from the parameter (recoding time length)
		! Im not certian, but I am pretty sure that the wavelet must be longer than the recorded data length.
		if(nt > ntw) call erexit("In the parameter file, nt is too small. It is currently smaller than the 'nt' of the wavelet. So either make the wavelet 'nt' bigger, or the parameter file nt smaller.")
		
		!----------------------------------------------------------------------------------------------
		if(sou_ox < 1 .or. (sou_nx-1)*sou_dx+sou_ox > nx) call erexit("sou axis values are not correct")
		if(sou_z  < 1 .or. sou_z > nz) call erexit("sou_z value is not correct")
		if(rec_ox < 1 .or. (rec_nx-1)*rec_dx+rec_ox > nx) call erexit("rec axis values are not correct")
		if(rec_z  < 1 .or. rec_z > nz) call erexit("rec_z value is not correct")


	  	!----------------------------------------------------------------------------------------------
		call mem_alloc1d(wavelet, nt)
		call sep_read(wavelet, "waveletin")


		!----------------------------------------------------------------------------------------------
		call initWH1(nz, oz, dz, nx, ox, dx, bc1, bc2, bc3, bc4)
		call initWH2(sou_nx, sou_ox, sou_dx, sou_z, rec_nx, rec_ox, rec_dx, rec_z, edge_par_shot,edge_par_rec)
		call initWH3(nt, dt, nh, wfld_j)
		call init_WHalloc(wavelet)

		!----------------------------------------------------------------------------------------------
		! call levelset_init(nz, nx, v1, nt, rec_nx, sou_nx, dx, dz, dt, smooth_rad, mrange, vel_max, vel_min, nltest, numgamma, lmax)
		! call calc_init( nz, nx, nt, rec_nx, sou_nx, tdelay, vc, v1, ((sou_ox-1)*dx), (sou_dx*dx), ((rec_ox-1)*dx), (rec_dx*dx), dt, dx, dz, oz, ox, ((sou_z-1)*dz)) ! should these be ((sou_ox-1)*dx)??
		call calc_init( nz, nx, nt, rec_nx, sou_nx, 0.1, vc, v1, ((sou_ox-1)*dx), (sou_dx*dx), ((rec_ox-1)*dx), (rec_dx*dx), dt, dx, dz, oz, ox, ((sou_z-1)*dz)) ! should these be ((sou_ox-1)*dx)??
		call secder_init(nz, nx, nt, dt)
		!call drlse_init(nz, nx, dz, dx)
		! call drlse_init(nz, nx, 1.0, 1.0)
		! call NLnewton_init( nz, nx, nt, rec_nx, sou_nx, 0.0000001, lmax, nltest, ltest, v1, mrange,vel_max, vel_min)
		! call Lnewton_init( nz, nx, nt, rec_nx, sou_nx, 0.0000001, lmax, nltest, ltest, v1, mrange,vel_max, vel_min)
		!----------------------------------------------------------------------------------------------


		if (exist_file('wflds0')) then
			call to_history("n1"    ,nz		,'wflds0')
			call to_history("n2"    ,nx		,'wflds0')
			call to_history("n3"    ,nt		,'wflds0')  ! To account for the first frame being phi0
			call to_history("d1"    ,dz		,'wflds0')
			call to_history("d2"    ,dx		,'wflds0')
			call to_history("d3"    ,dt		,'wflds0')
			call to_history("o1"    ,oz		,'wflds0')
			call to_history("o2"    ,ox		,'wflds0')
			call to_history("o3"    ,ot		,'wflds0')
			call to_history("label1",'z [km]'	,'wflds0')
			call to_history("label2",'x [km]'	,'wflds0')
			call to_history("label3",'time'		,'wflds0')
		end if


		

		if (exist_file('temp1')) then
			call to_history("n1"    ,nz		,'temp1')
			call to_history("n2"    ,nx		,'temp1')
			call to_history("n3"    ,nt		,'temp1')  ! To account for the first frame being phi0
			call to_history("d1"    ,dz		,'temp1')
			call to_history("d2"    ,dx		,'temp1')
			call to_history("d3"    ,dt		,'temp1')
			call to_history("o1"    ,oz		,'temp1')
			call to_history("o2"    ,ox		,'temp1')
			call to_history("o3"    ,ot		,'temp1')
			call to_history("label1",'z [km]'	,'temp1')
			call to_history("label2",'x [km]'	,'temp1')
			call to_history("label3",'time'		,'temp1')
		end if


		if (exist_file('datascat')) then
			call to_history("n1"    ,nz		,'datascat')
			call to_history("n2"    ,nx		,'datascat')
			call to_history("n3"    ,nt		,'datascat')  ! To account for the first frame being phi0
			call to_history("d1"    ,dz		,'datascat')
			call to_history("d2"    ,dx		,'datascat')
			call to_history("d3"    ,dt		,'datascat')
			call to_history("o1"    ,oz		,'datascat')
			call to_history("o2"    ,ox		,'datascat')
			call to_history("o3"    ,ot		,'datascat')
			call to_history("label1",'z [km]'	,'datascat')
			call to_history("label2",'x [km]'	,'datascat')
			call to_history("label3",'time'		,'datascat')
		end if



		if (exist_file('imagebuild')) then
			call to_history("n1"    ,nz		,'imagebuild')
			call to_history("n2"    ,nx		,'imagebuild')
			call to_history("n3"    ,nt		,'imagebuild')  ! To account for the first frame being phi0
			call to_history("d1"    ,dz		,'imagebuild')
			call to_history("d2"    ,dx		,'imagebuild')
			call to_history("d3"    ,dt		,'imagebuild')
			call to_history("o1"    ,oz		,'imagebuild')
			call to_history("o2"    ,ox		,'imagebuild')
			call to_history("o3"    ,ot		,'imagebuild')
			call to_history("label1",'z [km]'	,'imagebuild')
			call to_history("label2",'x [km]'	,'imagebuild')
			call to_history("label3",'time'		,'imagebuild')
		end if



		if (exist_file('temp1i')) then
			call to_history("n1"    ,nz			,'temp1i')
			call to_history("n2"    ,nx			,'temp1i')
			call to_history("n3"    ,nt2		,'temp1i')  ! To account for the first frame being phi0
			call to_history("d1"    ,dz			,'temp1i')
			call to_history("d2"    ,dx			,'temp1i')
			call to_history("d3"    ,dt			,'temp1i')
			call to_history("o1"    ,oz			,'temp1i')
			call to_history("o2"    ,ox			,'temp1i')
			call to_history("o3"    ,ot			,'temp1i')
			call to_history("label1",'z [km]'	,'temp1i')
			call to_history("label2",'x [km]'	,'temp1i')
			call to_history("label3",'time'		,'temp1i')
		end if



		if (exist_file('temp2')) then
			call to_history("n1"    ,nz		,'temp2')
			call to_history("n2"    ,nx		,'temp2')
			call to_history("n3"    ,nt		,'temp2')  ! To account for the first frame being phi0
			call to_history("n4"    ,niter		,'temp2')  ! To account for the first frame being phi0
			call to_history("d1"    ,dz		,'temp2')
			call to_history("d2"    ,dx		,'temp2')
			call to_history("d3"    ,dt		,'temp2')
			call to_history("d4"    ,1.0		,'temp2')
			call to_history("o1"    ,oz		,'temp2')
			call to_history("o2"    ,ox		,'temp2')
			call to_history("o3"    ,ot		,'temp2')
			call to_history("o4"    ,0.0		,'temp2')
			call to_history("label1",'z [km]'	,'temp2')
			call to_history("label2",'x [km]'	,'temp2')
			call to_history("label3",'time'		,'temp2')
			call to_history("label4",'iteration'	,'temp2')
		end if





		if (exist_file('res_wave')) then
			call to_history("n1"    ,nz			,'res_wave')
			call to_history("n2"    ,nx			,'res_wave')
			call to_history("n3"    ,nt			,'res_wave')  ! To account for the first frame being phi0
			call to_history("d1"    ,dz			,'res_wave')
			call to_history("d2"    ,dx			,'res_wave')
			call to_history("d3"    ,dt			,'res_wave')
			call to_history("o1"    ,oz			,'res_wave')
			call to_history("o2"    ,ox			,'res_wave')
			call to_history("o3"    ,ot			,'res_wave')
			call to_history("label1",'z [km]'	,'res_wave')
			call to_history("label2",'x [km]'	,'res_wave')
			call to_history("label3",'time'		,'res_wave')
		end if





		if (exist_file('scat_wave')) then
			call to_history("n1"    ,nz			,'scat_wave')
			call to_history("n2"    ,nx			,'scat_wave')
			call to_history("n3"    ,nt			,'scat_wave')  ! To account for the first frame being phi0
			call to_history("d1"    ,dz			,'scat_wave')
			call to_history("d2"    ,dx			,'scat_wave')
			call to_history("d3"    ,dt			,'scat_wave')
			call to_history("o1"    ,oz			,'scat_wave')
			call to_history("o2"    ,ox			,'scat_wave')
			call to_history("o3"    ,ot			,'scat_wave')
			call to_history("label1",'z [km]'	,'scat_wave')
			call to_history("label2",'x [km]'	,'scat_wave')
			call to_history("label3",'time'		,'scat_wave')
		end if

		if (exist_file('image')) then
			call to_history("n1"    ,nz			,'image')
			call to_history("n2"    ,nx			,'image')
			call to_history("d1"    ,dz			,'image')
			call to_history("d2"    ,dx			,'image')
			call to_history("o1"    ,oz			,'image')
			call to_history("o2"    ,ox			,'image')
			call to_history("label1",'z [km]'	,'image')
			call to_history("label2",'x [km]'	,'image')
		end if


		if (exist_file('startmod')) then
			call to_history("n1"    ,nz			,'startmod')
			call to_history("n2"    ,nx			,'startmod')
			call to_history("d1"    ,dz			,'startmod')
			call to_history("d2"    ,dx			,'startmod')
			call to_history("o1"    ,oz			,'startmod')
			call to_history("o2"    ,ox			,'startmod')
			call to_history("label1",'z [km]'	,'startmod')
			call to_history("label2",'x [km]'	,'startmod')
		end if


		if (exist_file('wemva1')) then
			call to_history("n1"    ,nz			,'wemva1')
			call to_history("n2"    ,nx			,'wemva1')
			call to_history("d1"    ,dz			,'wemva1')
			call to_history("d2"    ,dx			,'wemva1')
			call to_history("o1"    ,oz			,'wemva1')
			call to_history("o2"    ,ox			,'wemva1')
			call to_history("label1",'z [km]'	,'wemva1')
			call to_history("label2",'x [km]'	,'wemva1')
		end if


		if (exist_file('wemva2')) then
			call to_history("n1"    ,nz			,'wemva2')
			call to_history("n2"    ,nx			,'wemva2')
			call to_history("d1"    ,dz			,'wemva2')
			call to_history("d2"    ,dx			,'wemva2')
			call to_history("o1"    ,oz			,'wemva2')
			call to_history("o2"    ,ox			,'wemva2')
			call to_history("label1",'z [km]'	,'wemva2')
			call to_history("label2",'x [km]'	,'wemva2')
		end if

	end subroutine

end module



