
module pad_mod
    implicit none
    integer, private                   :: nth, sou_z, sou_x, rec_nx, nt, nz, nx
    real, private                      :: rec_dx, sou_ox, rec_ox, rec_z
    contains



    subroutine pad_init(sou_z_in, sou_x_in, rec_nx_in, rec_dx_in, sou_ox_in, rec_ox_in, rec_z_in, nz_in, nx_in, nt_in)
        integer                        :: sou_z_in, sou_x_in, rec_nx_in, rec_dx_in, rec_z_in, nz_in, nx_in, nt_in
        integer                        :: sou_ox_in,  rec_ox_in
        integer,external               :: omp_get_num_threads
        !$OMP PARALLEL
        nth=omp_get_num_threads()
        !$OMP END PARALLEL
        sou_z = sou_z_in 
        sou_x = sou_x_in
        rec_z = rec_z_in 
        rec_nx = rec_nx_in
        rec_dx = rec_dx_in
        sou_ox = sou_ox_in
        rec_ox = rec_ox_in
        rec_z = rec_z_in
        nt = nt_in
        nz = nz_in
        nx = nx_in
    end subroutine



    function Spad_op(adj, add, model, data) result(stat)
        logical,intent(in)             :: adj, add
        real,dimension(:)  :: model, data
        integer                        :: stat
        call Spad_op2(adj, add, model, data, nth)
        stat=0
    end function



    ! Source Trace --------->> Padded wavefield cube
    subroutine Spad_op2(adj, add, model, data,ith)
        logical,intent(in)                         :: adj, add 
        real, dimension(nt)            :: model
        real, dimension(nz,nx,nt,nth)  :: data
        integer                                    :: it, ith
        ith=1
        if(adj) then
            if(.not. add) model = 0.
            do it=1,nt
                 model(it) =  model(it) + data(sou_z,sou_x,it,ith)
            end do
        else
            if(.not. add) data = 0.
            do it=1,nt
                data(sou_z,sou_x,it,ith) = data(sou_z,sou_x,it,ith) + model(it)
            end do
        end if
    end subroutine



    function Rpad_op(adj, add, model, data) result(stat)
        logical,intent(in)             :: adj, add
        real,dimension(:)  :: model, data
        integer                        :: stat
        call Rpad_op2(adj, add, model, data, nth)
        stat=0
    end function



    ! Receiver Trace --------->> Padded wavefield cube
    subroutine Rpad_op2(adj, add, model, data,ith)
        logical,intent(in)                         :: adj, add 
        real, dimension(nt,rec_nx)     :: model
        real, dimension(nz,nx,nt,nth)  :: data
        integer                                    :: it, ith, irec, rec_x
        ith=1
        if(adj) then
            if(.not. add) model = 0.
            do irec=1,rec_nx
                rec_x = (irec-1)*rec_dx + abs(sou_ox-rec_ox) + sou_x 
                do it=1,nt
                     model(it,irec) =  model(it,irec) + data(rec_z,rec_x,it,ith)
                end do
            end do
        else
            if(.not. add) data = 0.
            do irec=1,rec_nx
                rec_x = (irec-1)*rec_dx + abs(sou_ox-rec_ox) + sou_x 
                do it=1,nt
                    data(rec_z,rec_x,it,ith) = data(rec_z,rec_x,it,ith) + model(it,irec)
                end do
            end do
        end if
    end subroutine


end module





! !==================================================================================
! !               DOUBLE PRECISION
! !==================================================================================


! module pad_mod
!     implicit none
!     integer, private                   :: nth, sou_z, sou_x, rec_nx, nt, nz, nx
!     real, private                      :: rec_dx, sou_ox, rec_ox, rec_z
!     contains



!     subroutine pad_init(sou_z_in, sou_x_in, rec_nx_in, rec_dx_in, sou_ox_in, rec_ox_in, rec_z_in, nz_in, nx_in, nt_in)
!         integer                        :: sou_z_in, sou_x_in, rec_nx_in, rec_dx_in, rec_z_in, nz_in, nx_in, nt_in
!         integer                        :: sou_ox_in,  rec_ox_in
!         integer,external               :: omp_get_num_threads
!         !$OMP PARALLEL
!         nth=omp_get_num_threads()
!         !$OMP END PARALLEL
!         sou_z = sou_z_in 
!         sou_x = sou_x_in
!         rec_z = rec_z_in 
!         rec_nx = rec_nx_in
!         rec_dx = rec_dx_in
!         sou_ox = sou_ox_in
!         rec_ox = rec_ox_in
!         rec_z = rec_z_in
!         nt = nt_in
!         nz = nz_in
!         nx = nx_in
!     end subroutine



!     function Spad_op(adj, add, model, data) result(stat)
!         logical,intent(in)             :: adj, add
!         double precision,dimension(:)  :: model, data
!         integer                        :: stat
!         call Spad_op2(adj, add, model, data, nth)
!         stat=0
!     end function



!     ! Source Trace --------->> Padded wavefield cube
!     subroutine Spad_op2(adj, add, model, data,ith)
!         logical,intent(in)                         :: adj, add 
!         double precision, dimension(nt)            :: model
!         double precision, dimension(nz,nx,nt,nth)  :: data
!         integer                                    :: it, ith
!         ith=1
!         if(adj) then
!             if(.not. add) model = 0.
!             do it=1,nt
!                  model(it) =  model(it) + data(sou_z,sou_x,it,ith)
!             end do
!         else
!             if(.not. add) data = 0.
!             do it=1,nt
!                 data(sou_z,sou_x,it,ith) = data(sou_z,sou_x,it,ith) + model(it)
!             end do
!         end if
!     end subroutine



!     function Rpad_op(adj, add, model, data) result(stat)
!         logical,intent(in)             :: adj, add
!         double precision,dimension(:)  :: model, data
!         integer                        :: stat
!         call Rpad_op2(adj, add, model, data, nth)
!         stat=0
!     end function



!     ! Receiver Trace --------->> Padded wavefield cube
!     subroutine Rpad_op2(adj, add, model, data,ith)
!         logical,intent(in)                         :: adj, add 
!         double precision, dimension(nt,rec_nx)     :: model
!         double precision, dimension(nz,nx,nt,nth)  :: data
!         integer                                    :: it, ith, irec, rec_x
!         ith=1
!         if(adj) then
!             if(.not. add) model = 0.
!             do irec=1,rec_nx
!                 rec_x = (irec-1)*rec_dx + abs(sou_ox-rec_ox) + sou_x 
!                 do it=1,nt
!                      model(it,irec) =  model(it,irec) + data(rec_z,rec_x,it,ith)
!                 end do
!             end do
!         else
!             if(.not. add) data = 0.
!             do irec=1,rec_nx
!                 rec_x = (irec-1)*rec_dx + abs(sou_ox-rec_ox) + sou_x 
!                 do it=1,nt
!                     data(rec_z,rec_x,it,ith) = data(rec_z,rec_x,it,ith) + model(it,irec)
!                 end do
!             end do
!         end if
!     end subroutine


! end module
