module solver_mod  
! 0 = F m - d
  use sep
  implicit none
  real, private      :: threshold, scale, resval
  contains


  subroutine solver_init(threshold_in)
    real             :: threshold_in
    threshold = threshold_in
  end subroutine


  subroutine solver_tiny( m,d, Fop, stepper, niter, m0,resd)
    optional ::                                     m0,resd
!--------------------begin definitions -------
    interface  
      integer function Fop(adj,add,m,d)
        real,dimension(:) :: m,d
        logical,intent(in):: adj,add
      end function 
      integer function stepper(forget,m,g,rr,gg)
        real,dimension(:) ::          m,g,rr,gg    
        logical           ::   forget
      end function
    end interface 



    real, dimension(:),    intent(in)  :: d       ! data
    real, dimension(:),    intent(out) :: m       ! model
    real, dimension(:),    intent(in)  :: m0      ! initial model
    integer,               intent(in)  :: niter   ! number of iterations
    integer                            :: iter    ! iteration number
    real, dimension(size( m))          :: g       ! gradient (dm)
    real, dimension(size( d)),target   :: rr      ! data residual ( vector)
    real, dimension(:),       pointer  :: rd      ! data residual (pointer)
    real, dimension(size( d)),target   :: gg      ! conj grad     ( vector)
    real, dimension(:),       pointer  :: gd      ! conj grad     (pointer)
    integer                            :: stat    ! status flag
    integer                            :: wc, srite    ! <<<<<<<<<<<<<<<<<<<<<<<<<<   function stuff
    real, dimension(:),    intent(out) :: resd    ! residual
    logical                            :: forget

!------------------------------------------
    call sep_init()
    if (exist_file('rnorm')) then
        call to_history("n2", 1, "rnorm")
        call to_history("o1", 1., "rnorm")
        call to_history("d1", 1., "rnorm")
    end if
!------------------------------------------

    rd => rr(1:size( d))
    gd => gg(1:size( d))
!--------- begin initialization -----------
    rd = -d                               ! Rd =      - d
    m  = 0
    if (present( m0)) then
      m = m0                              !  m =        m0
      stat = Fop(.false.,.true.,m,rd)     ! Rd = Rd + F m0
    end if
    write(0,*) "       iter      rd*rd            g*g"
!--------- begin iterations -----------
    do iter = 1,niter  
      stat = Fop( .true.,.false.,g,rd)        !  g =  F' Rd
      stat = Fop(.false.,.false.,g,gd)        ! Gd =  F  g
      stat = stepper(forget, m,g, rr,gg)     ! m = m+dm; R = R + dR

!------------------------------------------
! Check if reached residual threshold
      if (iter==1) then
        scale=sum(rd*rd)
        resval=100.0
      else
        resval=sum(rd*rd)*100.0/scale
        if (resval<threshold) exit
      end if
      write(0,*) "resval = ", resval
      write(0,*) "threshold = ", threshold
!------------------------------------------
      if (exist_file('rnorm')) then 
          call to_history("n1", iter,"rnorm")
          wc = srite("rnorm",sum(rd*rd),4)
      end if
!------------------------------------------

      write(0,*) " ===================================================================="
      write(0,*) iter, sum(rd*rd), sum(g*g)

    end do 
    if (present( resd)) then
      resd = rd
    end if

  end subroutine 








end module 
