

module step_mod

  use comb_mod
  use mem_mod
  implicit none
  integer, private                                            :: n1, n2, nt2, it2,nth
  real, dimension(:,:), pointer, private          :: wt
  real, dimension(:,:,:), allocatable, private    :: temp
  contains




  subroutine step_init(n1_in, n2_in, nt2_in, bc1_in, bc2_in, bc3_in, bc4_in, d1_in, d2_in, veldt2_in, vwt_in, wt_in)
    integer                                    :: n1_in, n2_in, nt2_in, bc1_in, bc2_in, bc3_in, bc4_in
    real                                       :: d1_in, d2_in
    real, dimension(:,:), target   :: veldt2_in, vwt_in, wt_in
    integer,external                           :: omp_get_num_threads
    !$OMP PARALLEL
    nth=omp_get_num_threads()
    !$OMP END PARALLEL
    n1 = n1_in
    n2 = n2_in
    nt2 = nt2_in
    wt => wt_in
    allocate(temp(n1,n2,nth))
    call comb_init(n1_in, n2_in, bc1_in, bc2_in, bc3_in, bc4_in, d1_in, d2_in, veldt2_in, vwt_in, wt_in)
  end subroutine





  function step_op(adj, add, model, data) result(stat)
    logical,intent(in) :: adj, add
    real,dimension(:)  :: model, data
    integer            :: stat
    call step_op2(adj, add, model, data,1)
    stat=0
  end function




  subroutine step_op2(adj, add, model, data,ith) ! This operator is the recursive G operator (equation 9)
    logical,intent(in)         :: adj, add 
    real, dimension(n1,n2,nt2) :: model
    real, dimension(n1,n2,nt2) :: data
    integer :: ith
    if(adj) then
      if(.not. add) model = 0.
      call comb_op2(adj, .false., temp(:,:,ith), model(:,:,nt2),ith)
      model(:,:,nt2-1) = model(:,:,nt2-1) + data(:,:,nt2) + temp(:,:,ith)
      do it2=nt2-2,1,-1
        call comb_op2(adj, .false., temp(:,:,ith), model(:,:,it2+1),ith)
        model(:,:,it2) = model(:,:,it2) + data(:,:,it2+1) + temp(:,:,ith) + (wt-1.)*model(:,:,it2+2)
      end do
    else
      if(.not. add) data = 0.
      call comb_op2(adj, .false., data(:,:,1), temp(:,:,ith),ith)
      data(:,:,2) = data(:,:,2) + model(:,:,1) + temp(:,:,ith)
      do it2=3,nt2
        call comb_op2(adj, .false., data(:,:,it2-1), temp(:,:,ith),ith)
        data(:,:,it2) = data(:,:,it2) + model(:,:,it2-1) + temp(:,:,ith) + (wt-1.)*data(:,:,it2-2)
      end do
    end if
  end subroutine




end module




!==================================================================================
!               DOUBLE PRECISION
!==================================================================================


! module step_mod

!   use comb_mod
!   use mem_mod
!   implicit none
!   integer, private                                            :: n1, n2, nt2, it2,nth
!   double precision, dimension(:,:), pointer, private          :: wt
!   double precision, dimension(:,:,:), allocatable, private    :: temp
!   contains




!   subroutine step_init(n1_in, n2_in, nt2_in, bc1_in, bc2_in, bc3_in, bc4_in, d1_in, d2_in, veldt2_in, vwt_in, wt_in)
!     integer                                    :: n1_in, n2_in, nt2_in, bc1_in, bc2_in, bc3_in, bc4_in
!     real                                       :: d1_in, d2_in
!     double precision, dimension(:,:), target   :: veldt2_in, vwt_in, wt_in
!     integer,external                           :: omp_get_num_threads
!     !$OMP PARALLEL
!     nth=omp_get_num_threads()
!     !$OMP END PARALLEL
!     n1 = n1_in
!     n2 = n2_in
!     nt2 = nt2_in
!     wt => wt_in
!     allocate(temp(n1,n2,nth))
!     call comb_init(n1_in, n2_in, bc1_in, bc2_in, bc3_in, bc4_in, d1_in, d2_in, veldt2_in, vwt_in, wt_in)
!   end subroutine





!   function step_op(adj, add, model, data) result(stat)
!     logical,intent(in) :: adj, add
!     double precision,dimension(:)  :: model, data
!     integer            :: stat
!     call step_op2(adj, add, model, data,1)
!     stat=0
!   end function




!   subroutine step_op2(adj, add, model, data,ith) ! This operator is the recursive G operator (equation 9)
!     logical,intent(in)         :: adj, add 
!     double precision, dimension(n1,n2,nt2) :: model
!     double precision, dimension(n1,n2,nt2) :: data
!     integer :: ith
!     if(adj) then
!       if(.not. add) model = 0.
!       call comb_op2(adj, .false., temp(:,:,ith), model(:,:,nt2),ith)
!       model(:,:,nt2-1) = model(:,:,nt2-1) + data(:,:,nt2) + temp(:,:,ith)
!       do it2=nt2-2,1,-1
!         call comb_op2(adj, .false., temp(:,:,ith), model(:,:,it2+1),ith)
!         model(:,:,it2) = model(:,:,it2) + data(:,:,it2+1) + temp(:,:,ith) + (wt-1.)*model(:,:,it2+2)
!       end do
!     else
!       if(.not. add) data = 0.
!       call comb_op2(adj, .false., data(:,:,1), temp(:,:,ith),ith)
!       data(:,:,2) = data(:,:,2) + model(:,:,1) + temp(:,:,ith)
!       do it2=3,nt2
!         call comb_op2(adj, .false., data(:,:,it2-1), temp(:,:,ith),ith)
!         data(:,:,it2) = data(:,:,it2) + model(:,:,it2-1) + temp(:,:,ith) + (wt-1.)*data(:,:,it2-2)
!       end do
!     end if
!   end subroutine




! end module
