module wemva_mod

  use sep
  use mem_mod
  use tomo_mod
  use sscale_mod
  use lint_mod
  use step_mod
  use scats0_sso_mod
  use scatr0_mod
  use secder_mod

  implicit none
  integer, private                                     	:: nt, nz, nx, nt2, wfld_j, it, nth, nh, np
  integer, private                                     	:: sou_nx, sou_ox, sou_dx, sou_z, isou, sou_x
  integer, private                                     	:: rec_nx, rec_ox, rec_dx, rec_z, irec, rec_x, edge_par_shot, edge_par_rec
  integer, private                                     	:: bc1, bc2, bc3, bc4
  real, private                                        	:: dt, d1, o1, d2, o2, dtw
  real, dimension(:,:), allocatable, target, private   	:: veldt2, vwt, wt !,vel, refl, !Not present in main: veldt2, vwt, wt. ! Not sure refl is necessary
  real, dimension(:,:,:,:),allocatable, target, private :: temp1,temp2,temp1i,temp2i
  real, dimension(:,:,:), allocatable, target, private 	:: datab, modelb, wflds0
  real, dimension(:,:,:), pointer, private 			    :: src_wave, scat_src_wave
  real, dimension(:,:), pointer, private 			    :: coor_src, coor_rec

!--------------------------------------------------------------------------------------
  real, dimension(:), allocatable		      	        :: wavelet ! This is a copy of what is in main. Could use a pointer instead if you want.

!--------------------------------------------------------------------------------------
  contains



	subroutine initWH1(nz_in, o1_in, d1_in, nx_in, o2_in, d2_in, bc1_in, bc2_in, bc3_in, bc4_in) ! Copying variables 1
		integer			:: nz_in, nx_in, bc1_in, bc2_in, bc3_in, bc4_in
		real			:: d1_in, o1_in, d2_in, o2_in
		nz = nz_in
		nx = nx_in
		d1 = d1_in
		o1 = o1_in
		d2 = d2_in
		o2 = o2_in
		bc1 = bc1_in
		bc2 = bc2_in
		bc3 = bc3_in
		bc4 = bc4_in
	end subroutine



	subroutine initWH2(sou_nx_in, sou_ox_in, sou_dx_in, sou_z_in, rec_nx_in, rec_ox_in, rec_dx_in, rec_z_in, edge_par_shot_in, edge_par_rec_in) ! Copying variables 2
		integer                             :: sou_nx_in, sou_ox_in, sou_dx_in, sou_z_in
		integer                             :: rec_nx_in, rec_ox_in, rec_dx_in, rec_z_in, edge_par_shot_in, edge_par_rec_in
		sou_nx = sou_nx_in
		sou_ox = sou_ox_in
		sou_dx = sou_dx_in
		sou_z = sou_z_in
		rec_nx = rec_nx_in
		rec_ox = rec_ox_in
		rec_dx = rec_dx_in
		rec_z = rec_z_in
		edge_par_shot = edge_par_shot_in
		edge_par_rec = edge_par_rec_in
	end subroutine



	subroutine initWH3(nt_in, dt_in, nh_in, wfld_j_in) ! Copying variables 3
		integer				:: nt_in, wfld_j_in, nh_in
		real				:: dt_in
		nt = nt_in
		nh = nh_in
		wfld_j = wfld_j_in
		dt = dt_in
	end subroutine



	subroutine init_WHalloc(wavelet_in) ! Allocations
		real, dimension(:)				:: wavelet_in 
		integer, external				:: omp_get_num_threads

		!$OMP PARALLEL
		nth=omp_get_num_threads()
		!$OMP END PARALLEL

		write(0,*) " ALLOCATED nth= ", nth
		write(0,*) " ALLOCATED nh= ", nh

		nt2 = (nt-1)*wfld_j+1
		dtw = dt / wfld_j

		call mem_alloc2d(veldt2, nz, nx)
		call mem_alloc2d(vwt, nz, nx)
		call mem_alloc2d(wt, nz, nx)
		call mem_alloc3d(wflds0, np, nt,nth)
		call mem_alloc4d(temp1, nz, nx, nt,nth)
		call mem_alloc4d(temp2, nz, nx, nt,nth)
		call mem_alloc4d(temp1i, nz, nx, nt2,nth)
		call mem_alloc4d(temp2i, nz, nx, nt2,nth)
		call mem_alloc3d(datab, nt, rec_nx, nth)
		call mem_alloc3d(modelb, nz, nx, nth)
		call mem_alloc1d(wavelet, nt)

		wavelet = wavelet_in 

		call sscale_init(nz, nx, nt, dtw, veldt2, wt)
		call lint_init(nz, nx, nt, wfld_j)
		call step_init(nz, nx, nt2, bc1, bc2, bc3, bc4, d1, d2, veldt2, vwt, wt)
		call scats0_init(nz, nx, nt, nh, wflds0, coor_src, np)
		call secder_init(nz, nx, nt, dt)

	end subroutine




	subroutine model_updateWH(vel_total) ! Copy velocity model and prepare boundary condition stuff
		real, dimension(:,:)		:: vel_total
		vwt = 0.
		wt = 0.
		call AbcWt(vel_total, vwt, wt, dtw, d1, d2, bc1, bc2, bc3, bc4)
		veldt2 = vel_total*vel_total*dtw*dtw
	end subroutine





	subroutine init_saved_wavefields(src_wave_in, scat_src_wave_in) ! Reading in the source wavefield
		real, dimension(:,:,:), target 				:: src_wave_in, scat_src_wave_in
		src_wave => src_wave_in
		scat_src_wave => scat_src_wave_in
	end subroutine




	subroutine init_wave_coor(coor_src_in, coor_rec_in, np_in) ! Reading in the source wavefield
		real, dimension(:,:), target 				:: coor_src_in, coor_rec_in
		integer                                     :: np_in
		coor_rec => coor_rec_in
		coor_src => coor_src_in
		np=np_in
	end subroutine






	subroutine wemva1(adj, add, image_in, image_out, rec_wave, isou) ! RECEIVER SIDE
		logical, intent(in)						:: adj, add
		real, dimension(:, :)					:: image_in, image_out
		real, dimension(np, nt,1)			    :: rec_wave
		integer									:: i, ith, isou, ih

		ith=1
		modelb=0.0
		temp1=0.0
		image_out=0.0

		! Imaging condition
		if(adj) then

			! Start with perturbed model
		    modelb(:,:,ith) = image_in

			! Make the scattered source wavefield
			call scats0_init(nz, nx, nt, nh, src_wave, coor_src, np)  ! Initialize the reciever wavefield
		    call scats0_op2(.false., .false., modelb(:,:,ith), temp1(:,:,:,ith),ith) ! Scatter the source wavefield
	        call secder_op2(.false., .false., temp1(:,:,:,ith), temp2(:,:,:,ith)) ! T operator
	        call sscale_op2(.false., .false., temp2(:,:,:,ith), temp1(:,:,:,ith)) ! S operator
	        call lint_op2(.false., .false., temp1(:,:,:,ith), temp1i(:,:,:,ith)) ! L operator
	        call step_op2(.false., .false., temp1i(:,:,:,ith), temp2i(:,:,:,ith),ith) ! G operator
	        call lint_op2(.true., .false., temp1(:,:,:,ith), temp2i(:,:,:,ith)) ! L' operator

			modelb=0.0
			! Cross-correlate the scattered source wavefield and reciever wavefield
			call scats0_init(nz, nx, nt, nh, rec_wave, coor_rec, np)  ! Initialize the source wavefield
		    call scats0_op2(.true., .false., modelb(:,:,ith), temp1(:,:,:,ith),ith) ! Scatter the reciever wavefield

			! Stack over all threads
			do i=1,nth
				image_out=image_out + modelb(:,:,i)
			end do



		!   else
		 !    call scats0_op2(.false., .true., modelb(:,:,:,ith), temp1(:,:,:,ith),ith) 
			! ! Stack over all threads
			! do i=1,nth
			! 	data=data+datab(:,:,i)
			! end do
		end if

	end subroutine








	subroutine wemva2(adj, add, image_in, image_out, rec_wave, isou) ! SOURCE SIDE
	    logical, intent(in)				:: adj, add
	    real, dimension(:,:)			:: image_in, image_out
	    real, dimension(np, nt,1)	    :: rec_wave
	    integer							:: i, ith, isou, ih

		ith=1
		modelb=0.0
		temp1=0.0
		image_out=0.0

	    if(adj) then ! Adjoint: First order scattered data => perturbation

			! Start with perturbed model
		    modelb(:,:,ith) = image_in

			! Make the scattered reciever wavefield
			call scats0_init(nz, nx, nt, nh, rec_wave, coor_rec, np)  ! Initialize the reciever wavefield
	        call scats0_op2_REV(.false., .false., modelb(:,:,ith), temp1(:,:,:,ith),ith) ! Scatter the reciever wavefield
	        call secder_op2(.false., .false., temp1(:,:,:,ith), temp2(:,:,:,ith)) ! T operator
	        call sscale_op2(.false., .false., temp2(:,:,:,ith), temp1(:,:,:,ith)) ! S operator
	        call lint_op2(.false., .false., temp1(:,:,:,ith), temp1i(:,:,:,ith)) ! L operator
	        call step_op2(.false., .false., temp1i(:,:,:,ith), temp2i(:,:,:,ith),ith) ! G operator
	        call lint_op2(.true., .false., temp1(:,:,:,ith), temp2i(:,:,:,ith)) ! L' operator

			! Cross-correlate the scattered reciever wavefield and source wavefield
			modelb=0.0
			call scats0_init(nz, nx, nt, nh, src_wave, coor_src, np)  ! Initialize the source wavefield
	        call scats0_op2_REV(.true., .false., modelb(:,:,ith), temp1(:,:,:,ith),ith) ! Scatter the reciever wavefield

			! Stack over all threads
			do i=1,nth
				image_out=image_out + modelb(:,:,i)
			end do


	    else ! Forward: Perturbation => Scattered data
		 !    modelb(:,:,:,ith) = model

			! call scats0_init(nz, nx, nt, nh, rec_wave)  ! Initialize the reciever wavefield
	  !       call scats0_op2(.true., .true., modelb(:,:,:,ith), temp1(:,:,:,ith),ith) ! Scatter the reciever wavefield

			! call scats0_init(nz, nx, nt, nh, src_wave)  ! Initialize the source wavefield
	  !       call scats0_op2(.false., .true., modelb(:,:,:,ith), temp1(:,:,:,ith),ith) ! Scatter the reciever wavefield

			! do i=1,nth
			! 	model=model + modelb(:,:,:,i)
			! end do

	    end if


	end subroutine












end module 
