#!/usr/bin/python
import copy
import random
import os
import sepbase
import tempfile
import time
import unittest
import math
import subprocess

from batch_task_executor import *
import pbs_util

### An example to define a batch task through inheriting the BatchTaskComposer class.
### IMPORTANT: Read /home/zyang03/script/batch_task_executor.py to learn the detailed specifications/explanations of the interfaces defined in BatchTaskComposer.

# Not everything is best placed in parameter files, since the python program cannot easily read them. Might be better to 
# have num_shots, etc set up in the makefile rule that calls the python script. 






#=======================================================================================================
#=======================================================================================================
#=======================================================================================================


class GN_HESS_OP(BatchTaskComposer):
  def __init__(self, param_reader):
    BatchTaskComposer.__init__(self)
 
    #Read in the job parameters
    self.dict_args = param_reader.dict_args
    self.T_BIN = self.dict_args['T_BIN']
    self.wavelet = self.dict_args['wavelet']
    self.nshots = int(self.dict_args['nfiles'])
    self.nsh_perjob = int(self.dict_args['nfiles_perjob'])
    self.ish_beg = int(self.dict_args['ish_beg'])
    self.path_out = param_reader.path_out
    self.jobprefix = "GN_Hessian"
    self.levelset = self.dict_args['levelset']
    self.acq = self.dict_args['acq']
    self.model_geom = self.dict_args['model_geom']
    self.path_keep = self.dict_args['path_keep']
    self.vel_mod_path = self.dict_args['vel_mod_path']
    self.hess_in = self.dict_args['hess_in']

    # Create all subjobs info
    self.njobs = int(math.ceil(self.nshots/float(self.nsh_perjob)))
    self.subjobids = range(0, self.njobs)
    self.subjobfns_list = [None]*self.njobs
    self.subjobIND_list = range(self.ish_beg, self.nshots+1, self.nsh_perjob) # List of actual shot number (starting @ 0)

    # Make a list of all the files for each job ( shots per job)
    for i in range(0,self.njobs):
      ish_start = self.subjobIND_list[i]
      nsh = min(self.nsh_perjob, self.nshots-ish_start)
      self.subjobfns_list[i] = [None]*nsh
      for j in range(0,nsh):
        self.subjobfns_list[i][j] = '%s/%s-%s.H' % (self.path_out,self.jobprefix,ish_start+j+1)
    return


  def GetSubjobScripts(self, subjobid):
    #'''Return the scripts content for that subjob.'''
    subjobfns = self.subjobfns_list[subjobid]
    # Build scripts that create the files designated in subjobfns
    scripts = []
    cnt = 0
    nf_act = len(subjobfns) # number of files

    for fn in subjobfns:
      shot = self.nsh_perjob*subjobid+cnt+1
      cnt += 1
      if cnt > nf_act: break
      operator = "%s/%s" % (self.T_BIN,'GNHESSIAN.x')
      inputs1 = 'par=%s par=%s par=%s isou=%s nh=0 adjoint=0 ' % (self.acq,self.model_geom,self.levelset,shot)
      inputs2 = 'waveletin=%s velmod=%s insrc=%s/src_wavefield%s.H  coor=%s/coor_src%s.H' % (self.wavelet,self.vel_mod_path,self.path_out,shot,self.path_out,shot)
      outputs = 'imageA=%s  imageB=%s ' % (self.hess_in,fn)
      cmd = '%s %s %s %s \n' % (operator, inputs1, inputs2, outputs) # Note: need the \n at the end... also cant use < in string itself.
      scripts.append(cmd)
    return scripts, str(subjobid)


  def GetSubjobsInfo(self):
    '''Should return two lists, subjobids and subjobfns.'''
    return copy.deepcopy(self.subjobids), copy.deepcopy(self.subjobfns_list)






class make_GN_hessian2(object):
  def __init__(self, param_reader):

    #-------------------------------------------------------------------------------------
    dict_args = param_reader.dict_args
    self.hess_out = dict_args['hess_out']
    fn_seph_list = ' '
    
    #----- GN HESSIAN APPLIED TO RTM GRADIENT --------------------------------------------
    job5 = GN_HESS_OP(param_reader)
    bte = BatchTaskExecutor(param_reader)
    prefix = param_reader.prefix
    print "--------- Submitting/Running GN HESSIAN FORWARD ---------------------------"
    bte.LaunchBatchTask(prefix, job5)

    #------ Combine the files to one single file -----------------------------------------
    _, fns_list = job5.GetSubjobsInfo()
    fn_seph_list = [fn for fns in fns_list for fn in fns]
    args=''
    for i in range(0,len(fn_seph_list)):
      args = '%s %s' % (args,fn_seph_list[i])
    cmd = 'Cat %s | Transp plane=23 | Stack > %s out=%s@ ' % (args,self.hess_out,self.hess_out)
    print "--------- Stacking GN HESSIAN files  ------------------------------"
    print(cmd)
    subprocess.call(cmd,shell=True,stdout=f)

    #----- Removing ADJOINT and FORWARD BORN output files from wrk directory -------------
    ars_hess=''
    for i in range(0,len(fn_seph_list)):
      ars_hess = '%s %s' % (ars_hess,fn_seph_list[i])
    cmd = 'rm -rf %s' % (ars_hess)
    subprocess.call(cmd,shell=True,stdout=f)
    print "--------- Removing ADJOINT and FORWARD BORN output files from wrk directory  ------------------------------"
    print(cmd)
    
    return






if __name__ == '__main__':
  print "Run BatchTaskExecutor_test with params:", sys.argv
  # Check basic environment variable setup.
  assert 'SEP' in os.environ, '!Environment var SEP is not set yet, should have your seplib enviroment set up first! Check paths like /opt/SEP/'
  assert 'RSF' in os.environ, '!Environment var RSF is not set yet, should have your RSF environment set up first! Check paths like /opt/RSF'
  eq_args_from_cmdline,args = sepbase.parse_args(sys.argv)
  param_reader = pbs_util.JobParamReader(eq_args_from_cmdline)
  f = open("hessian_screen_out.txt",'a+')

  #------- APPLY GAUSS NEWTON HESSIAN TO GRADIENT ----------------
  start_time = time.time()
  make_GN_hessian2(param_reader)
  # make_GN_hessian(param_reader)
  elapsed_time = (time.time() - start_time)/60.0
  print("ELAPSED TIME FOR BTE Hessian (min) : ")
  print(elapsed_time)









# class BORN(BatchTaskComposer):
#   def __init__(self, param_reader, vel_mod_path, jobprefix, adjointin, datapath, modelpath):
#     BatchTaskComposer.__init__(self)
 
#     #Read in the job parameters
#     self.dict_args = param_reader.dict_args
#     self.T_BIN = self.dict_args['T_BIN']
#     self.wavelet = self.dict_args['wavelet']
#     self.nshots = int(self.dict_args['nfiles'])
#     self.nsh_perjob = int(self.dict_args['nfiles_perjob'])
#     self.ish_beg = int(self.dict_args['ish_beg'])
#     self.path_out = param_reader.path_out
#     self.jobprefix = jobprefix
#     # self.residual = self.dict_args['residual']
#     self.levelset = self.dict_args['levelset']
#     self.acq = self.dict_args['acq']
#     self.model_geom = self.dict_args['model_geom']
#     self.path_keep = self.dict_args['path_keep']
#     # self.res_name = self.dict_args['res_name']
#     self.vel_mod_path = vel_mod_path
#     self.adjoint = int(adjointin)
#     self.datapath = datapath
#     self.modelpath = modelpath


#     # Create all subjobs info
#     self.njobs = int(math.ceil(self.nshots/float(self.nsh_perjob)))
#     self.subjobids = range(0, self.njobs)
#     self.subjobfns_list = [None]*self.njobs
#     self.subjobIND_list = range(self.ish_beg, self.nshots+1, self.nsh_perjob) # List of actual shot number (starting @ 0)

#     # Make a list of all the files for each job ( shots per job)
#     for i in range(0,self.njobs):
#       ish_start = self.subjobIND_list[i]
#       nsh = min(self.nsh_perjob, self.nshots-ish_start)
#       self.subjobfns_list[i] = [None]*nsh
#       for j in range(0,nsh):
#         self.subjobfns_list[i][j] = '%s/%s-%s.H' % (self.path_out,self.jobprefix,ish_start+j+1)
#     return


#   def GetSubjobScripts(self, subjobid):
#     #'''Return the scripts content for that subjob.'''
#     subjobfns = self.subjobfns_list[subjobid]
#     # Build scripts that create the files designated in subjobfns
#     scripts = []
#     cnt = 0
#     nf_act = len(subjobfns) # number of files

#     for fn in subjobfns:
#       # shot_pos = self.fs + self.ds*(cnt+self.subjobIND_list[subjobid])
#       shot = self.nsh_perjob*subjobid+cnt+1
#       cnt += 1
#       # if shot > (self.nshots): break
#       if cnt > nf_act: break
#       operator = "%s/%s" % (self.T_BIN,'BORN.x')
#       inputs = 'par=%s par=%s par=%s isou=%s nh=0 adjoint=%s waveletin=%s velmod=%s ' % (self.acq,self.model_geom,self.levelset,shot,self.adjoint,self.wavelet,self.vel_mod_path)
#       # inputs = 'par=%s par=%s par=%s isou=%s nh=0 adjoint=%s waveletin=%s velmod=%s insrc=%s/src_wavefield%s.H' % (self.acq,self.model_geom,self.levelset,shot,self.adjoint,self.wavelet,self.vel_mod_path,self.path_out,shot)
#       # inputs = 'par=%s par=%s par=%s isou=%s nh=0  nt=%d  adjoint=%s waveletin=%s velmod=%s insrc=%s/src_wavefield%s.H' % (self.acq,self.model_geom,self.levelset,shot,self.nt,self.adjoint,self.wavelet,self.vel_mod_path,self.path_out,shot)

#       if self.adjoint==1:
#           outputs = 'data=%s  model=%s' % (self.datapath, fn)
#       else:
#           outputs = 'data=%s  model=%s scat_wave=%s/scat_wave%s.H' % (fn, self.modelpath,self.path_out,shot)
#       cmd = '%s %s %s \n' % (operator, inputs, outputs) # Note: need the \n at the end... also cant use < in string itself.
#       scripts.append(cmd)
#     return scripts, str(subjobid)


#   def GetSubjobsInfo(self):
#     '''Should return two lists, subjobids and subjobfns.'''
#     return copy.deepcopy(self.subjobids), copy.deepcopy(self.subjobfns_list)





# class make_GN_hessian(object):
#   def __init__(self, param_reader):

#     #-------------------------------------------------------------------------------------
#     dict_args = param_reader.dict_args
#     T_BIN = dict_args['T_BIN']
#     self.path_keep = dict_args['path_keep']
#     self.vel_mod_path = dict_args['vel_mod_path']
#     self.datapath = dict_args['datapath']
#     self.hess_in = dict_args['hess_in']
#     self.hess_out = dict_args['hess_out']
#     self.path_out = dict_args['path_out']
#     fn_seph_list = ' '
    

#     #----- FORWARD BORN APPLIED TO RTM GRADIENT -------------------
#     modelpath='%s' % (self.hess_in)
#     job5 = BORN(param_reader, self.vel_mod_path, 'forwardborn', 0.0, self.datapath, modelpath)
#     bte = BatchTaskExecutor(param_reader)
#     prefix = param_reader.prefix
#     print "--------- Submitting/Running BORN FORWARD ---------------------------"
#     bte.LaunchBatchTask(prefix, job5)
#     # Combine the files to one single file.
#     _, fns_list = job5.GetSubjobsInfo()
#     fn_seph_list = [fn for fns in fns_list for fn in fns]
#     args=''
#     for i in range(0,len(fn_seph_list)):
#       args = '%s %s' % (args,fn_seph_list[i])
#     cmd = 'Cat %s > %s' % (args,self.datapath)
#     subprocess.call(cmd,shell=True,stdout=f)
#     print "--------- Stacking FORWARD BORN files  ------------------------------"
#     print(cmd)


#     #----- Make list of all Born forward output files -------------------
#     args_for=''
#     for i in range(0,len(fn_seph_list)):
#       args_for = '%s %s' % (args_for,fn_seph_list[i])


#     #----- ADJOINT BORN APPLIED TO FORWARD BORN DATA -------------------
#     modelpath='dummy_string'
#     job5 = BORN(param_reader, self.vel_mod_path, 'adjointborn', 1.0, self.datapath, modelpath)
#     bte = BatchTaskExecutor(param_reader)
#     prefix = param_reader.prefix
#     print "--------- Submitting/Running BORN ADJOINT ------------------------------"
#     bte.LaunchBatchTask(prefix, job5)
    
#     # Combine the files to one single file.
#     _, fns_list = job5.GetSubjobsInfo()
#     fn_seph_list = [fn for fns in fns_list for fn in fns]
#     args=''
# # modelpath='%s/GNapplied.H' % (self.path_keep)
#     for i in range(0,len(fn_seph_list)):
#       args = '%s %s' % (args,fn_seph_list[i])
#     cmd = 'Cat %s | Transp plane=23 | Stack > %s out=%s@ ' % (args,self.hess_out,self.hess_out)
#     print "--------- Stacking ADJOINT BORN files  ------------------------------"
#     print(cmd)
#     subprocess.call(cmd,shell=True,stdout=f)


#     #----- Removing ADJOINT and FORWARD BORN output files from wrk directory -------------------
#     args_adj=''
#     for i in range(0,len(fn_seph_list)):
#       args_adj = '%s %s' % (args_adj,fn_seph_list[i])
#     cmd = 'rm -rf %s %s' % (args_for,args_adj)
#     subprocess.call(cmd,shell=True,stdout=f)
#     print "--------- Removing ADJOINT and FORWARD BORN output files from wrk directory  ------------------------------"
#     print(cmd)
    
#     return





