#!/usr/local/anaconda/bin/python


### !/usr/bin/python
###/usr/local/anaconda/bin/python ( for cees-cluster, since regular python has broken numpy)
import copy
import random
import os
import sepbase
import tempfile
import time
import unittest
import math
import subprocess
import numpy as np
from numpy.linalg import inv

from batch_task_executor import *
import pbs_util

### An example to define a batch task through inheriting the BatchTaskComposer class.
### IMPORTANT: Read /home/zyang03/script/batch_task_executor.py to learn the detailed specifications/explanations of the interfaces defined in BatchTaskComposer.



# Not everything is best placed in parameter files, since the python program cannot easily read them. Might be better to 
# have num_shots, etc set up in the makefile rule that calls the python script. 



















class forward_modeling(BatchTaskComposer):
    def __init__(self, param_reader, vel_mod_path,jobprefix):
        BatchTaskComposer.__init__(self)

        #Read in the job parameters
        self.dict_args = param_reader.dict_args
        self.T_BIN = self.dict_args['T_BIN']
        self.wavelet = self.dict_args['wavelet']
        self.vel_mod_path = vel_mod_path
        self.nshots = int(self.dict_args['nfiles'])
        self.nsh_perjob = int(self.dict_args['nfiles_perjob'])
        self.ish_beg = int(self.dict_args['ish_beg'])
        self.path_out = param_reader.path_out
        self.jobprefix = jobprefix
        self.acq = self.dict_args['acq']
        self.model_geom = self.dict_args['model_geom']
        self.path_keep = self.dict_args['path_keep']
        self.nt = int(self.dict_args['nt'])

        # Create all subjobs info
        self.njobs = int(math.ceil(self.nshots/float(self.nsh_perjob)))
        self.subjobids = range(0, self.njobs)
        self.subjobfns_list = [None]*self.njobs
        self.subjobIND_list = range(self.ish_beg, self.nshots+1, self.nsh_perjob) # List of actual shot number (starting @ 0)

        # Make a list of all the files for each job ( shots per job)
        for i in range(0,self.njobs):
            ish_start = self.subjobIND_list[i]
            nsh = min(self.nsh_perjob, self.nshots-ish_start)
            self.subjobfns_list[i] = [None]*nsh
            for j in range(0,nsh):
                self.subjobfns_list[i][j] = '%s/%s-%s.H' % (self.path_out,self.jobprefix,ish_start+j+1)
        return


    def GetSubjobScripts(self, subjobid):
        #'Return the scripts content for that subjob.'
        subjobfns = self.subjobfns_list[subjobid]
        # Build scripts that create the files designated in subjobfns
        scripts = []
        cnt = 0
        nf_act = len(subjobfns) # number of files
        for fn in subjobfns:
            shot = self.nsh_perjob*subjobid+cnt+1
            cnt += 1
            if cnt > nf_act: break
            operator = "%s/%s" % (self.T_BIN,'FORWARD_SYN_REC.x')
            inputs = 'par=%s par=%s ns=1 nh=1 nt=%d isou=%d waveletin=%s velmodel=%s' % (self.acq,self.model_geom,self.nt,shot,self.wavelet,self.vel_mod_path)
            # outputs = 'forward_data=%s src_wavefield=%s/src_wavefield%s.H' % (fn,self.path_out,shot)
            outputs = 'forward_data=%s src_wavefield=%s/src_wavefield%s_full.H' % (fn,self.path_out,shot)
            compression = '%s/%s wavefield_in=%s/src_wavefield%s_full.H  mask=%s/s1.H  comp_wave=%s/src_wavefield%s.H   coor=%s/coor_src%s.H' % (self.T_BIN,'WAVEFIELD_COMP.x',self.path_out,shot,self.path_keep,self.path_out,shot,self.path_out,shot)
            # cmd = '%s %s %s \n' % (operator, inputs, outputs)
            cmd = '%s %s %s ; %s \n' % (operator, inputs, outputs, compression) 
            # Note: need the \n at the end... also cant use < in string itself.
            scripts.append(cmd)
            # print(cmd)
        return scripts, str(subjobid)


    def GetSubjobsInfo(self):
        '''Should return two lists, subjobids and subjobfns.'''
        return copy.deepcopy(self.subjobids), copy.deepcopy(self.subjobfns_list)







class BORN(BatchTaskComposer):
    def __init__(self, param_reader, vel_mod_path, jobprefix, adjointin, datapath, modelpath):
        BatchTaskComposer.__init__(self)
     
        #Read in the job parameters
        self.dict_args = param_reader.dict_args
        self.T_BIN = self.dict_args['T_BIN']
        self.wavelet = self.dict_args['wavelet']
        self.nshots = int(self.dict_args['nfiles'])
        self.nsh_perjob = int(self.dict_args['nfiles_perjob'])
        self.ish_beg = int(self.dict_args['ish_beg'])
        self.path_out = param_reader.path_out
        self.jobprefix = jobprefix
        # self.residual = self.dict_args['residual']
        self.levelset = self.dict_args['levelset']
        self.acq = self.dict_args['acq']
        self.model_geom = self.dict_args['model_geom']
        self.path_keep = self.dict_args['path_keep']
        # self.res_name = self.dict_args['res_name']
        self.vel_mod_path = vel_mod_path
        self.adjoint = int(adjointin)
        self.datapath = datapath
        self.modelpath = modelpath
        self.nt = int(self.dict_args['nt'])

        # Create all subjobs info
        self.njobs = int(math.ceil(self.nshots/float(self.nsh_perjob)))
        self.subjobids = range(0, self.njobs)
        self.subjobfns_list = [None]*self.njobs
        self.subjobIND_list = range(self.ish_beg, self.nshots+1, self.nsh_perjob) # List of actual shot number (starting @ 0)

        # Make a list of all the files for each job ( shots per job)
        for i in range(0,self.njobs):
            ish_start = self.subjobIND_list[i]
            nsh = min(self.nsh_perjob, self.nshots-ish_start)
            self.subjobfns_list[i] = [None]*nsh
            for j in range(0,nsh):
                self.subjobfns_list[i][j] = '%s/%s-%s.H' % (self.path_out,self.jobprefix,ish_start+j+1)
        return


    def GetSubjobScripts(self, subjobid):
        #'''Return the scripts content for that subjob.'''
        subjobfns = self.subjobfns_list[subjobid]
        # Build scripts that create the files designated in subjobfns
        scripts = []
        cnt = 0
        nf_act = len(subjobfns) # number of files

        for fn in subjobfns:
            # shot_pos = self.fs + self.ds*(cnt+self.subjobIND_list[subjobid])
            shot = self.nsh_perjob*subjobid+cnt+1
            cnt += 1
            # if shot > (self.nshots): break
            if cnt > nf_act: break
            operator = "%s/%s" % (self.T_BIN,'BORN.x')
            inputs = 'par=%s par=%s par=%s isou=%s nh=0  nt=%d  adjoint=%s waveletin=%s velmod=%s insrc=%s/src_wavefield%s.H  coor_src=%s/coor_src%s.H ' % (self.acq,self.model_geom,self.levelset,shot,self.nt,self.adjoint,self.wavelet,self.vel_mod_path,self.path_out,shot,self.path_out,shot)
            # inputs = 'par=%s par=%s par=%s isou=%s nh=0  nt=%d  adjoint=%s waveletin=%s velmod=%s                           ' % (self.acq,self.model_geom,self.levelset,shot,self.nt,self.adjoint,self.wavelet,self.vel_mod_path)
            if self.adjoint==1:
                outputs = 'data=%s  model=%s   rec_wave=%s/rec_wave%s_full.H ' % (self.datapath, fn, self.path_out,shot)
                # outputs = 'forward_data=%s src_wavefield=%s/src_wavefield%s_full.H' % (fn,self.path_out,shot)
                compression = '%s/%s wavefield_in=%s/rec_wave%s_full.H  mask=%s/s1.H  comp_wave=%s/rec_wave%s.H   coor=%s/coor_rec%s.H' % (self.T_BIN,'WAVEFIELD_COMP.x',self.path_out,shot,self.path_keep,self.path_out,shot,self.path_out,shot)
            else:
                outputs = 'data=%s  model=%s' % (fn, self.modelpath)
            cmd = '%s %s %s; %s \n' % (operator, inputs, outputs, compression) # Note: need the \n at the end... also cant use < in string itself.
            scripts.append(cmd)
        return scripts, str(subjobid)


    def GetSubjobsInfo(self):
        '''Should return two lists, subjobids and subjobfns.'''
        return copy.deepcopy(self.subjobids), copy.deepcopy(self.subjobfns_list)






class make_gradient(object):
    def __init__(self, param_reader):
        #----------------------------------------------------------------
        dict_args = param_reader.dict_args
        self.res_name = dict_args['res_name']
        self.cleandata_name = dict_args['cleandata_name']
        self.syndata_name = dict_args['syndata_name']
        self.rtm_name = dict_args['rtm_name']
        self.T_BIN = dict_args['T_BIN']
        self.path_keep = dict_args['path_keep']
        self.path_out = dict_args['path_out']
        self.vel_mod_path = dict_args['vel_mod_path']
  
        MV_flags_in = dict_args['MODVEL_flags_in']
        MODVEL_flags_in = MV_flags_in.split()
        self.prefix = param_reader.prefix

        self.PBS_input_path = dict_args['PBS_input_path']
        self.PBS_inputs= dict_args['PBS_inputs']
        self.hess_iter= dict_args['hess_iter']
        self.hess_name= dict_args['hess_name']
        self.hess_in_path= dict_args['hess_in_path']
        self.hess_out_path= dict_args['hess_out_path']
        self.threshold= dict_args['hess_thresh']

        # nt = dict_args['nt']
        # model_geom = dict_args['model_geom']
        # acq = dict_args['acq']
        # levelset = dict_args['levelset']
        return



    def syn_data(self, param_reader,debug,prefixA):
        #----- MAKE SYNTHENTIC DATA -------------------
        syndata = forward_modeling(param_reader,self.vel_mod_path,prefixA)
        bte = BatchTaskExecutor(param_reader)
        if (debug):
            print('-----------------------------------------------------------------')
            print "MAKING SYNTHENTIC DATA "
        param_reader.prefix=prefixA
        bte.LaunchBatchTask(prefixA, syndata)
        # Combine the files to one single file.
        _, fns_list = syndata.GetSubjobsInfo()
        fn_seph_list = [fn for fns in fns_list for fn in fns]
        args=''
        for i in range(0,len(fn_seph_list)):
          args = '%s %s' % (args,fn_seph_list[i])
        cmd = 'Cat %s > %s/%s.H' % (args,self.path_keep,prefixA)
        if (debug):
            print('-----------------------------------------------------------------')
            print " CONCATENATING SYNTHETIC FORWARD DATA FILES"
            print(cmd)
        subprocess.call(cmd,shell=True)
        return



    def residual(self, param_reader,debug,prefixA):
        # #----- MAKE RESIDUAL  ------------------------------------------
        cmd = "Math file1=%s/%s.H file2=%s/%s.H exp='(file1-file2)' > %s/%s.H " % (self.path_keep,prefixA,self.path_keep,self.cleandata_name,self.path_keep,self.res_name) # Need residual to mave meters in d2 and o2 for MUTE_DA to work properly
        if (debug):
            print('-----------------------------------------------------------------')
            print " COMPUTING RESIDUAL "
            print(cmd)
        subprocess.call(cmd,shell=True)
        return



    def residual_real(self, param_reader,debug,prefixA):
        # #----- MAKE RESIDUAL  (and MUTE THE DIRECT ARRIVAL)  -------------------
        cmd1 = ' %s/SRC_CONVOLVE.x wavelet=%s/waveletC.H < %s/%s.H > synconv.h ' % (self.T_BIN,self.path_keep,self.path_keep,syndata_name)
        cmd2 = ' %s/SRC_CONVOLVE.x wavelet=%s/waveletC.H < %s/%s.H > obsconv.h ' % (self.T_BIN,self.path_keep,self.path_keep,cleandata_name)

        cmd3 = ' %s/RES_MUTE.x v1=1900.0  t1=2.25 < synconv.h weights=weights.H radius=25  > syn1.h ' % (self.T_BIN)
        cmd4 = ' %s/RES_MUTE.x v1=1900.0  t1=2.25 < obsconv.h weights=weights.H radius=25  > obs1.h' % (self.T_BIN)

        cmd5 = "Math file1=syn1.h exp='@ABS(file1)' | Scale > tmp.h"
        cmd6 = 'Clip pclip=90 dclip=1 retry=y chop=clip to=1.0 < tmp.h > tmp2.h'
        cmd7 = 'Smooth rect1=4 rect2=4 repeat=4 < tmp2.h | Scale > eventselector.h'
        cmd8 = "Math file1=obs1.h file2=eventselector.h exp='file1*file2' > obs2.h"

        cmd9 = "Scale axis=1 < syn1.h > syn1 ; Scale axis=1 < obs2.h > obs2 " 
        cmd10 = "Math file1=syn1 file2=obs2 exp='(file1-file2)' > %s/res_dead.H " % (self.path_keep) # Need residual to mave meters in d2 and o2 for MUTE_DA to work properly
        cmd11 = "%s/DEAD_TRACE_RMV.x < %s/res_dead.H >  %s/%s.H " % (self.T_BIN,self.path_keep,self.path_keep,self.res_name)

        cmd  = '%s; %s; %s; %s; %s; %s; %s; %s; %s; %s; %s; ' % (cmd1, cmd2, cmd3, cmd4, cmd5, cmd6, cmd7, cmd8, cmd9, cmd10, cmd11)
        subprocess.call(cmd,shell=True)
        print(cmd)
        print "--------- Computed residual ------------------------------"

    #===================================================================================
        if (debug):
            print('-----------------------------------------------------------------')
            print " COMPUTING RESIDUAL "
            print(cmd)
        subprocess.call(cmd,shell=True)
        return




    def born(self, param_reader,debug):
        #----- ADJOINT BORN APPLIED TO RESIDUAL DATA -------------------
        modelpath='dummy_string'
        self.datapath = "%s/%s.H" % (self.path_keep,self.res_name)
        self.stacked_RTM = "%s/%s.H" % (self.path_keep,self.rtm_name)
        job5 = BORN(param_reader, self.vel_mod_path, 'primary_gradient', 1.0, self.datapath, modelpath)
        bte = BatchTaskExecutor(param_reader)
        if (debug):
            print('-----------------------------------------------------------------')
            print " RUNNING BORN ADJOINT"
        prefixA="BORN_OPERATOR"
        param_reader.prefix=prefixA
        bte.LaunchBatchTask(prefixA, job5)
        # Combine the files to one single file.
        _, fns_list = job5.GetSubjobsInfo()
        fn_seph_list = [fn for fns in fns_list for fn in fns]
        args=''

        for i in range(0,len(fn_seph_list)):
          args = '%s %s' % (args,fn_seph_list[i])
        # stacked_RTM = "%s/%s" % (self.path_keep,self.rtm_name)
        cmd = 'Cat %s | Transp plane=23 | Stack > %s out=%s@ ' % (args,self.stacked_RTM,self.stacked_RTM)
        if (debug):
            print('-----------------------------------------------------------------')
            print " STACKING ADJOINT BORN FILES"
            print(cmd)
        subprocess.call(cmd,shell=True)
        return




    def GN_inversion(self, param_reader,debug):
       # -------------------------------------------------------------------------------------
        # Make the PBS parameter file
        text_file = open(self.PBS_input_path, "w")
        text_file.write("%s" % PBS_inputs)
        text_file.close()

        # MKL routine
        # cmd = '%s/GN_inversionTJ2.x < %s/%s.H  threshold=%s  lberg=0 lambda=0.001 intermediate=intermediate.H  rnorm=rnorm.H niter=%s  PBS_inputs=%s  hess_in_path=%s   hess_out_path=%s >  %s/%s.H' % (T_BIN,path_keep,rtm_name,threshold, hess_iter,PBS_input_path,hess_in_path,hess_out_path,path_keep,hess_name) 
       
        # CG routine
        cmd1 = '%s/GN_inversionCG.x < %s/%s.H  threshold=%s  lberg=0 lambda=0.001  rnorm=rnorm.H ' % (self.T_BIN,self.path_keep,self.rtm_name,self.threshold) 
        cmd2 = ' niter=%s  PBS_inputs=%s  hess_in_path=%s   hess_out_path=%s >  %s/%s.H' % (self.hess_iter,self.PBS_input_path,self.hess_in_path,self.hess_out_path,self.path_keep,self.hess_name) 
        cmd  = ' %s %s' & (cmd1, cmd2)
        print(cmd)
        subprocess.call(cmd,shell=True)

        return







   


class MAIN_FUNCTIONS(object):
    def __init__(self, param_reader):
        dict_args = param_reader.dict_args
        self.path_keep = dict_args['path_keep']
        self.hess_name= dict_args['hess_name']
        self.res_name = dict_args['res_name']

        S1_flags_in = dict_args['SERIAL1_flags_in']
        self.SERIAL1_flags_in = S1_flags_in.split()
        S1_flags_out = dict_args['SERIAL1_flags_out']
        self.SERIAL1_flags_out = S1_flags_out.split()
        self.model_geom = dict_args['model_geom']
        self.acq = dict_args['acq']
        self.T_BIN = dict_args['T_BIN']

        S2_flags_in = dict_args['SERIAL2_flags_in']
        self.SERIAL2_flags_in = S2_flags_in.split()
        S2_flags_out = dict_args['SERIAL2_flags_out']
        self.SERIAL2_flags_out = S2_flags_out.split()
        self.levelset = dict_args['levelset']

        S2_flags_in = dict_args['SERIAL2_flags_in']
        self.SERIAL2_flags_in = S2_flags_in.split()
        S2_flags_out = dict_args['SERIAL2_flags_out']
        self.SERIAL2_flags_out = S2_flags_out.split()

        MV_flags_in = dict_args['MODVEL_flags_in']
        self.MODVEL_flags_in = MV_flags_in.split()

        self.cleandata_name = dict_args['cleandata_name'] 

        self.fulltime_it = int(dict_args['fulltime_it'])
        self.path_out = dict_args['path_out']
        self.phi_grad1 = dict_args['phi_grad1']
        self.vel_mod_path = dict_args['vel_mod_path']

        return


    def INIT_INTERMEDIATE_OUTPUTS(self, nt0, debug):
        #------- INITIALIZE THE INTERMEDIATE RESULTS -----------------------
        cmd1 = 'cp  %s/scaling.H  scaling_comp.H'                  % (self.path_keep)
        cmd2 = 'cp  %s/s1.H  s1_comp.H'                            % (self.path_keep)
        cmd3 = 'cp  %s/drlse_term1.H  drlse_comp.H'                % (self.path_keep)
        cmd4 = 'Pad < %s/residual.H n1out=%s > residual_comp.H'    % (self.path_keep,nt0)
        cmd5 = 'cp  %s/%s.H  GNHessian_comp.H'                     % (self.path_keep,self.hess_name)
        cmd6 = 'cp  %s/RTMgrad.H  RTMgrad_comp2.H'                 % (self.path_keep)
        cmd7 = 'cp  %s/tomo_grad.H  tomograd_comp.H'               % (self.path_keep)
        cmd8 = 'cp  %s/phi_grad1.H  phigrad_comp.H'                % (self.path_keep)
        cmd9 = 'cp  %s/objfunc.H  objfunc_comp.H'                  % (self.path_keep)
        cmd10 = 'cp  rnorm.H  rnorm_comp.H'
        cmd11 = '%s/INITIALIZE.x  grad_in=%s/%s.H    alpha_cg=alpha_comp.H  beta_cg=beta_comp.H   maxbeta=maxbeta_comp.H  par=%s' % (T_BIN, path_keep, phi_grad1, model_geom)
        args = '%s ;%s ;%s ;%s ;%s ;%s ;%s ;%s ; %s; %s;' % (cmd1, cmd2, cmd3, cmd4, cmd5, cmd6, cmd7, cmd8, cmd9, cmd10)
        if (debug):
            print('-----------------------------------------------------------------')
            print('INIT_INTERMEDIATE_OUTPUTS')
            print(args)
        subprocess.call(args,shell=True)
        return




    def SERIAL2(self, debug):
        #----- SERIAL2:   normalize, find maxbetas -------------------
        # Make a list of all the input files for the job
        inputs = ' '
        for i in range(0,len(self.SERIAL2_flags_in)):
            inputs = '%s %s=%s/%s.H ' % (inputs,self.SERIAL2_flags_in[i],self.path_keep,self.SERIAL2_flags_in[i])
        par_in = 'par=%s par=%s par=%s' % (self.model_geom, self.acq, self.levelset)
        inputs = '%s %s secondphi=0.0 phi2=%s/phi1.H' % (inputs, par_in, self.path_keep)

        outputs = ' '
        for i in range(0,len(self.SERIAL2_flags_out)):
            outputs = '%s %s=%s/%s.H ' % (outputs,self.SERIAL2_flags_out[i],self.path_keep,self.SERIAL2_flags_out[i])
        outputs = '%s  weight2=watermask.H  wm_vel=1700.0 >&log_serial2b ' % (outputs)  # Doesn't depend of hfree_suffix 
        # 1500 doesnt work really. 1600 or 1700 [m/s] is what seems to work for the data Yi Shen gave me. I'm gonna do 2000 in order to get rid of some acquisition artifacts that are hard to get rid of otherwise.
        operator = '%s/SERIAL_2b.x ' % (self.T_BIN)
        args = '%s %s %s ' % (operator, inputs, outputs)
        if (debug):
            print('-----------------------------------------------------------------')
            print('SERIAL2')
            print(args)
        subprocess.call(args,shell=True)
        return




    def SERIAL2Hess(self, debug):
        #----- SERIAL2:   normalize, find maxbetas -------------------
        # Make a list of all the input files for the job
        inputs = ' '
        for i in range(0,len(self.SERIAL2_flags_in)):
            inputs = '%s %s=%s/%s.H ' % (inputs,self.SERIAL2_flags_in[i],self.path_keep,self.SERIAL2_flags_in[i])
        par_in = 'par=%s par=%s par=%s' % (self.model_geom, self.acq, self.levelset)
        inputs = '%s %s secondphi=0.0 phi2=%s/phi1.H' % (inputs, par_in, self.path_keep)

        outputs = ' '
        for i in range(0,len(self.SERIAL2_flags_out)):
            outputs = '%s %s=%s/%s.H ' % (outputs,self.SERIAL2_flags_out[i],self.path_keep,self.SERIAL2_flags_out[i])
        outputs = '%s  weight2=watermask.H  wm_vel=1700.0 >&log_serial2b_hess ' % (outputs)  # Doesn't depend of hfree_suffix 
        # 1500 doesnt work really. 1600 or 1700 [m/s] is what seems to work for the data Yi Shen gave me. I'm gonna do 2000 in order to get rid of some acquisition artifacts that are hard to get rid of otherwise.
        operator = '%s/SERIAL_2bHess.x ' % (self.T_BIN)
        args = '%s %s %s ' % (operator, inputs, outputs)
        if (debug):
            print('-----------------------------------------------------------------')
            print('SERIAL2Hess')
            print(args)
        subprocess.call(args,shell=True)
        return









    def CLEAN_DATA(self, param_reader, debug):
        #----- MAKE "CLEAN" DATA (since we dont have real field data) -------------------
        vel_mod_est = '%s/true_model.H' % (self.path_keep)
        cleandata = forward_modeling(param_reader,vel_mod_est,self.cleandata_name)
        bte = BatchTaskExecutor(param_reader)
        prefix = param_reader.prefix
        if (debug):
            print('-----------------------------------------------------------------')
            print(' MAKING "CLEAN" SYNTHETIC DATA')
        bte.LaunchBatchTask(prefix, cleandata)
        # Combine the files to one single file.
        _, fns_list = cleandata.GetSubjobsInfo()
        fn_seph_list = [fn for fns in fns_list for fn in fns]
        args=''
        for i in range(0,len(fn_seph_list)):
          args = '%s %s' % (args,fn_seph_list[i])
        cmd1 = 'Cat %s > %s/%s_full.H' % (args,self.path_keep,self.cleandata_name)
        # cmd2 = 'rm -rf %s/*'  % (path_out)
        # cmd = '%s; %s' % (cmd1, cmd2)
        cmd = '%s;' % (cmd1)
        if (debug):
            print('-----------------------------------------------------------------')
            print(cmd)
        subprocess.call(cmd,shell=True)
        return



    def CALC_MAXBETA(self, debug):
        # --------  Calculate the maxbeta   ----------------------
        inputs1 = 'tomo_grad=%s/%s.H  drlse_term1=%s/drlse_term1.H  phi_grad1=%s/phi_grad1.H '   % (self.path_keep,tomo_grad,self.path_keep,self.path_keep)
        inputs2 = 'par=%s  par=%s vel_full=%s/vel_full.H   phi2=%s/phi2.H   secondphi=0.0 '   % (self.model_geom,self.levelset,self.path_keep,self.path_keep)
        outputs = ' maxbeta=maxbeta.H    >&logMAXBETA' 
        operator = '%s/MAXBETA.x ' % (self.T_BIN)
        cmd = '%s %s %s %s' % (operator, inputs1, inputs2, outputs)
        if (debug):
            print('-----------------------------------------------------------------')
            print('CALC_MAXBETA')
            print(cmd)
        subprocess.call(cmd,shell=True)
        # --------  Grab the maxbeta   ----------------------
        file = open('maxbeta1.txt')
        for line in file:
            maxbeta = float(line)
        # --------  Grab the maxalpha   ----------------------
        file = open('maxalpha.txt')
        for line in file:
            maxalpha = float(line)
        if (debug):
            print('MAXBETA = %s          MAXALPHA = %s ') % (maxbeta, maxalpha)
        return maxbeta, maxalpha



    def SAVE_INTERMEDIATE_VALUES(self, nt0, debug):
        #------- Save the gradient intermediate values -----------------------
        cmd1 = 'Cat scaling_comp.H %s/scaling.H > %s/out.H' % (self.path_keep,self.path_out)
        cmd2 = 'rm scaling_comp.H; mv %s/out.H scaling_comp.H' % (self.path_out)
        cmd3 = 'Cat s1_comp.H %s/s1.H > %s/out.H' % (self.path_keep,self.path_out)
        cmd4 = 'rm s1_comp.H; mv %s/out.H s1_comp.H' % (self.path_out)
        cmd5 = 'Cat phigrad_comp.H %s/phi_grad1.H > %s/out.H' % (self.path_keep,self.path_out)
        cmd6 = 'rm phigrad_comp.H; mv %s/out.H phigrad_comp.H' % (self.path_out)
        cmd7 = 'Cat tomograd_comp.H %s/tomo_grad.H > %s/out.H' % (self.path_keep,self.path_out)
        cmd8 = 'rm tomograd_comp.H; mv %s/out.H tomograd_comp.H' % (self.path_out)
        cmd9 = 'Cat RTMgrad_comp2.H %s/RTMgrad.H > %s/out.H' % (self.path_keep,self.path_out)
        cmd10 = 'rm RTMgrad_comp2.H; mv %s/out.H RTMgrad_comp2.H' % (self.path_out)
        cmd11 = 'Cat GNHessian_comp.H %s/%s.H > %s/out.H' % (self.path_keep,self.hess_name,self.path_out)
        cmd12 = 'rm GNHessian_comp.H; mv %s/out.H GNHessian_comp.H' % (self.path_out)
        cmd13 = 'Cat drlse_comp.H %s/drlse_term1.H > %s/out.H' % (self.path_keep,self.path_out)
        cmd14 = 'rm drlse_comp.H; mv %s/out.H drlse_comp.H' % (self.path_out)
        cmd15 = 'Pad < %s/residual.H n1out=%s > res_pad.H ; Cat axis=4 residual_comp.H res_pad.H > %s/out.H' % (self.path_keep,nt0,self.path_out)
        cmd16 = 'rm residual_comp.H; mv %s/out.H residual_comp.H' % (self.path_out)
        cmd17 = 'Cat objfunc_comp.H %s/objfunc.H > %s/out.H' % (self.path_keep, self.path_out)
        cmd18 = 'rm objfunc_comp.H; mv %s/out.H objfunc_comp.H' % (self.path_out)
        cmd19 = 'Cat maxbeta_comp.H maxbeta.H > %s/out.H' % (self.path_out)
        cmd20 = 'rm maxbeta_comp.H; mv %s/out.H maxbeta_comp.H' % (self.path_out)
        cmd21 = 'Cat rnorm_comp.H rnorm.H > %s/out.H' % (self.path_out)
        cmd22 = 'rm rnorm_comp.H; mv %s/out.H rnorm_comp.H' % (self.path_out)
        cmdA = '%s; %s; %s; %s; %s; %s; %s; %s; %s; %s;' % (cmd1, cmd2, cmd3, cmd4, cmd5, cmd6, cmd7, cmd8, cmd9, cmd10 )
        cmdB = '%s; %s; %s; %s; %s; %s; %s; %s; %s; %s; %s; %s;' % (cmd11, cmd12, cmd13, cmd14, cmd15, cmd16, cmd17, cmd18, cmd19, cmd20, cmd21, cmd22 )
        cmd = ' %s %s' % (cmdA, cmdB)
        if (debug):
            print('-----------------------------------------------------------------')
            print('SAVE_INTERMEDIATE_VALUES')
            print(cmd)
        subprocess.call(cmd,shell=True)
        return



    def CALC_OBJ_VAL(self, debug):
        #----- CALCULATE THE OBJECTIVE FUNCTION VALUE  -------------------
        cmd = ' %s/CALC_OBJ_VAL.x residual=%s/residual.H  > %s/objfunc.H ' % (self.T_BIN,self.path_keep,self.path_keep)
        if (debug):
            print('-----------------------------------------------------------------')
            print('CALC_OBJ_VAL')
            print(cmd)

        subprocess.call(cmd,shell=True)
        #------- GET NEW OBJ FUNCTION VALUE  ----------------
        file = open('objfuncval.txt')
        for line in file:
            new_objfuncval = float(line)

        return new_objfuncval





    def SET_TIME_SAMPLES(self, param_reader, it, debug):
        #----------- Choose the value for NT (number of time samples, for TOP-DOWN algorithm) --------------
        if (it<=self.fulltime_it):
            difft=float(nt0)-float(ot)
            # nt=float(difft)*(i/float(fulltime_it)) + float(ot)
        # Window the clean-data
        cmd = 'Window n1=%s < %s/%s_full.H > %s/%s.H' % (nt,self.path_keep,self.cleandata_name,self.path_keep,self.cleandata_name)
        if (debug):
            print('-----------------------------------------------------------------')
            print('SET_TIME_SAMPLES')
            print(cmd)

        subprocess.call(cmd,shell=True)
        # Rewrite nt to the param_reader
        dict_args['nt']=nt
        param_reader.dict_args = dict_args
        # write debug output

        return nt




    def SMOOTH_VEL_MODEL(self, debug):
        #------------------------------------------------------
        cmd0 = 'cp %s/vel_full.H  %s/vel_full_ns.H' % (self.path_keep,self.path_keep)
        cmd1 = 'Smooth repeat=10 rect1=4 rect2=4 < %s/vel_full.H > %s/vel_full_smooth.H' % (self.path_keep,self.path_keep)
        cmd2 = 'mv %s/vel_full_smooth.H  %s/vel_full.H' % (self.path_keep,self.path_keep)
        cmd = '%s; %s; %s;' % (cmd0, cmd1, cmd2)   
        if (debug):
            print('-----------------------------------------------------------------')
            print('SMOOTH_VEL_MODEL')
            print(cmd)
        subprocess.call(cmd,shell=True)
        return



    def SAVE_VEL_MODEL(self, debug):
        #----- Write the velmodel to output (current iteration) --------------------
        cmd1 = 'Cat vel_comp.H %s/vel_full.H > %s/out.H' % (self.path_keep,self.path_out)
        cmd2 = 'rm vel_comp.H; mv %s/out.H vel_comp.H' % (self.path_out)
        cmd = '%s; %s; ' % (cmd1, cmd2)
        if (debug):
            print('-----------------------------------------------------------------')
            print('SAVE_VEL_MODEL')
            print(cmd)
        subprocess.call(cmd,shell=True)
        return



    def CLEAN_WRK(self, debug):
        #----- Write the velmodel to output (current iteration) --------------------
        cmd = 'rm %s/*' % (self.path_out)
        if (debug):
            print('-----------------------------------------------------------------')
            print('CLEAN_WRK')
            print(cmd)
        subprocess.call(cmd,shell=True)
        return








            
          

    def BUILD_MODEL(self, debug, p1, buildonly):
        # ------- BUILD MODEL FROM NEW PHI   ------------------------
        inputs = ''
        for i in range(0,len(self.MODVEL_flags_in)):
            inputs = '%s %s=%s/%s.H ' % (inputs,self.MODVEL_flags_in[i],self.path_keep,self.MODVEL_flags_in[i])
        par_in = 'par=%s drlseonly=0.0 buildonly=%s' % (self.model_geom,buildonly)
        if (buildonly==1.0):
            inputs = '%s %s p1=%s p2=%s alpha=%s ' % (inputs,par_in,p1,0.0,0.0)
        else:
            inputs = '%s %s p1=%s p2=%s alpha=%s phi_grad1=%s/%s.H ' % (inputs,par_in,p1,0.0,0.0,self.path_keep,self.phi_grad1)
    
        # ------ Ignore the phi2 and the alpha iterations. Just do Hessian free for the phi1, no splitting. 
        outputs = 'vel_full_test=%s/vel_full.H   phi1_out=%s/phi1.H    vel_back_out=%s/vel_back.H  >&log_MODVEL2' % (path_keep,path_keep,path_keep)
        operator = '%s/MOD_VEL2.x ' % (self.T_BIN)
        args = '%s %s %s ' % (operator, inputs, outputs)

        # if (buildonly==0.0):
            # file = open('MODVELnorm.txt')
            # for line in file:
            #     MODVELnorm = float(line)
            #     print("sum(velmodel) = %s") % (MODVELnorm)

        if (debug):
            print('-----------------------------------------------------------------')
            print('BUILD_MODEL')
            print(args)
        subprocess.call(args,shell=True)
        return
  



    def SERIAL1(self, single_well, debug):
        singlewell = int(single_well)
        #----- SERIAL1:   Make DRLSE and scaling terms -------------------
        # Make a list of all the input files for the job
        inputs = ' '
        for i in range(0,len(self.SERIAL1_flags_in)):
            inputs = '%s %s=%s/%s.H ' % (inputs,self.SERIAL1_flags_in[i],self.path_keep,self.SERIAL1_flags_in[i])
        par_in = 'par=%s par=%s singlewell=%s' % (self.model_geom, self.acq, singlewell)
        inputs = '%s %s ' % (inputs, par_in)
        # Make a list of all the output files for the job
        outputs = ' '
        for i in range(0,len(self.SERIAL1_flags_out)):
            outputs = '%s %s=%s/%s.H ' % (outputs,self.SERIAL1_flags_out[i],self.path_keep,self.SERIAL1_flags_out[i])
        outputs = '%s >&log_serial1 ' % (outputs)
        # Build the command argument
        args = ' '
        operator = '%s/SERIAL_1.x ' % (self.T_BIN)
        args = '%s %s %s ' % (operator, inputs, outputs)
        if (debug):
            print('-----------------------------------------------------------------')
            print('SERIAL1')
            print(args)
        subprocess.call(args,shell=True)
        return






    def DRLSE(self, debug):
        # ------- MAKE MODEL  ------------------------
        inputs = ''
        for i in range(0,len(self.MODVEL_flags_in)):
            inputs = '%s %s=%s/%s.H ' % (inputs,self.MODVEL_flags_in[i],self.path_keep,self.MODVEL_flags_in[i])
        par_in = 'drlseonly=1.0  par=%s buildonly=0.0' % (self.model_geom)
        inputs = '%s %s p1=%s p2=%s alpha=%s ' % (inputs,par_in,0.0,0.0,0.0)
        # ------ Ignore the phi2 and the alpha iterations. Just do Hessian free for the phi1, no splitting. 
        outputs = 'vel_full_test=%s/vel_full.H   phi1_out=%s/phi1.H   >&logDRLSE_modvel2' % (path_keep,path_keep)
        operator = '%s/MOD_VEL2.x ' % (self.T_BIN)
        args = '%s %s %s ' % (operator, inputs, outputs)

        if (debug):
            print('-----------------------------------------------------------------')
            print('DRLSE')
            print(args)
        subprocess.call(args,shell=True)
            # file = open('MODVELnorm.txt')
            # for line in file:
            #     MODVELnorm = float(line)
            #     print("sum(velmodel) = %s") % (MODVELnorm)
            # file = open('phi1norm.txt')
            # for line in file:
            #     phi1norm = float(line)
            #     print("sum(phi1) = %s") % (phi1norm)
        return



    def COPY_TO_OLD_PHI(self, debug):
        args = 'Cp %s/phi1.H  %s/phi1_old.H ' % (path_keep,path_keep)
        subprocess.call(args,shell=True)
        if (debug):
            print('-----------------------------------------------------------------')
            print('COPY_TO_OLD_PHI')
            print(args)
        return


    def COPY_FROM_OLD_PHI(self, debug):
        args = 'Cp %s/phi1_old.H  %s/phi1.H   ' % (path_keep,path_keep)
        subprocess.call(args,shell=True)
        if (debug):
            print('-----------------------------------------------------------------')
            print('COPY_FROM_OLD_PHI')
            print(args)
        return


    def DONE_ITERATING(self, debug):
        #----- LAST ITERATION:  Write the velmodel to output --------------------
        cmd1 = 'Cat vel_comp.H %s/vel_full.H > %s/out.H' % (path_keep,path_out)
        cmd2 = 'rm vel_comp.H; mv %s/out.H vel_comp.H' % (path_out)
        cmd = '%s; %s; ' % (cmd1, cmd2)
        subprocess.call(cmd,shell=True)
        if (debug):
            print('-----------------------------------------------------------------')
            print('DONE_ITERATING')
            print(cmd)
        return




    def LINE_SEARCH(self, param_reader, debug, NLiter, maxbeta, lastval):
        # Initialize the lists
        testP=[0]*NLiter
        new_objfuncval=[0]*NLiter
        X=[0]*NLiter
        Y=[0]*NLiter

        # Go through a set of test values
        if (NLiter<4): print("You should increase the valeu of NLiter; kinda small.")
        for ip in range(0,NLiter-1):
            dp = maxbeta/(NLiter-1)
            testP[ip] = ip*dp
            if (debug):  print(">>>>>>>>>>   TEST BETA =  %s") % (testP[ip])
            #------ Grab the old Phi ------------------------------------------
            main.COPY_FROM_OLD_PHI(debug)
            main.BUILD_MODEL(debug=True,p1=testP[ip],buildonly=0.0) # Build the new model
            prefixA = "syndata_LINESEARCH_%s" % (ip)
            gradient_maker.syn_data(param_reader,debug,prefixA) # Make the synthetic data
            gradient_maker.residual(param_reader,debug,prefixA) # Make the residual
            new_objfuncval[ip]=main.CALC_OBJ_VAL(debug) # Calc the objective function value
        
        # Manually set the last values (since we calculated it earlier; saves a forward prop)
        testP[NLiter-1] = (NLiter-1)*dp
        new_objfuncval[NLiter-1] = lastval

        # Find the best objective function value and the Beta that made it
        bestval=max(new_objfuncval)
        maxip= (NLiter-1)
        if (debug): print("range(0,NLiter) = %s    maxip = %s    dp =%s       NLiter=%s") % (range(0,NLiter),maxip, dp, NLiter)
        for ip in range(0,NLiter):
            if (debug): print("beta = %s       objfuncval= %s   ip=%s") % (testP[ip],  new_objfuncval[ip], ip)
            if (new_objfuncval[ip]<bestval):
                bestval=new_objfuncval[ip]
                best_ip=ip

        # Bracket the three best values
        if (maxip>best_ip>0):
            X[1] = testP[best_ip-1] 
            X[2] = testP[best_ip] 
            X[3] = testP[best_ip+1] 
            Y[1] = new_objfuncval[best_ip-1] 
            Y[2] = new_objfuncval[best_ip] 
            Y[3] = new_objfuncval[best_ip+1]
        # (Best ip is the LAST one)
        elif(best_ip==maxip):
            X[1] = testP[best_ip-2] 
            X[2] = testP[best_ip-1] 
            X[3] = testP[best_ip] 
            Y[1] = new_objfuncval[best_ip-2] 
            Y[2] = new_objfuncval[best_ip-1] 
            Y[3] = new_objfuncval[best_ip]
        # (Best ip is the FIRST one)
        elif(best_ip==0):
            X[1] = testP[best_ip] 
            X[2] = testP[best_ip+1] 
            X[3] = testP[best_ip+2] 
            Y[1] = new_objfuncval[best_ip] 
            Y[2] = new_objfuncval[best_ip+1] 
            Y[3] = new_objfuncval[best_ip+2]

        # Invert the Adjoint : (X^T)-1
        a = np.array(  [ [ (X[1]**2.), X[1], 1 ], [ (X[2]**2.), X[2], 1 ], [ (X[3]**2.), X[3], 1 ] ]  )
        b = np.array(  [ [ Y[1], Y[2], Y[3] ] ]  )
        ainv = inv(a)

        # Print if debug
        if (debug): print("ainv = %s") % (ainv)
        if (debug): print("b = %s") % (b)

        # Multiply : ((X^T)-1)*d
        M = np.dot(ainv, np.transpose(b))
        # Find the minimum value
        final_beta = -1.0*M[1][0]/(2.*M[0][0])

        # Put checks on final_beta
        if (final_beta<=0.0):
            print("final_beta = %s. The MAXBETA value is negative or zero. That means the update (gradient + drlseterm) isn't that good. Going to use maxbeta/(NLiter-1) as the step to try and get out of the local minima. ") % (final_beta)
            final_beta = dp
        elif (final_beta>maxbeta):
            print("optimal beta =  %s.       maxbeta = %s.    Reverting to maxbeta. ") % (final_beta, maxbeta)
            final_beta = maxbeta

        # Check the objfunc_val guess
        final_obj= (final_beta**2.)*M[0][0]  +  (final_beta)*M[1][0]  +  M[2][0]

        # Not necessarily a minimum!! Could be a maximum!!
        if (final_obj>bestval):
            print("Optimal beta is likely at a maximum, not a minimum.  final_beta=%s, final_obj=%s   best_ip=%s,  bestval=%s. Using best_ip instead. ") % (final_beta,final_obj,best_ip,bestval)
            final_beta = testP[best_ip]
            if (final_beta<=0.0):
                print("final_beta = %s. The MAXBETA value is negative or zero. That means the update (gradient + drlseterm) isn't that good. Going to use maxbeta/(NLiter-1) as the step to try and get out of the local minima. ") % (final_beta)
                final_beta = dp

        # Redo the objfunc_val guess
        final_obj= (final_beta**2.)*M[0][0]  +  (final_beta)*M[1][0]  +  M[2][0]

        return final_beta, final_obj






    def SOURCE_inversion(self,debug):
        # -------------------------------------------------------------------------------------
        cmd1 = '%s/SOURCE_INVERSION.x  par=%s  par=%s  par=%s ' % (self.T_BIN, self.model_geom, self.acq, self.levelset)
        cmd2 = ' isou=1 ns=1  velmodel=%s  data=%s/%s.H  source=%s/delta_source.H  waveletin=%s/waveletC.H  >/dev/null ' %  (self.vel_mod_path, self.path_keep,self.res_name, self.path_keep, self.path_keep)
        cmd = '%s  %s'  % (cmd1, cmd2)
        subprocess.call(cmd,shell=True)
        if (debug):
            print('-----------------------------------------------------------------')
            print('SOURCE_inversion')
            print(cmd)
        return








#XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
#XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
#XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX


if __name__ == '__main__':
  print "Run BatchTaskExecutor_test with params:", sys.argv
  # Check basic environment variable setup.
  assert 'SEP' in os.environ, '!Environment var SEP is not set yet, should have your seplib enviroment set up first! Check paths like /opt/SEP/'
  assert 'RSF' in os.environ, '!Environment var RSF is not set yet, should have your RSF environment set up first! Check paths like /opt/RSF'
  eq_args_from_cmdline,args = sepbase.parse_args(sys.argv)
  param_reader = pbs_util.JobParamReader(eq_args_from_cmdline)

  dict_args = param_reader.dict_args
  cleandata_name = dict_args['cleandata_name'] 
  path_keep = dict_args['path_keep']
  levelset_iter = int(dict_args['levelset_iter'])
  phi_grad1 = dict_args['phi_grad1']
  model_geom = dict_args['model_geom']
  T_BIN = dict_args['T_BIN']
  MV_flags_in = dict_args['MODVEL_flags_in']
  MODVEL_flags_in = MV_flags_in.split()
  rtm_name = (dict_args['rtm_name'])
  phi1=dict_args['phi1']
  path_out = dict_args['path_out']
  path_save= dict_args['path_save']
  levelset = dict_args['levelset']
  tomo_grad = dict_args['tomo_grad']
  hess_name= dict_args['hess_name']

  fulltime_it= int(dict_args['fulltime_it'])
  nt = dict_args['nt']
  numdrlse = int(dict_args['numdrlse'])

  #------ INITIALIZE SOME STUFF ----------------------
  ot=500.0
  nt0=nt
  old_objfuncval=(10**20)
  istomo=0
  sou_up=1
  best_eps=0.0
  first_it=1.0
  NLiter=5
  best_BETA=0.0
  objvaldone=False
  force=False

  debug=True


  # INITIALIZING THE MAIN FUNCTION SPACE. 
  # FUNCTIONS THAT CALL OTHER FUNCTIONS RELY ON THIS OBJECT INSTANCE
  main=MAIN_FUNCTIONS(param_reader)

  #----------- START BY DOING SOME DRLSE FOR A FEW ROUNDs ---------------
  for i in range(0,numdrlse):
      main.SERIAL1(single_well=0.0,debug=debug)  # 1.0 is singlewell, 0.0 is double well
      main.DRLSE(debug)
      words="Doing DRLSE on initial phi %s/%s" % (i,numdrlse)
      print(words)


  # #----------- MAKE THE CLEAN DATA (SYNTHETIC) --------------------------
  # main.CLEAN_DATA(param_reader,debug)
  # main.CLEAN_WRK(debug) # Dont want those saved wavefields interfering

  # ------- BUILD INITIAL MODEL FROM NEW PHI AND MAXBETA  -------------------
  main.BUILD_MODEL(debug,p1=best_BETA,buildonly=1.0)

#######################################################################################
  #----------- BEGIN LEVELSET ITERATIONS ---------------------
  for i in range(0,levelset_iter):
      print("###################################################################################################")
      print("########################      LEVELSET ITERATION   %s / %s   first_it=%s  ############################") % (i, (levelset_iter-1), first_it)
      print("###################################################################################################")

      #----------- Choose the value for NT (number of time samples, for TOP-DOWN algorithm) --------------
      nt = main.SET_TIME_SAMPLES(param_reader, i, debug)

      # ------- If the iteration is even, do tomoupdate   ---------------
      # if ((i % 2) == 0):
      #   istomo=1.0
      # print(istomo)


     # ------- If the iteration is even, do source update   ---------------
      if ((i % 2) == 0):
        sou_up=1.0
      print(sou_up)


      # #------- SMOOTH MODEL IF IT's A TOMO ITERATION ------------------
      # if (istomo):
      #     main.SMOOTH_VEL_MODEL(debug)


      #----- CALCULATE THE OBJECTIVE FUNCTION VALUE  --------------------
      if (not(objvaldone)):
          #------- MAKE SYNTHETIC DATA AND RESIDUAL -------------------------
          gradient_maker=make_gradient(param_reader)
          gradient_maker.syn_data(param_reader,debug,"syndata_MAIN")
          gradient_maker.residual(param_reader,debug,"syndata_MAIN")
          new_objfuncval=main.CALC_OBJ_VAL(debug)
      objvaldone=False


      #----- CALCULATE THE SOURCE PERTURBATION --------------------
      if (sou_up):
          main.SOURCE_inversion(debug)



      #VVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVV
      #------- CHECK NEW OBJ FUNCTION VALUE  ----------------------------
      print("VVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVV")
      print("new_objfuncval = %s            old_objfuncval = %s ") % (new_objfuncval, old_objfuncval)
      if ((new_objfuncval<old_objfuncval) or (force)):
          if (force): print(" FORCING TO MAKE A NEW GRADIENT, SINCE THE PREVIOUS LINE SEARCH DID THE BEST IT COULD.")
      # if (first_it):

          #------- MAKE GRADIENT -----------------------------------------
          gradient_maker.born(param_reader,debug)

          # ------- APPLY GAUSS NEWTON HESSIAN TO GRADIENT ----------------
          GN_inversion(param_reader)

          #----- SERIAL2: Make scaled down phi_grad1 and tomo_grad -------
          # main.SERIAL2(debug)
          main.SERIAL2Hess(debug)

          maxbeta,maxalpha = main.CALC_MAXBETA(debug)
          best_BETA = maxbeta

          #----- Save the output files --------------------
          main.SAVE_VEL_MODEL(debug)
          if (first_it):
              main.INIT_INTERMEDIATE_OUTPUTS(nt0,debug)
          else:
              main.SAVE_INTERMEDIATE_VALUES(nt0, debug)

          # ------- COPY TO OLD OUTPUTS  --------------------------------------
          main.COPY_TO_OLD_PHI(debug)
          old_objfuncval = new_objfuncval

          # ------- BUILD MODEL FROM NEW PHI AND MAXBETA  -------------------
          main.BUILD_MODEL(debug,p1=best_BETA,buildonly=0.0)
         
          #------- MAKE DRLSE, S1, and SCALING ---------------------------
          main.SERIAL1(single_well=0.0,debug=debug)  # 1.0 is singlewell, 0.0 is double well

          # ------- CLEAN THE WORKING DIRECTORY   -------------------------------
          main.CLEAN_WRK(debug)

          # Force to make a new gradient using this update? NO
          force=False

      else:

          print('VVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVV')
          best_BETA, new_objfuncvalguess = main.LINE_SEARCH(param_reader, debug, NLiter, maxbeta, new_objfuncval)
          words='Optimally chosen BETA = %s  Corresponding (estimated) objfunc_val = %s    istomo = %s' % (best_BETA, new_objfuncvalguess, istomo)
          print(words)

          # ------- BUILD NEW MODEL FROM NEW PHI AND MAXBETA  ---------------
          main.COPY_FROM_OLD_PHI(debug)
          main.BUILD_MODEL(debug,p1=best_BETA,buildonly=0.0)
         
          #------- MAKE DRLSE, S1, and SCALING ---------------------------
          main.SERIAL1(single_well=0.0,debug=debug)  # 1.0 is singlewell, 0.0 is double well


          # ------- CLEAN THE WORKING DIRECTORY   -------------------------------
          main.CLEAN_WRK(debug)
          

          gradient_maker=make_gradient(param_reader)
          gradient_maker.syn_data(param_reader,debug,"syndata_MAIN")
          gradient_maker.residual(param_reader,debug,"syndata_MAIN")
          # Why? Because the local estimate of the objfunc_val is not necessarily
          # the real objfunc_val

          #----- CALCULATE THE OBJECTIVE FUNCTION VALUE  --------------------
          new_objfuncval=main.CALC_OBJ_VAL(debug)
          objvaldone=True

          # ------- SAVE  OUTPUTS  ------------------------------------------
          main.SAVE_VEL_MODEL(debug)
          main.COPY_TO_OLD_PHI(debug)
          if (first_it):
              main.INIT_INTERMEDIATE_OUTPUTS(nt0,debug)
          else:
              main.SAVE_INTERMEDIATE_VALUES(nt0, debug)

          # Force to make a new gradient using this update? YES
          force=True

      #VVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVV
      first_it=0

#===============================================================================



  main.DONE_ITERATING(debug)


