!!---------------------------------------------------------------------------
!! Author: Yaxun Tang, tang@sep.stanford.edu
!! 
!! Revision: Yang Zhang, add linear-wfld interpolation, 2014.04
!!---------------------------------------------------------------------------
module bwi_ssrepi_mod
  implicit none

contains
  !----------------------------------------------------------------------------------------------------------------------
  !.. Multiple reference velocity extrapolation plus interpolation
  !----------------------------------------------------------------------------------------------------------------------
  integer function ssr_epi(adj, Fxop, wfld_slice, vmod_slice, vref_slice, kx2, ky2, ws, dz, nx, ny, nvrf, opsign, &
                           iwrk2d, cwrk3d, plan_2dxF, plan_2dxI, fwrk2d) result(stat)
    logical                   :: adj
    complex, dimension(:,:)   :: wfld_slice
    real   , dimension(:,:)   :: vmod_slice
    real   , dimension(:)     :: vref_slice
    real   , dimension(:)     :: kx2, ky2
    real                      :: ws
    real                      :: dz
    integer                   :: nx, ny, nvrf
    integer                   :: opsign
    integer, dimension(:,:)           :: iwrk2d
    real, dimension(:,:),optional     :: fwrk2d
    complex, dimension(:,:,:) :: cwrk3d
    integer*8                 :: plan_2dxF
    integer*8                 :: plan_2dxI

    interface
      integer function Fxop(adj, wfld_slice, vmod_slice, kx2, ky2, vref, ws, dz, nx, ny, opsign, plan_2dxF, plan_2dxI) result(stat)
        logical                 :: adj
        complex, dimension(:,:) :: wfld_slice
        real   , dimension(:,:) :: vmod_slice
        real   , dimension(:)   :: kx2, ky2
        real                    :: vref
        real                    :: ws
        real                    :: dz
        integer                 :: nx, ny
        integer                 :: opsign
        integer*8               :: plan_2dxF
        integer*8               :: plan_2dxI
      end function
    end interface

    integer                     :: ix, iy, itr, ivrf, jvrf
    real                        :: vmin, vmax, dvel, vref, vdif1, vdif2
    logical   :: b_linear_interp

    b_linear_interp = present(fwrk2d)
    vmax = maxval(vmod_slice)
    
    !write(0,*) "b_linear_interp",b_linear_interp, "vmax", vmax, "nvrf",nvrf

    if (nvrf == 1) then
       vref = vref_slice(1)
       stat = FXop(adj, wfld_slice, vmod_slice, kx2, ky2, vref, ws, dz, nx, ny, opsign, plan_2dxF, plan_2dxI)
    else
       iwrk2d = 1; 
!      calculate the velocity map for wavefield interpolation, here switch from nearest-neighbor interpolation to linear interploation of the velocity value (maybe slowness is a better idea.)
!      ------------------------------------------------------
       if (.not. b_linear_interp) then
         !call erexit("!should use linear interpolation in bwi_ssrepi_mod.f90!")
         !write(0,*) "FF3.1"
         do iy=1, ny
         do ix=1, nx
            jvrf = 1
            ! Trying to locate the nearest vref values, for linear interpolation, it is not enough
            do ivrf=1, nvrf-1
               vdif1 = abs(vmod_slice(ix,iy)-vref_slice(ivrf)  )
               vdif2 = abs(vmod_slice(ix,iy)-vref_slice(ivrf+1))
               if (vdif1<vdif2) then
                  jvrf = ivrf
                  exit
               else
                  jvrf = ivrf+1
               end if
            end do
            iwrk2d(ix,iy) = jvrf
          end do
          end do
       else 
         fwrk2d = 1.0
         !! Do linear interpolation, iwrk2d will store the bigger vref velocity in the range of (vr_min,vr_max), and fwrk2d will store the linear interp weight for iwrk2d
         do iy=1, ny
         do ix=1, nx
           if (vmod_slice(ix,iy) <= vref_slice(1)) then !Check if vmod is smaller than the lowest vref(1)
             iwrk2d(ix,iy) = 1
             fwrk2d(ix,iy) = 1.
           else if (vmod_slice(ix,iy) >= vref_slice(nvrf)) then !check if vmod is very large
             iwrk2d(ix,iy) = nvrf
             fwrk2d(ix,iy) = 1.
           else !Search the range of vrefs and find the weight
            do ivrf=nvrf-1,1,-1
               vdif1 = vmod_slice(ix,iy)-vref_slice(ivrf)
               if (vdif1 >= 0.) then !calculate the weights
                 vdif2 = abs(vmod_slice(ix,iy)-vref_slice(ivrf+1))
                 iwrk2d(ix,iy) = ivrf+1
                 if (vdif1+vdif2 > 0.005*vmax) then 
                  fwrk2d(ix,iy) = vdif1/(vdif1+vdif2)
                 else  !probably v1 and v2 are the same. so do not do division
                  fwrk2d(ix,iy) = 1.0
                 end if
                 exit
               end if
            end do
           end if
         end do
         end do
         if (minval(fwrk2d)<0. .or. maxval(fwrk2d)>1.0) then
           write(0,*) "!!! Sth is wrong about fwrk2d!", minval(fwrk2d) ,"/", maxval(fwrk2d)
           call erexit("")
         end if
       end if

       if (minval(iwrk2d)<1 .or. maxval(iwrk2d)>nvrf) then
         write(0,*) "!!! Sth is wrong about iwrk2d!", minval(iwrk2d) ,":", maxval(iwrk2d)
         call erexit("!!WRONG!!")
       end if
       !write(0,*) "FF3.5"

       if (adj) then
!         *********************
!         ADJOINT INTERPOLATION
!         *********************
!         copy current interpolated wavefield into a backup array
!         -------------------------------------------------------
          cwrk3d(:,:,1) = wfld_slice
!         zero out buffer
!         ---------------
          cwrk3d(:,:,2) = 0.
          
          if (.not. b_linear_interp) then
  !         loop over reference velocities
  !         ------------------------------
            do ivrf=1, nvrf
               vref = vref_slice(ivrf)
  !            adjoint of nearest-neighbour interpolation
  !            ------------------------------------------ 
               wfld_slice = 0.
               where (iwrk2d == ivrf)
                     wfld_slice = cwrk3d(:,:,1)
               end where
  !            adjoint propagation
  !            -------------------
               stat = FXop(adj, wfld_slice, vmod_slice, kx2, ky2, vref, ws, dz, nx, ny, opsign, plan_2dxF, plan_2dxI)
  !            adjoint of copying wavefields
  !            -----------------------------
               cwrk3d(:,:,2) = cwrk3d(:,:,2) + wfld_slice
            end do
  !         copy the wavefield back
  !         ----------------------- 
            wfld_slice = cwrk3d(:,:,2)
          else !under linear interpolation
  !         loop over reference velocities
  !         ------------------------------
            do ivrf=1, nvrf
               vref = vref_slice(ivrf)
  !            adjoint of linear interpolation
  !            ------------------------------------------ 
               wfld_slice = 0.

              !! where (iwrk2d == ivrf)
              !!   wfld_slice = cwrk3d(:,:,1)*fwrk2d
              !! end where
              !! where (iwrk2d == ivrf+1)
              !!   wfld_slice = wfld_slice + cwrk3d(:,:,1)*(1-fwrk2d)
              !! end where
              do iy=1, ny
              do ix=1, nx
                if (iwrk2d(ix,iy) == ivrf) then
                  wfld_slice(ix,iy) = cwrk3d(ix,iy,1)*fwrk2d(ix,iy)
                end if
                if (iwrk2d(ix,iy) == ivrf+1) then
                  wfld_slice(ix,iy) = wfld_slice(ix,iy)+cwrk3d(ix,iy,1)*(1.-fwrk2d(ix,iy))
                end if
              end do
              end do

  !            adjoint propagation
  !            -------------------
               stat = FXop(adj, wfld_slice, vmod_slice, kx2, ky2, vref, ws, dz, nx, ny, opsign, plan_2dxF, plan_2dxI)
  !            adjoint of copying wavefields
  !            -----------------------------
               cwrk3d(:,:,2) = cwrk3d(:,:,2) + wfld_slice
            end do
  !         copy the wavefield back
  !         ----------------------- 
            wfld_slice = cwrk3d(:,:,2)
          end if
        else !if adj==false
!         *********************
!         FORWARD INTERPOLATION
!         *********************

!         copy current wavefield into a backup array
!         ------------------------------------------
          cwrk3d(:,:,1) = wfld_slice
!         zero out buffers
!         ----------------
          cwrk3d(:,:,2) = 0.
          wfld_slice    = 0.

          if (.not. b_linear_interp) then
  !         loop over reference velocities
  !         ------------------------------
            do ivrf=1, nvrf
  !            propagate wavefields according to different reference velocities
  !            ----------------------------------------------------------------
               vref = vref_slice(ivrf)
               wfld_slice = cwrk3d(:,:,1)
               stat = FXop(adj, wfld_slice, vmod_slice, kx2, ky2, vref, ws, dz, nx, ny, opsign, plan_2dxF, plan_2dxI)
  !            nearest-neighbor interpolation
  !            ------------------------------ 
               where (iwrk2d == ivrf)
                     cwrk3d(:,:,2) = wfld_slice
               end where
            end do
  !         copy the interpolated wavefield back to current wavefield
  !         ---------------------------------------------------------
            wfld_slice = cwrk3d(:,:,2)
            !write(0,*) "FF4"
          else
  !         loop over reference velocities
  !         ------------------------------
            do ivrf=1, nvrf
  !            propagate wavefields according to different reference velocities
  !            ----------------------------------------------------------------
               vref = vref_slice(ivrf)
               wfld_slice = cwrk3d(:,:,1)
               stat = FXop(adj, wfld_slice, vmod_slice, kx2, ky2, vref, ws, dz, nx, ny, opsign, plan_2dxF, plan_2dxI)

  !            linear interpolation
  !            ------------------------------ 
              !! where (iwrk2d == ivrf)
              !!   cwrk3d(:,:,2) = cwrk3d(:,:,2) + wfld_slice*fwrk2d
              !! end where
              !! where (iwrk2d == ivrf+1)
              !!   cwrk3d(:,:,2) = cwrk3d(:,:,2) + wfld_slice*(1-fwrk2d)
              !! end where
              do iy=1, ny
              do ix=1, nx
                if (iwrk2d(ix,iy) == ivrf) then
                  cwrk3d(ix,iy,2) = cwrk3d(ix,iy,2) + wfld_slice(ix,iy)*fwrk2d(ix,iy)
                end if
                if (iwrk2d(ix,iy) == ivrf+1) then
                  cwrk3d(ix,iy,2) = cwrk3d(ix,iy,2) + wfld_slice(ix,iy)*(1.-fwrk2d(ix,iy))
                end if
              end do
              end do
            end do
  !         copy the interpolated wavefield back to current wavefield
  !         ---------------------------------------------------------
            wfld_slice = cwrk3d(:,:,2)
          end if !end if not b_linear_interp
          !write(0,*) "FF5"
        end if !end if adjoing
      end if !end if nvrf==1
   
   !write(0,*) "FF6"
   stat = 0
  end function
end module

