!!.. read nzs depth step velocity or slowness
module bwi_velref_mod
  implicit none
  integer, private :: nvrf
  integer, private :: ncalls,ndepth,nsztt
contains

!=====================================================================
!
  integer function bwi_velref_init(nvrf_in) result(stat)
    integer :: nvrf_in
    integer :: nx_in
    integer :: ny_in

    nvrf  = nvrf_in
    stat  = 0
    !ncalls = 0; ndepth = 0; nsztt = 0
  end function
!======================================================================
  integer function bwi_velref_clos() result(stat)
    nvrf  = 0
    stat  = 0
  end function

!=================================================================
  !-----------------------------------------------
  !               Quantile function: Find the kth smallest value from array a
  !		eg. if k=1		 : Find the smallest value, Zero spikeness
  !			k=size(a)	: Find the biggest value, Full spikenss.
  ! This may cause STACK OVERFLOW because the result of pack is 
  !   is a temp array established on the stack, not in memory heap, MUST use with heap-arrays option in compiling
  !-----------------------------------------------
  recursive function quantile(k,a) result( value)
    integer,             intent (in)  :: k          
    ! position in array
    real, dimension (:), intent (in)  :: a
    real                              :: value      
    ! output value of quantile
    integer                           :: j
    real                              :: ak
    !if (k>size(a)) then
    !  write(0,*) "error!! k>size(a) in quantile", k, size(a); call seperr("error!! k>size(a) in quantile")
    !end if
    
    !ndepth = ndepth + 1; ncalls = ncalls + 1; nsztt  = nsztt + size(a)
    !write(0,*) "ndepth",ndepth, ncalls,k,size(a),nsztt

    ak = a(k)
    j = count( a < ak)  
    ! how many a(:) < ak
    if ( j >= k) then
       value = quantile( k, pack( a, a < ak))
    else
       j = count( a > ak) + k - size( a)
       if ( j > 0) then
          value = quantile( j, pack( a, a > ak))
       else
          value = ak
       end if
    end if
  end function

!======================================================================
!.. select the reference velocity for the current depth step
!.. !!! Lloyd's method will be used, THIS IS NOT Lloyd algorithm, but it does handle salt velocity case where there is a gap between salt velocity (vmax) and the sorrouding sediments.
!   !!! Add a small patch to delineate the salt velocity
!----------------------------------------------------------------------
  integer function bwi_velref_oper(vmod_slice, vref_slice) result(stat)
    real, dimension(:,:),intent(in)  :: vmod_slice
    real, dimension(:),intent(out)   :: vref_slice
    real, dimension(:),allocatable   :: vslice1d !,vslice1d_2
    integer              :: iref, nvrf_sec, ix, iy, ic, nx,ny,nsel1,nsel2,k1,k2
    real                 :: vmin_pri, vmax_pri, dv_min,dv_max, qntl
    real                 :: vmin_sec, vmax_sec, dvel, t
    logical              :: b_gap_vmin, b_gap_vmax

    stat = 1
    if(nvrf > size(vref_slice)) call seperr("ERROR: nref larger than the size of vref_slice(:)")
    vref_slice=0.0

    if (nvrf==1) then
       vref_slice(1) = sum(vmod_slice)/size(vmod_slice) !Use average velocity

    else if (nvrf==2) then
       vref_slice(1) = minval(vmod_slice)
       vref_slice(2) = maxval(vmod_slice)

    else if (nvrf==3) then !Use min/max and average velocity
       vref_slice(1) = minval(vmod_slice)
       vref_slice(2) = sum(vmod_slice)/size(vmod_slice)
       vref_slice(3) = maxval(vmod_slice)

    else
       vmin_pri = minval(vmod_slice)
       vmax_pri = maxval(vmod_slice)
       vref_slice(1)    = vmin_pri
       vref_slice(nvrf) = vmax_pri
       nx = size(vmod_slice,1); ny = size(vmod_slice,2)
       !write(0,*) "nx,ny=",nx,":",ny
       allocate(vslice1d(nx*ny))!,vslice1d_2(nx*ny))
       do iy=1,ny
       do ix=1,nx
         vslice1d((iy-1)*nx+ix) = vmod_slice(ix,iy) 
       end do
       end do
     ! !If there is a big gap between values within [min,max] and the two end values [min,max], this would remove the gap, so the vmin_sec and vmax_sec can better represent the values in between, this does not work, because the bins are nearly zero but not exactly zero. Use quantile instead, 99%
     !  b_gap_vmin = .true.; b_gap_vmax = .true.
     !  dv_min = 100.; dv_max = 200.; qntl = 0.99
     !  nsel1 = count(vslice1d>vmin_pri+dv_min)
     !  nsel2 = count(vslice1d<vmax_pri-dv_max)
     !  k1 = int(nsel1*(1.-qntl)+0.5); k2 = int(nsel2*(qntl)+0.5)
     !  if (k1<=1) then
     !    b_gap_vmin = .false.; vmin_sec = vmin_pri
     !  end if
     !  if (k2<=1) then
     !    b_gap_vmax = .false.; vmax_sec = vmax_pri
     !  end if
     !  
     !  !write(0,*) nsel1,":",nsel2,";",k1,":",k2,";",nx*ny,":", vmax_pri,vmin_pri, "; isNaN=",count(isNaN(vslice1d))
     !  
     !  if (b_gap_vmax) then !try to detect if there is actually a gap 
     !    ndepth = 0; nsztt = 0
     !    vmax_sec = quantile(k2,pack(vslice1d, vslice1d<vmax_pri-dv_max))
     !    !vmax_sec = vmax_pri*0.9
     !    !Detect if there is gap
     !    if (vmax_sec - vmax_pri > -2*dv_max) then !There seems no gap on the vmax end
     !      b_gap_vmax = .false.
     !    end if
     !  end if
     !  if (b_gap_vmin) then !try to detect if there is actually a gap 
     !    !vslice1d_2(1:nsel1) = pack(vslice1d, vslice1d>vmin_pri+dv_min)
     !    ndepth = 0; nsztt = 0
     !    vmin_sec = quantile(k1,pack(vslice1d, vslice1d>vmin_pri+dv_min))
     !    !Detect if there is gap
     !    if (vmin_sec -vmin_pri < +3*dv_min) then !There seems no gap on the vmin end
     !      b_gap_vmin = .false.
     !    end if
     !  end if
     !  
     !  if (vmin_sec >= vmax_sec) then
     !    t = vmin_sec
     !    vmin_sec = vmax_sec
     !    vmax_sec = t
     !  end if
     !  if (b_gap_vmin .and. b_gap_vmax) then
     !    write(0,*) "vmin,vmin_sec,vmax_sec,vmax:",b_gap_vmin,":",b_gap_vmax," ; ", vmin_pri,":",vmin_sec," : ",vmax_sec,":",vmax_pri
     !  end if

     !  nvrf_sec         = nvrf - 2 
     ! 
     !  if ((.not. b_gap_vmin) .and. (.not. b_gap_vmax)) then
     !    vmin_sec = (vmax_pri-vmin_pri)/(nvrf-1)  + vmin_pri
     !    vmax_sec = -(vmax_pri-vmin_pri)/(nvrf-1) + vmax_pri
     !  else if ( (b_gap_vmin) .and. (.not. b_gap_vmax)) then !only vmin has a gap
     !    vmax_sec = -(vmax_pri-vmin_sec)/(nvrf-2) + vmax_pri
     !  else if ( .not.(b_gap_vmin) .and. (b_gap_vmax)) then  !only vmax has a gap
     !    vmin_sec = (vmax_sec-vmin_pri)/(nvrf-2)  + vmin_pri
     !  end if

     !  dvel=(vmax_sec-vmin_sec)/(nvrf_sec-1)
     !  do iref=1, nvrf_sec
     !     vref_slice(iref+1)=vmin_sec+(iref-1)*dvel
     !  end do

      ! Use Quantile to determine the vref.
       do iref=2, nvrf-1
         qntl = 1.0/(nvrf-1)*(iref-1)
         k1 = int((nx*ny*qntl)+0.5)
         vref_slice(iref) = quantile(k1,vslice1d)
       end do
       deallocate(vslice1d)
       write(0,*) "vrefs by quantile:", vref_slice

       !vmin_sec = minval(vmod_slice, MASK=vmod_slice-vmin_pri>+50.)
       !vmax_sec = maxval(vmod_slice, MASK=vmod_slice-vmax_pri<-100.)

       !if (abs(vmin_sec)<vmin_pri .or. abs(vmin_sec)>vmax_pri) vmin_sec = vmin_pri
       !if (abs(vmax_sec)<vmin_pri .or. abs(vmax_sec)>vmax_pri) vmax_sec = vmax_pri

       !vref_slice(1)    = vmin_pri
       !vref_slice(nvrf) = vmax_pri
       !nvrf_sec         = nvrf - 2 

       !dvel=(vmax_sec-vmin_sec)/(nvrf_sec-1)
       !do iref=1, nvrf_sec
       !   vref_slice(iref+1)=vmin_sec+(iref-1)*dvel
       !end do
       !write(0,*) "vrefs by uniform:", vref_slice
    end if
    stat = 0 
  end function
end module

