#include "mpi.h"
#include "string.h"
#include <stdio.h>
#include <math.h>
#include "seplib.h"
#include "bwi-files-mpi.c"
#include "bwi-commd-mpi.c"

#define LMAX_DIR 256
#define LMAX_FNS 256
#define LMAX_CMD 1024


void ssystem(char* cmd) {
  fprintf("executing cmd: %s \n", cmd);
}

int main(int argc, char* argv[]) {
char bindir[LMAX_DIR], datdir[LMAX_DIR], resdir[LMAX_DIR], tmpdir[LMAX_DIR], datapath[LMAX_DIR];
char fncsou[LMAX_FNS], fncrec[LMAX_FNS], fnbvel[LMAX_FNS], fnbimg[LMAX_FNS];
char fncsou_orig[LMAX_FNS], fncrec_orig[LMAX_FNS], fnbvel_orig[LMAX_FNS], fnbimg_orig[LMAX_FNS];
char fngrdo_orig[LMAX_FNS], fngrdp_orig[LMAX_FNS], fnsrch_orig[LMAX_FNS], fnfobj_orig[LMAX_FNS];
char fncsou_temp[LMAX_FNS], fncrec_temp[LMAX_FNS], fnbvel_temp[LMAX_FNS], fnbimg_temp[LMAX_FNS];
char fngrdo_temp[LMAX_FNS], fngrdp_temp[LMAX_FNS], fnsrch_temp[LMAX_FNS], fnfobj_temp[LMAX_FNS];
char cpy_csou[LMAX_CMD], cpy_crec[LMAX_CMD], cpy_bvel[LMAX_CMD], cpy_bimg[LMAX_CMD];
char cpy_grdo[LMAX_CMD], cpy_grdp[LMAX_CMD], cpy_srch[LMAX_CMD], cpy_fobj[LMAX_CMD];
char cpy_mmov[LMAX_CMD], cpy_fmov[LMAX_CMD];
char cmd_bornmodl[LMAX_CMD];
char par_bornmodl[LMAX_CMD];
char suffix[20], shot_string[24], numstr[4];
char nsx_string[20], dsx_string[20], osx_string[20];
char nsy_string[20], dsy_string[20], osy_string[20];

axis asx={1, 1., 0.}, asy={1, 1., 0.};

int irank, nrank, nsb;

MPI_Init(&argc,&argv);
MPI_Comm_rank(MPI_COMM_WORLD, &irank);
MPI_Comm_size(MPI_COMM_WORLD, &nrank);

initpar(argc,argv);
doc(" ");

if (0==getch("bindir","s",&bindir)) strcpy(bindir,"./");
if (0==getch("datdir","s",&datdir)) strcpy(datdir,"./");
if (0==getch("resdir","s",&resdir)) strcpy(resdir,"./");
if (0==getch("tmpdir","s",&tmpdir)) strcpy(tmpdir,"./");

if (0==getch("suffix","s",&suffix)) strcpy(suffix,"./");

if (0==getch("fncsou","s",&fncsou)) strcpy(fncsou,"CSOU.H");
if (0==getch("fncrec","s",&fncrec)) strcpy(fncrec,"CREC.H");
if (0==getch("fnbvel","s",&fnbvel)) strcpy(fnbvel,"BVEL.H");
if (0==getch("fnbimg","s",&fnbimg)) strcpy(fnbimg,"BIMG.H");

if (0==getch("par_bornmodl","s",&par_bornmodl)) strcpy(par_bornmodl," ");

if (0==getch("nsb","i",&nsb)) nsb = 1  ;

if (nrank!=nsb){
  fprintf(stderr, "ERROR: nrank != nsb,  %d %d \n", nrank, nsb);
  MPI_FINALIZE();
  exit(-1);
}

sprintf(numstr, "%d", irank+1);
strcpy(shot_string, suffix); strcat(shot_string, numstr);

strcpy(fnbimg_orig, resdir); strcat(fnbimg_orig, fnbimg); 
strcpy(fnbimg_temp, tmpdir); strcat(fnbimg_temp, fnbimg); strcat(fnbimg_temp, shot_string);

strcpy(fncsou_orig, datdir); strcat(fncsou_orig, fncsou); strcat(fncsou_orig, shot_string);
strcpy(fncsou_temp, tmpdir); strcat(fncsou_temp, fncsou); strcat(fncsou_temp, shot_string);

strcpy(fncrec_orig, datdir); strcat(fncrec_orig, fncrec); strcat(fncrec_orig, shot_string);
strcpy(fncrec_temp, tmpdir); strcat(fncrec_temp, fncrec); strcat(fncrec_temp, shot_string);

strcpy(fnbvel_orig, datdir); strcat(fnbvel_orig, fnbvel);
strcpy(fnbvel_temp, tmpdir); strcat(fnbvel_temp, fnbvel); strcat(fnbvel_temp, shot_string);

////////////////////////////
// copy source to all nodes
////////////////////////////
create_cpy_one2all(cpy_csou, tmpdir, fncsou_orig, fncsou_temp);
create_cpy_one2all(cpy_bimg, tmpdir, fnbimg_orig, fnbimg_temp);
create_cpy_one2all(cpy_bvel, tmpdir, fnbvel_orig, fnbvel_temp);

create_cpy_all2one(cpy_crec, fncrec_orig, fncrec_temp);

// copy sources
if (0!=system(cpy_csou)) {fprintf(stderr, "irank=%d, %s\n", irank, cpy_csou); MPI_Finalize(); exit(1);}

// copy images
if (0!=system(cpy_bimg)) {fprintf(stderr, "irank=%d, %s\n", irank, cpy_bimg); MPI_Finalize(); exit(1);}

// copy background velocity
if (0!=system(cpy_bvel)) {fprintf(stderr, "irank=%d, %s\n", irank, cpy_bvel); MPI_Finalize(); exit(1);}

// read source parameters
auxinout(fncsou_temp);
hinaxis(fncsou_temp, &asx, 4);
hinaxis(fncsou_temp, &asy, 5);
auxclose(fncsou_temp);

sprintf(nsx_string, "%d", asx.n); sprintf(dsx_string, "%f", asx.d); sprintf(osx_string, "%f", asx.o);
sprintf(nsy_string, "%d", asy.n); sprintf(dsy_string, "%f", asy.d); sprintf(osy_string, "%f", asy.o);

strcat(par_bornmodl, " nsx="); strcat(par_bornmodl, nsx_string); 
strcat(par_bornmodl, " dsx="); strcat(par_bornmodl, dsx_string);
strcat(par_bornmodl, " osx="); strcat(par_bornmodl, osx_string);

strcat(par_bornmodl, " nsy="); strcat(par_bornmodl, nsy_string);
strcat(par_bornmodl, " dsy="); strcat(par_bornmodl, dsy_string);
strcat(par_bornmodl, " osy="); strcat(par_bornmodl, osy_string);

create_cmd_bornmodl(cmd_bornmodl, bindir, tmpdir, fncsou_temp, fncrec_temp, fnbvel_temp, fnbimg_temp, par_bornmodl);

// compute background image
if (0!=system(cmd_bornmodl)) {fprintf(stderr, "irank=%d, %s\n", irank, cmd_bornmodl); MPI_Finalize(); exit(1);}

// copy modeled data back
if (0!=system(cpy_crec)) {fprintf(stderr, "irank=%d, %s\n", irank, cpy_crec); MPI_Finalize(); exit(1);}

if (irank==0){
   fprintf(stderr, "********************************\n");
   fprintf(stderr, " FINISHED BORN MODELING         \n");
   fprintf(stderr, "********************************\n");
}

MPI_Finalize();

exit(0);

}




