!---------------------------------------------
! One-way wave-equation shot-profile migration
!---------------------------------------------
! Author : Yaxun Tang
! Version: 2.0
! Date   : Jan. 1, 2008
program main
  use bwi_kernel_mod
  use bwi_splits_mod
  implicit none
  include 'mpif.h'

  type(dat_grid) :: dgrid
  type(pro_grid) :: pgrid
  type(vel_grid) :: vgrid
  type(img_grid) :: igrid
  type(hes_grid) :: hgrid
  type(bwi_ftag) :: ftag_ 
  type(block)    :: sblock
  type(axis)     :: as, ass
  logical        :: adj, add 
  integer        :: stat

  complex, dimension(:,:), allocatable :: crec
  real   , dimension(:,:), allocatable :: hsou
  type(axis)         :: agx, agy, aw, asx, asy
  integer            :: ntrg, nshg, ntrs, nshs, ish, isb, iss
  logical            :: comsou
  character(len=256) :: csou_, crec_ 
  character(len=256) :: csou_split_, crec_split_
  
  integer :: nrank, irank, ierr, rc

  call MPI_INIT(ierr)
  if (ierr .ne. MPI_SUCCESS) then
     write(0,*)"Error starting MPI program. Terminating."
     call MPI_ABORT(MPI_COMM_WORLD, rc, ierr)
  end if

  call MPI_COMM_RANK(MPI_COMM_WORLD, irank, ierr)
  call MPI_COMM_SIZE(MPI_COMM_WORLD, nrank, ierr)

  call sep_init()
  call getch_add_string("head=/dev/null")
  call getch_add_string("noheader=y")
  call set_no_putch()
  call sep_begin_prog()

  call from_param("add", add, .false.)


  as%n = nrank-1
  as%d = 1
  as%o = 1 
  call base_blkcal(sblock, as)

  if (adj) then
!    migration, distribute sources, receivers and velocity
     call auxinout(csou_)
     call auxinout(crec_)
     call auxinout(csou_split_)
     call auxinout(crec_split_)
     call bwi_splits_imgadj(hsou, agx, agy, aw, asx, asy, ntrg, nshg, ntrs, nshs, comsou, csou_, crec_)
     allocate(crec(ntrg,aw%n))
     ish = 0

!    send the binary data
!    --------------------
     do isb=1, sblock%nb
        ass%o = sblock%os(isb); ass%d = sblock%ds; ass%n = sblock%blen(isb) 
        
        do iss=1, ass%n
!       absolute shot index
           ish =  ish + 1
           if (irank==0) then 
              call bwi_splits_iocrec(crec, ntrg, aw%n, nshg, ish, crec_      , "reed")
              call MPI_SEND(crec, ntrg*aw%n, MPI_COMPLEX, isb, ish, MPI_COMM_WORLD, ierr)
           else
              call MPI_RECV(crec, ntrg*aw%n, MPI_COMPLEX, isb, ish, MPI_COMM_WORLD, stat, ierr)
              call bwi_splits_iocrec(crec, ntrg, aw%n, nshg, ish, crec_split_, "rite")
           end if
        end do
     end do

!    create the header files
!    -----------------------
     if (irank/=0) then
        call houaxis(crec_split_, agx, 1)
        call houaxis(crec_split_, agy, 2)
        call houaxis(crec_split_, aw , 3)
        call houaxis(crec_split_, ass, 4) 
     end if 

     deallocate(crec)
  else
!    modeling, distribute sources, images and velocity
  end if




!  stat = bwi_kernel_init(dgrid, pgrid, vgrid, igrid, hgrid, ftag_)
!
!  if (pgrid%mode=="imgfwd" .or. pgrid%mode=="imgadj") then
!     if (pgrid%mode=="imgfwd") adj = .false.
!     if (pgrid%mode=="imgadj") adj = .true.
!
!     stat = bwi_kerimg_oper(adj, add, ftag_%csou_, ftag_%crec_, ftag_%bvel_, ftag_%bimg_)
!
!  else
!     write(0,*)"ERROR: WRONG MODE"
!     STOP
!  end if

  write(0,*)"***********************"
  write(0,*)"      Finished!!!      "
  write(0,*)"***********************"

  call sep_end_prog()

  call exit()

end program
