program main
  use sep
  use base_struct_mod
  use base_ioutil_mod
  use base_random_mod
  implicit none
  complex, dimension(:,:), allocatable :: efns, efnr
  real   , dimension(:,:), allocatable :: efns_tsft, efnr_tsft
  real   , dimension(:,:), allocatable :: rts, tfns
  real   , dimension(:,:), allocatable :: rtr, tfnr
  type(axis)                           :: asx, asy, arx, ary, aws
  type(axis)                           :: apsx, apsy, asx_chk, asy_chk
  type(axis)                           :: aprx, apry, arx_chk, ary_chk
  real                                 :: psx, psy, sx0, sy0
  real                                 :: prx, pry, rx0, ry0
  real                                 :: sx0_pos, sx0_neg, sy0_pos, sy0_neg
  real                                 :: rx0_pos, rx0_neg, ry0_pos, ry0_neg
  real                                 :: psx_min, psx_max, psy_min, psy_max
  real                                 :: prx_min, prx_max, pry_min, pry_max
  integer                              :: iws, isx, isy, ipsx, ipsy, rdseed
  integer                              :: irx, iry, iprx, ipry
  real                                 :: phase, sx, sy, rx, ry, ws, PI 
  real                                 :: rt_min, rt_max
  logical                              :: wntsft
  character(len=6)                     :: domain, enctyp
  character(len=256)                   :: efns_, efnr_, tfns_, tfnr_
  character(len=256)                   :: efns_tsft_, efnr_tsft_ 

  PI = 3.14159265359

  call sep_init(SOURCE)
  call getch_add_string("head=/dev/null")
  call getch_add_string("noheader=y")
  call set_no_putch()
  call sep_begin_prog()

  efns_ = "efns"
  efnr_ = "efnr"
  tfns_ = "tfns"
  tfnr_ = "tfnr"
  efns_tsft_ = "efns_tsft"
  efnr_tsft_ = "efnr_tsft"

  call from_param("domain", domain, "encsou")
  call from_param("enctyp", enctyp, "planes")
  call from_param("wntsft", wntsft, .false. )
  call from_param("nws"   , aws%n , 1       )
  call from_param("dws"   , aws%d , 1.      )
  call from_param("ows"   , aws%o , 0.      )

  if (domain=="encsou") then
     call from_param("nsx" , asx%n , 1); call from_param("dsx" , asx%d , 1.); call from_param("osx" , asx%o , 0.)
     call from_param("nsy" , asy%n , 1); call from_param("dsy" , asy%d , 1.); call from_param("osy" , asy%o , 0.)
     call from_param("psx_min", psx_min, 0.); call from_param("psx_max", psx_max, 0.); call from_param("npsx", apsx%n, 1)
     call from_param("psy_min", psy_min, 0.); call from_param("psy_max", psy_max, 0.); call from_param("npsy", apsy%n, 1)
     apsx%o = psx_min; apsx%d = 1.
     apsy%o = psy_min; apsy%d = 1.
     if (apsx%n>1) apsx%d = (psx_max-psx_min)/(apsx%n-1)
     if (apsy%n>1) apsy%d = (psy_max-psy_min)/(apsy%n-1) 
     allocate(efns(asx%n,asy%n))
     if (wntsft) allocate(efns_tsft(asx%n,asy%n))
  end if

  if (domain=="encrec") then
     call from_param("nrx", arx%n, 1); call from_param("drx", arx%d, 1.); call from_param("orx", arx%o, 0.)
     call from_param("nry", ary%n, 1); call from_param("dry", ary%d, 1.); call from_param("ory", ary%o, 0.)
     call from_param("prx_min", prx_min, 0.); call from_param("prx_max", prx_max, 0.); call from_param("nprx", aprx%n, 1)
     call from_param("pry_min", pry_min, 0.); call from_param("pry_max", pry_max, 0.); call from_param("npry", apry%n, 1)
     aprx%o = prx_min; aprx%d = 1.
     apry%o = pry_min; apry%d = 1.
     if (aprx%n>1) aprx%d = (prx_max-prx_min)/(aprx%n-1)
     if (apry%n>1) apry%d = (pry_max-pry_min)/(apry%n-1) 
     allocate(efnr(arx%n,ary%n))
     if (wntsft) allocate(efnr_tsft(arx%n,ary%n))
  end if 

  if (domain=="encsou") then
     if (enctyp=="planes") then                            !!.. plane-wave encoding
        call from_param("sx0_pos", sx0_pos, 0.)
        call from_param("sy0_pos", sy0_pos, 0.)
        call from_param("sx0_neg", sx0_neg, 0.)
        call from_param("sy0_neg", sy0_neg, 0.)
     else if (enctyp=="randws".or.enctyp=="randts".or.enctyp=="binary") then  !!.. random encoding
        sx0_pos = 0.; sy0_pos = 0.
        sx0_neg = 0.; sy0_neg = 0.
        call from_param("rt_min", rt_min, 0.1 )
        call from_param("rt_max", rt_max, 1.0 ) 
        call from_param("rdseed", rdseed, 1234)
     else if (enctyp=="anythg") then
        sx0_pos = 0.; sy0_pos = 0.                       !!.. user defined encoding
        sx0_neg = 0.; sy0_neg = 0.
        allocate(tfns(asx%n,asy%n))
        call auxin(tfns_)
        call hinaxis(tfns_, asx_chk, 1)
        call hinaxis(tfns_, asy_chk, 2)
        if (asx_chk%n/=asx%n .or. asx_chk%d/=asx%d .or. asx_chk%o/=asx%o) call seperr("asx_chk and asx do not agree")
        if (asy_chk%n/=asy%n .or. asy_chk%d/=asy%d .or. asy_chk%o/=asy%o) call seperr("asy_chk and asy do not agree") 
        call sep_read(tfns, tfns_)
     end if

     call auxinout(efns_)   
     call houaxis(efns_, asx , 1)
     call houaxis(efns_, asy , 2)
     call houaxis(efns_, aws , 3)
     call houaxis(efns_, apsx, 4)
     call houaxis(efns_, apsy, 5)
     call auxputch("esize" , "i", 8     , efns_)
     call auxputch("label1", "s", "sx"  , efns_)
     call auxputch("label2", "s", "sy"  , efns_)
     call auxputch("label3", "s", "freq", efns_)
     call auxputch("label4", "s", "psx" , efns_)
     call auxputch("label5", "s", "psy" , efns_)

     if (wntsft) then
        call auxinout(efns_tsft_)
        call houaxis(efns_tsft_, asx , 1)
        call houaxis(efns_tsft_, asy , 2)
        call houaxis(efns_tsft_, apsx, 3)
        call houaxis(efns_tsft_, apsy, 4)
        call auxputch("label1", "s", "sx"  , efns_tsft_)
        call auxputch("label2", "s", "sy"  , efns_tsft_)
        call auxputch("label3", "s", "psx" , efns_tsft_)
        call auxputch("label4", "s", "psy" , efns_tsft_)
     end if

     call random_seed() 

     if (enctyp=="randts".or.enctyp=="binary") allocate(rts(asx%n,asy%n))

     do ipsy=1, apsy%n; psy=apsy%o+(ipsy-1)*apsy%d; sy0=sy0_pos; if(psy<0.) sy0=sy0_neg
     do ipsx=1, apsx%n; psx=apsx%o+(ipsx-1)*apsx%d; sx0=sx0_pos; if(psx<0.) sx0=sx0_neg
        if (enctyp=="randts") then
           do isy=1, asy%n
           do isx=1, asx%n
              call random_number(rts(isx,isy))
              rts(isx,isy) = rts(isx,isy)*(rt_max-rt_min)+rt_min
           end do
           end do
           write(0,*)"ts max", maxval(rts), "ts min", minval(rts)
        end if

        if (enctyp=="binary") then
           do isy=1, asy%n
           do isx=1, asx%n
              call random_number(rts(isx,isy))
              if (rts(isx,isy)<0.5) then
                 rts(isx,isy) = -1
              else
                 rts(isx,isy) = +1
              end if
           end do
           end do
        end if

        do iws=1, aws%n; ws=aws%o+(iws-1)*aws%d 
        do isy=1, asy%n; sy=asy%o+(isy-1)*asy%d
        do isx=1, asx%n; sx=asx%o+(isx-1)*asx%d
           if (enctyp=="planes") then
              phase = ws*(psx*(sx-sx0) + psy*(sy-sy0))
           else if (enctyp=="randws") then
              call random_number(phase)
              phase = phase*2*PI
           else if (enctyp=="randts") then
              phase = ws*rts(isx,isy)
           else if (enctyp=="anythg") then
              phase = ws*tfns(isx,isy)
           end if

           if (enctyp=="binary") then
              efns(isx,isy) = cmplx(rts(isx,isy),0.)
           else
              efns(isx,isy) = cmplx(cos(phase),sin(phase))
           end if

           if (wntsft) then
              if (enctyp/="randws") then
                 efns_tsft(isx,isy) = phase/ws
              else
                 efns_tsft(isx,isy) = phase
              end if
           end if 
        end do
        end do
        call sep_write(efns, efns_, esize=8)
        if (wntsft.and.iws==1) call sep_write(efns_tsft, efns_tsft_, esize=4)
        end do
     end do
     end do
  end if

  if (domain=="encrec") then
     if (enctyp=="planes") then                            !!.. plane-wave encoding
        call from_param("rx0_pos", rx0_pos, 0.)
        call from_param("ry0_pos", ry0_pos, 0.)
        call from_param("rx0_neg", rx0_neg, 0.)
        call from_param("ry0_neg", ry0_neg, 0.)
     else if (enctyp=="randws".or.enctyp=="randts".or.enctyp=="binary") then  !!.. random encoding
        rx0_pos = 0.; ry0_pos = 0.
        rx0_neg = 0.; ry0_neg = 0.
        call from_param("rt_min", rt_min, 0.1 )
        call from_param("rt_max", rt_max, 1.0 )
        call from_param("rdseed", rdseed, 1234)
     else if (enctyp=="anythg") then
        rx0_pos = 0.; ry0_pos = 0.                       !!.. user defined encoding
        rx0_neg = 0.; ry0_neg = 0.
        allocate(tfnr(arx%n,ary%n))
        call auxin(tfns_)
        call hinaxis(tfnr_, arx_chk, 1)
        call hinaxis(tfnr_, ary_chk, 2)
        if (arx_chk%n/=arx%n .or. arx_chk%d/=arx%d .or. arx_chk%o/=arx%o) call seperr("asx_chk and asx do not agree")
        if (ary_chk%n/=ary%n .or. ary_chk%d/=ary%d .or. ary_chk%o/=ary%o) call seperr("asy_chk and asy do not agree")
        call sep_read(tfnr, tfnr_)
     end if

     call auxinout(efnr_)
     call houaxis(efnr_, arx , 1)
     call houaxis(efnr_, ary , 2)
     call houaxis(efnr_, aws , 3)
     call houaxis(efnr_, aprx, 4)
     call houaxis(efnr_, apry, 5)
     call auxputch("esize" , "i", 8     , efnr_)
     call auxputch("label1", "s", "rx"  , efnr_)
     call auxputch("label2", "s", "ry"  , efnr_)
     call auxputch("label3", "s", "freq", efnr_)
     call auxputch("label4", "s", "prx" , efnr_)
     call auxputch("label5", "s", "pry" , efnr_)


     if (wntsft) then
        call auxinout(efnr_tsft_)
        call houaxis(efnr_tsft_, asx , 1)
        call houaxis(efnr_tsft_, asy , 2)
        call houaxis(efnr_tsft_, apsx, 3)
        call houaxis(efnr_tsft_, apsy, 4)
        call auxputch("label1", "s", "sx"  , efnr_tsft_)
        call auxputch("label2", "s", "sy"  , efnr_tsft_)
        call auxputch("label3", "s", "psx" , efnr_tsft_)
        call auxputch("label4", "s", "psy" , efnr_tsft_)
     end if


     call random_seed()

     if (enctyp=="randts") allocate(rtr(arx%n,ary%n))

     do ipry=1, apry%n; pry=apry%o+(ipry-1)*apry%d; ry0=ry0_pos; if(pry<0.) ry0=ry0_neg
     do iprx=1, aprx%n; prx=aprx%o+(iprx-1)*aprx%d; rx0=rx0_pos; if(prx<0.) rx0=rx0_neg
        if (enctyp=="randts") then
           do iry=1, ary%n
           do irx=1, arx%n
              call random_number(rtr(irx,iry))
              rtr(irx,iry) = rtr(irx,iry)*(rt_max-rt_min)+rt_min
           end do
           end do
           write(0,*)"t max", maxval(rtr), "t min", minval(rtr)
        end if

        if (enctyp=="binary") then
           do iry=1, ary%n
           do irx=1, arx%n
              call random_number(rtr(irx,iry))
              if (rtr(irx,iry)<0.5) then
                 rtr(irx,iry) = -1
              else
                 rtr(irx,iry) = +1
              end if 
           end do
           end do
        end if

        do iws=1, aws%n; ws=aws%o+(iws-1)*aws%d
        do iry=1, ary%n; ry=ary%o+(iry-1)*ary%d
        do irx=1, arx%n; rx=arx%o+(irx-1)*arx%d
           if (enctyp=="planes") then
              phase = ws*(prx*(rx-rx0) + pry*(ry-ry0))
           else if (enctyp=="randws") then
              call random_number(phase)
              phase = phase*2*PI
           else if (enctyp=="randts") then
              phase = ws*rtr(irx,iry)
           else if (enctyp=="anythg") then
              phase = ws*tfnr(irx,iry)
           end if

           if (enctyp=="binary") then
              efnr(irx,iry) = cmplx(rtr(irx,iry),0.)
           else
              efnr(irx,iry) = cmplx(cos(phase),sin(phase))
           end if
           if (wntsft) then
              if (enctyp/="randws") then
                 efnr_tsft(irx,iry) = phase/ws
              else
                 efnr_tsft(irx,iry) = phase
              end if
           end if  
        end do
        end do
        call sep_write(efnr, efnr_, esize=8)
        if (wntsft.and.iws==1) call sep_write(efnr_tsft, efnr_tsft_, esize=4)
        end do  
     end do   
     end do

  end if

end program

