module hesdct_fileio_mod
  use sep
  use hesdct_struct_mod
  implicit none

  integer, external, private :: sreed_window, srite_window

contains
!====================================================================================================================
  subroutine hesdct_fileio_init(filetag, hgrid)
    character(len=*)                :: filetag
    type(hes_grid)                  :: hgrid
    real, dimension(:), allocatable :: rtmp
    integer                         :: itr, ntr
    allocate(rtmp(hgrid%ax_modl%n*hgrid%ay_modl%n*hgrid%az_modl%n))
    ntr = hgrid%ax_offd%n*hgrid%ay_offd%n*hgrid%az_offd%n
    rtmp = 0.
    do itr=1, ntr
       call sep_write(rtmp, filetag)
    end do
    deallocate(rtmp)
    return
  end subroutine
!====================================================================================================================
  subroutine hesdct_fileio_mask(mask, amask, filetag, flag)
    real, dimension(:,:,:,:) :: mask
    type(acq_mask)              :: amask
    character(len=*)            :: filetag
    character(len=4)            :: flag

    integer, dimension(4)       :: n_cube, j_wind, n_wind, f_wind
    integer                     :: n_dim_cube, err

    n_dim_cube = 4
    n_cube = (/amask%asx%n,amask%asy%n,amask%arx%n,amask%ary%n/)
    j_wind = (/1, 1, 1, 1/)
    n_wind = (/amask%asx%n,amask%asy%n,amask%arx%n,amask%ary%n/)
    f_wind = (/0, 0, 0, 0/)

    if (flag=="rite") then
       err = srite_window(filetag,n_dim_cube,n_cube,n_wind,f_wind,j_wind,4,mask)
       mask = 0
    else if (flag=="reed") then
       mask = 0
       err = sreed_window(filetag,n_dim_cube,n_cube,n_wind,f_wind,j_wind,4,mask)
    end if
    return
  end subroutine
!====================================================================================================================
  subroutine hesdct_fileio_hess(hess, hgrid, ix_modl, iy_modl, iz_modl, filetag, flag)
    real, dimension(:,:,:) :: hess
    type(hes_grid)         :: hgrid
    integer                :: ix_modl, iy_modl, iz_modl
    character(len=*)       :: filetag
    character(len=4)       :: flag

    integer, dimension(6)  :: n_cube, j_wind, n_wind, f_wind
    integer                :: n_dim_cube, err

    n_dim_cube = 6
    n_cube = (/hgrid%ax_offd%n, hgrid%ay_offd%n, hgrid%az_offd%n, &
               hgrid%ax_modl%n, hgrid%ay_modl%n, hgrid%az_modl%n/)
    j_wind = (/1, 1, 1, 1, 1, 1/)
    n_wind = (/hgrid%ax_offd%n, hgrid%ay_offd%n, hgrid%az_offd%n, 1, 1, 1/)
    f_wind = (/0, 0, 0, ix_modl-1, iy_modl-1, iz_modl-1/)

    if (flag == "rite") then
       err = srite_window(filetag,n_dim_cube,n_cube,n_wind,f_wind,j_wind,4,hess)
    else if (flag == "reed") then
       hess = 0.
       err = sreed_window(filetag,n_dim_cube,n_cube,n_wind,f_wind,j_wind,4,hess)
    else
       call seperr("you have to specify the flag for wei_hessian_io_hessian()")
    end if
    return
  end subroutine

  subroutine hesdct_fileio_hess2(hess, hgrid, ix_offd, iy_offd, iz_offd, filetag, flag)
    real, dimension(:,:,:) :: hess
    type(hes_grid)         :: hgrid
    integer                :: ix_offd, iy_offd, iz_offd
    character(len=*)       :: filetag
    character(len=4)       :: flag

    integer, dimension(6)  :: n_cube, j_wind, n_wind, f_wind
    integer                :: n_dim_cube, err

    n_dim_cube = 6
    n_cube = (/hgrid%ax_modl%n, hgrid%ay_modl%n, hgrid%az_modl%n, &
               hgrid%ax_offd%n, hgrid%ay_offd%n, hgrid%az_offd%n/)
    j_wind = (/1, 1, 1, 1, 1, 1/)
    n_wind = (/hgrid%ax_modl%n, hgrid%ay_modl%n, hgrid%az_modl%n, 1, 1, 1/)
    f_wind = (/0, 0, 0, ix_offd-1, iy_offd-1, iz_offd-1/)

    if (flag == "rite") then
       err = srite_window(filetag,n_dim_cube,n_cube,n_wind,f_wind,j_wind,4,hess)
    else if (flag == "reed") then
       hess = 0.
       err = sreed_window(filetag,n_dim_cube,n_cube,n_wind,f_wind,j_wind,4,hess)
    else
       call seperr("you have to specify the flag for wei_hessian_io_hessian()")
    end if
    return
  end subroutine

!-------------------------------------------------------------------------------------------------------------
  subroutine hesdct_fileio_grns(grns, ggrid, nws, iws, filetag, flag)
    complex, dimension(:,:,:,:,:,:) :: grns
    type(gre_grid)                  :: ggrid
    integer                         :: nws, iws
    character(len=*)                :: filetag
    character(len=4)                :: flag

    integer, dimension(6)           :: n_cube, j_wind, n_wind, f_wind
    integer                         :: n_dim_cube, err

    n_dim_cube = 6
    n_cube = (/ggrid%ax_surf%n, ggrid%ay_surf%n, ggrid%ax_modl%n, ggrid%ay_modl%n, ggrid%az_modl%n, ggrid%aw%n/)
    j_wind = (/1, 1, 1, 1, 1, 1/)
    n_wind = (/ggrid%ax_surf%n, ggrid%ay_surf%n, ggrid%ax_modl%n, ggrid%ay_modl%n, ggrid%az_modl%n, nws/)
    f_wind = (/0, 0, 0, 0, 0, iws-1/)

    if (flag == "rite") then
       err = srite_window(filetag,n_dim_cube,n_cube,n_wind,f_wind,j_wind,8,grns)
       grns = 0.
    else if (flag == "reed") then
       grns = 0.
       err = sreed_window(filetag,n_dim_cube,n_cube,n_wind,f_wind,j_wind,8,grns)
    else
       call seperr("you have to specify the flag for grns")
    end if
    return
  end subroutine

  subroutine hesdct_fileio_efns(efns, dgrid, nws, iws, filetag, flag)
    complex, dimension(:,:,:) :: efns
    type(dat_grid)            :: dgrid
    integer                   :: nws, iws
    character(len=*)          :: filetag
    character(len=4)          :: flag
    integer, dimension(3)     :: n_cube, j_wind, n_wind, f_wind
    integer                   :: n_dim_cube, err

    n_dim_cube = 3
    n_cube = (/dgrid%asx%n, dgrid%asy%n, dgrid%aw%n/)
    j_wind = (/1, 1, 1/)
    n_wind = (/dgrid%asx%n, dgrid%asy%n, nws/)
    f_wind = (/0, 0, iws-1/)

    if (flag == "rite") then
       err = srite_window(filetag,n_dim_cube,n_cube,n_wind,f_wind,j_wind,8,efns)
       efns = 0.
    else if (flag == "reed") then
       efns = 0.
       err = sreed_window(filetag,n_dim_cube,n_cube,n_wind,f_wind,j_wind,8,efns)
    else
       call seperr("you have to specify the flag for grns")
    end if
    return
  end subroutine

  subroutine hesdct_fileio_efnr(efnr, dgrid, nws, iws, filetag, flag)
    complex, dimension(:,:,:) :: efnr
    type(dat_grid)            :: dgrid
    integer                   :: nws, iws
    character(len=*)          :: filetag
    character(len=4)          :: flag
    integer, dimension(3)     :: n_cube, j_wind, n_wind, f_wind
    integer                   :: n_dim_cube, err

    n_dim_cube = 3
    n_cube = (/dgrid%agx%n, dgrid%agy%n, dgrid%aw%n/)
    j_wind = (/1, 1, 1/)
    n_wind = (/dgrid%agx%n, dgrid%agy%n, nws/)
    f_wind = (/0, 0, iws-1/)

    if (flag == "rite") then
       err = srite_window(filetag,n_dim_cube,n_cube,n_wind,f_wind,j_wind,8,efnr)
       efnr = 0.
    else if (flag == "reed") then
       efnr = 0.
       err = sreed_window(filetag,n_dim_cube,n_cube,n_wind,f_wind,j_wind,8,efnr)
    else
       call seperr("you have to specify the flag for grns")
    end if
    return
  end subroutine



!===============================================================================================================
  subroutine hesdct_fileio_grns2(grns, ggrid, nws, iws, filetag, flag)
    complex, dimension(:,:,:,:,:,:) :: grns
    type(gre_grid)                  :: ggrid
    integer                         :: nws, iws
    character(len=*)                :: filetag
    character(len=4)                :: flag

    integer, dimension(6)           :: n_cube, j_wind, n_wind, f_wind
    integer                         :: n_dim_cube, err

    n_dim_cube = 6
    n_cube = (/ggrid%ax_modl%n, ggrid%ay_modl%n, ggrid%az_modl%n, ggrid%ax_surf%n, ggrid%ay_surf%n, ggrid%aw%n/)
    j_wind = (/1, 1, 1, 1, 1, 1/)
    n_wind = (/ggrid%ax_modl%n, ggrid%ay_modl%n, ggrid%az_modl%n, ggrid%ax_surf%n, ggrid%ay_surf%n, nws/)
    f_wind = (/0, 0, 0, 0, 0, iws-1/)

    grns = 0.
    if (flag == "rite") then
       err = srite_window(filetag,n_dim_cube,n_cube,n_wind,f_wind,j_wind,8,grns)
    else if (flag == "reed") then
       err = sreed_window(filetag,n_dim_cube,n_cube,n_wind,f_wind,j_wind,8,grns)
    else
       call seperr("you have to specify the flag for grns")
    end if
    return
  end subroutine

!===============================================================================================================
  subroutine hesdct_fileio_csou(csou, dgrid, nws, iws, filetag, flag)
    complex, dimension(:) :: csou
    type(dat_grid)        :: dgrid
    integer               :: nws
    integer               :: iws
    character(len=*)      :: filetag
    character(len=4)      :: flag

    integer, dimension(1) :: n_cube, j_wind, n_wind, f_wind
    integer               :: n_dim_cube, err
    n_dim_cube = 1
    n_cube = (/dgrid%aw%n/)
    j_wind = (/1/)
    n_wind = (/nws/)
    f_wind = (/iws-1/)
    if (flag == "rite") then
       err = srite_window(filetag,n_dim_cube,n_cube,n_wind,f_wind,j_wind,8,csou)
    else if (flag == "reed") then
       err = sreed_window(filetag,n_dim_cube,n_cube,n_wind,f_wind,j_wind,8,csou)
    end if
    return
  end subroutine
end module
