module hesdct_histos_mod
  use sep
  use base_ioutil_mod
  use hesdct_struct_mod
contains

  subroutine hesdct_params_init(dgrid, ggrid, hgrid)
    type(dat_grid) :: dgrid
    type(gre_grid) :: ggrid
    type(hes_grid) :: hgrid
    real           :: xmin_modl, xmax_modl
    real           :: ymin_modl, ymax_modl
    real           :: zmin_modl, zmax_modl
    real           :: xmin_modl_default, xmax_modl_default
    real           :: ymin_modl_default, ymax_modl_default
    real           :: zmin_modl_default, zmax_modl_default

    call from_param("nsx", dgrid%asx%n, 1); call from_param("dsx", dgrid%asx%d, 1.); call from_param("osx", dgrid%asx%o, 0.)
    call from_param("nsy", dgrid%asy%n, 1); call from_param("dsy", dgrid%asy%d, 1.); call from_param("osy", dgrid%asy%o, 0.)

    call from_param("ngx", dgrid%agx%n, 1); call from_param("dgx", dgrid%agx%d, 1.); call from_param("ogx", dgrid%agx%o, 0.)
    call from_param("ngy", dgrid%agy%n, 1); call from_param("dgy", dgrid%agy%d, 1.); call from_param("ogy", dgrid%agy%o, 0.)

    dgrid%nshg = dgrid%asx%n*dgrid%asy%n
    dgrid%ntrg = dgrid%agx%n*dgrid%agy%n

    call from_param("nx_offd",hgrid%ax_offd%n,1); call from_param("dx_offd",hgrid%ax_offd%d,ggrid%ax_modl%d); call from_param("ox_offd",hgrid%ax_offd%o,0.)
    call from_param("ny_offd",hgrid%ay_offd%n,1); call from_param("dy_offd",hgrid%ay_offd%d,ggrid%ay_modl%d); call from_param("oy_offd",hgrid%ay_offd%o,0.)
    call from_param("nz_offd",hgrid%az_offd%n,1); call from_param("dz_offd",hgrid%az_offd%d,ggrid%az_modl%d); call from_param("oz_offd",hgrid%az_offd%o,0.)

    xmin_modl_default = ggrid%ax_modl%o
    xmax_modl_default = ggrid%ax_modl%o + (ggrid%ax_modl%n-1)*ggrid%ax_modl%d
    ymin_modl_default = ggrid%ay_modl%o
    ymax_modl_default = ggrid%ay_modl%o + (ggrid%ay_modl%n-1)*ggrid%ay_modl%d
    zmin_modl_default = ggrid%az_modl%o
    zmax_modl_default = ggrid%az_modl%o + (ggrid%az_modl%n-1)*ggrid%az_modl%d

    call from_param("xmin_modl", xmin_modl, xmin_modl_default)
    call from_param("xmax_modl", xmax_modl, xmax_modl_default)
    call from_param("ymin_modl", ymin_modl, ymin_modl_default)
    call from_param("ymax_modl", ymax_modl, ymax_modl_default)
    call from_param("zmin_modl", zmin_modl, zmin_modl_default)
    call from_param("zmax_modl", zmax_modl, zmax_modl_default)

    xmin_modl = ggrid%ax_modl%o + int(1.5+(xmin_modl-ggrid%ax_modl%o)/ggrid%ax_modl%d-1)*ggrid%ax_modl%d
    ymin_modl = ggrid%ay_modl%o + int(1.5+(ymin_modl-ggrid%ay_modl%o)/ggrid%ay_modl%d-1)*ggrid%ay_modl%d
    zmin_modl = ggrid%az_modl%o + int(1.5+(zmin_modl-ggrid%az_modl%o)/ggrid%az_modl%d-1)*ggrid%az_modl%d

    xmax_modl = ggrid%ax_modl%o + int(1.5+(xmax_modl-ggrid%ax_modl%o)/ggrid%ax_modl%d-1)*ggrid%ax_modl%d
    ymax_modl = ggrid%ay_modl%o + int(1.5+(ymax_modl-ggrid%ay_modl%o)/ggrid%ay_modl%d-1)*ggrid%ay_modl%d
    zmax_modl = ggrid%az_modl%o + int(1.5+(zmax_modl-ggrid%az_modl%o)/ggrid%az_modl%d-1)*ggrid%az_modl%d

    hgrid%ax_modl%o = xmin_modl
    hgrid%ax_modl%d = ggrid%ax_modl%d
    hgrid%ax_modl%n = 1.5 + (xmax_modl-xmin_modl)/hgrid%ax_modl%d

    hgrid%ay_modl%o = ymin_modl
    hgrid%ay_modl%d = ggrid%ay_modl%d
    hgrid%ay_modl%n = 1.5 + (ymax_modl-ymin_modl)/hgrid%ay_modl%d

    hgrid%az_modl%o = zmin_modl
    hgrid%az_modl%d = ggrid%az_modl%d
    hgrid%az_modl%n = 1.5 + (zmax_modl-zmin_modl)/hgrid%az_modl%d



    if (dgrid%aw%n /= ggrid%aw%n) then
       write(0,*)"dgrid%aw%n = ", dgrid%aw%n, " ggrid%aw%n = ", ggrid%aw%n
       call seperr("dgrid%aw%n .ne. ggrid%aw%n error!!!")
    end if

    if (dgrid%aw%d /= ggrid%aw%d) then
       write(0,*)"dgrid%aw%d = ", dgrid%aw%d, " ggrid%aw%d = ", ggrid%aw%d
       call seperr("dgrid%aw%d .ne. ggrid%aw%d error!!!")
    end if

    if (dgrid%aw%o /= ggrid%aw%o) then
       write(0,*)"dgrid%aw%o = ", dgrid%aw%o, " ggrid%aw%o = ", ggrid%aw%o
       call seperr("dgrid%aw%o .ne. ggrid%aw%o error!!!")
    end if
    return
  end subroutine

  subroutine hesdct_histos_grns(ftag_, ggrid)
    type(hesdct_ftag) :: ftag_
    type(gre_grid)    :: ggrid

    call hinaxis(ftag_%grns_, ggrid%ax_surf, 1)
    call hinaxis(ftag_%grns_, ggrid%ay_surf, 2)
    call hinaxis(ftag_%grns_, ggrid%ax_modl, 3)
    call hinaxis(ftag_%grns_, ggrid%ay_modl, 4)
    call hinaxis(ftag_%grns_, ggrid%az_modl, 5)
    call hinaxis(ftag_%grns_, ggrid%aw     , 6)

    ggrid%surf_xmin = ggrid%ax_surf%o
    ggrid%surf_xmax = ggrid%ax_surf%o+(ggrid%ax_surf%n-1)*ggrid%ax_surf%d 
    ggrid%surf_ymin = ggrid%ay_surf%o
    ggrid%surf_ymax = ggrid%ay_surf%o+(ggrid%ay_surf%n-1)*ggrid%ay_surf%d

    return
  end subroutine

  subroutine hesdct_histos_csou(ftag_, dgrid)
    type(hesdct_ftag) :: ftag_
    type(dat_grid)    :: dgrid
    call hinaxis(ftag_%csou_, dgrid%aw, 1)
    return
  end subroutine 

  subroutine hesdct_histos_hess(ftag_, hgrid)
    type(hesdct_ftag) :: ftag_
    type(hes_grid)    :: hgrid

    call houaxis(ftag_%hess_, hgrid%ax_modl, 1)
    call houaxis(ftag_%hess_, hgrid%ay_modl, 2)
    call houaxis(ftag_%hess_, hgrid%az_modl, 3)
    call houaxis(ftag_%hess_, hgrid%ax_offd, 4)
    call houaxis(ftag_%hess_, hgrid%ay_offd, 5)
    call houaxis(ftag_%hess_, hgrid%az_offd, 6)

    call auxputch("label1", "s", "x_modl", ftag_%hess_)
    call auxputch("label2", "s", "y_modl", ftag_%hess_)
    call auxputch("label3", "s", "z_modl", ftag_%hess_)
    call auxputch("label4", "s", "x_offd", ftag_%hess_)
    call auxputch("label5", "s", "y_offd", ftag_%hess_)
    call auxputch("label6", "s", "z_offd", ftag_%hess_)
    return
  end subroutine

  subroutine hesdct_histos_mask(ftag_, amask)
    type(hesdct_ftag) :: ftag_
    type(acq_mask)    :: amask

    call hinaxis(ftag_%mask_, amask%arx, 1)
    call hinaxis(ftag_%mask_, amask%ary, 2)
    call hinaxis(ftag_%mask_, amask%asx, 3)
    call hinaxis(ftag_%mask_, amask%asy, 4)

    amask%nsh = amask%asx%n*amask%asy%n
    amask%ntr = amask%arx%n*amask%ary%n

    amask%rxmin = amask%arx%o; amask%rxmax = amask%arx%o+(amask%arx%n-1)*amask%arx%d  
    amask%rymin = amask%ary%o; amask%rymax = amask%ary%o+(amask%ary%n-1)*amask%ary%d
    amask%sxmin = amask%asx%o; amask%sxmax = amask%asx%o+(amask%asx%n-1)*amask%asx%d
    amask%symin = amask%asy%o; amask%symax = amask%asy%o+(amask%asy%n-1)*amask%asy%d

    return
  end subroutine  

end module
