module hesdct_kernel_mod
  use sep
  use base_blkcal_mod
  use hesdct_struct_mod
  use hesdct_histos_mod
  use hesdct_fileio_mod
  implicit none

  type(dat_grid)                               ,private :: dgrid
  type(gre_grid)                               ,private :: ggrid
  type(hes_grid)                               ,private :: hgrid
  type(acq_mask)                               ,private :: amask
  type(block)                                  ,private :: wblock
  type(hesdct_ftag)                            ,private :: ftag_

  real   , dimension(:,:,:,:)    , allocatable, private :: mask
  real   , dimension(:,:,:,:)    , allocatable, private :: hess        !!.. (x_offd,y_offd,z_offd,nws) 
  complex, dimension(:,:,:,:,:,:), allocatable, private :: grns        !!.. (x_surf,y_surf,nws)
  complex, dimension(:)          , allocatable, private :: csou        !!.. source signature
  complex, dimension(:,:,:)      , allocatable, private :: efns        !!.. (x, y, ws)
  complex, dimension(:,:,:)      , allocatable, private :: efnr        !!.. (x, y, ws)

  integer, dimension(:)          , allocatable, private :: jx_modl, jy_modl, jz_modl
  integer, dimension(:)          , allocatable, private :: jx_offd, jy_offd, jz_offd

  logical                                     , private :: forwrd
  logical                                     , private :: offset
  logical                                     , private :: obcrec
  logical                                     , private :: wntwgt
 
  character(len=6)                            , private :: domain

  integer                                     , private :: ompchunk, nth
  integer  ,                          external, private :: omp_get_num_procs
  integer  ,                          external, private :: omp_get_num_threads
  integer  ,                          external, private :: omp_set_num_threads
  integer  ,                          external, private :: omp_get_thread_num

contains
!=======================================================================================
  integer function hesdct_kernel_init() result(stat)
    type(axis) :: asx, asy, arx, ary, aws
    stat = 1

    ftag_%grns_ = "grns"
    ftag_%hess_ = "hess"
    ftag_%csou_ = "csou"
    ftag_%efns_ = "efns"
    ftag_%efnr_ = "efnr"
    ftag_%mask_ = "mask"

    call auxinout(ftag_%grns_)
    call auxinout(ftag_%hess_)
    call auxinout(ftag_%csou_)
    call auxinout(ftag_%mask_)

    call hesdct_histos_csou(ftag_, dgrid)
    call hesdct_histos_grns(ftag_, ggrid)
    call hesdct_histos_mask(ftag_, amask)
    call hesdct_params_init(dgrid, ggrid, hgrid)
    call hesdct_histos_hess(ftag_, hgrid)
    call hesdct_fileio_init(ftag_%hess_, hgrid)


    if (amask%arx%d/=ggrid%ax_surf%d) call seperr("amask%arx%d/=ggrid%agx%d")
    if (amask%ary%d/=ggrid%ay_surf%d) call seperr("amask%ary%d/=ggrid%agy%d")


    call from_param("forwrd", forwrd   , .true.  )
    call from_param("offset", offset   , .false. )
    call from_param("obcrec", obcrec   , .false. )
    call from_param("wntwgt", wntwgt   , .false. )
    call from_param("nws"   , wblock%ns, 1       )
    call from_param("domain", domain   , "shtpro")

    if (domain=="encsou" .or. domain=="encsim") then 
       call auxinout(ftag_%efns_)
       call hinaxis(ftag_%efns_, asx, 1)
       call hinaxis(ftag_%efns_, asy, 2)
       call hinaxis(ftag_%efns_, aws, 3)
       if (asx%n/=amask%asx%n .or. asx%d/=amask%asx%d .or. asx%o/=amask%asx%o) call seperr("efns%asx/=amask%asx")
       if (asy%n/=amask%asy%n .or. asy%d/=amask%asy%d .or. asy%o/=amask%asy%o) call seperr("efns%asy/=amask%asy")
       if (aws%n/=ggrid%aw%n  .or. aws%d/=ggrid%aw%d  .or. aws%o/=ggrid%aw%o ) call seperr("efns%aw /=ggrid%aw ")
    end if
    if (domain=="encrec" .or. domain=="encsim") then
       call auxinout(ftag_%efnr_)
       call hinaxis(ftag_%efnr_, arx, 1)
       call hinaxis(ftag_%efnr_, ary, 2)
       call hinaxis(ftag_%efnr_, aws, 3)
       if (arx%n/=amask%arx%n .or. arx%d/=amask%arx%d .or. arx%o/=amask%arx%o) call seperr("efnr%asx/=amask%asx")
       if (ary%n/=amask%ary%n .or. ary%d/=amask%ary%d .or. ary%o/=amask%ary%o) call seperr("efnr%asy/=amask%asy")
       if (aws%n/=ggrid%aw%n  .or. aws%d/=ggrid%aw%d  .or. aws%o/=ggrid%aw%o ) call seperr("efnr%aw /=ggrid%aw ")
    end if

    if (offset) then
       dgrid%gxmin_save = dgrid%agx%o
       dgrid%gymin_save = dgrid%agy%o
    end if

!   setup OpenMP stuff
!   ------------------
    ompchunk = 1
    nth      = 1
   !$OMP PARALLEL
    if (omp_get_thread_num() == 0) then
       nth = omp_get_num_threads()
       write(0,*)"nthread =", nth, omp_get_num_procs()
    end if
    !$OMP END PARALLEL

    call base_blkcal(wblock, ggrid%aw)

    allocate(csou(wblock%ns))
    allocate(grns(ggrid%ax_surf%n,ggrid%ay_surf%n,ggrid%ax_modl%n,ggrid%ay_modl%n,ggrid%az_modl%n,wblock%ns))
    allocate(hess(hgrid%ax_modl%n,hgrid%ay_modl%n,hgrid%az_modl%n,nth))
    allocate(jx_modl(hgrid%ax_modl%n),jy_modl(hgrid%ay_modl%n),jz_modl(hgrid%az_modl%n))
    allocate(jx_offd(hgrid%ax_offd%n),jy_offd(hgrid%ay_offd%n),jz_offd(hgrid%az_offd%n))
    allocate(efns(amask%asx%n,amask%asy%n,wblock%ns))
    allocate(efnr(amask%arx%n,amask%ary%n,wblock%ns)) 
    allocate(mask(amask%arx%n,amask%ary%n,amask%asx%n,amask%asy%n))
    allocate(amask%index_mskgre_sou(2, amask%nsh))
    allocate(amask%index_mskgre_rec(2, amask%ntr))

    write(0,*)"-----------------------------------"
    write(0,*)"    DOMAIN:      ", domain
    write(0,*)"-----------------------------------" 
    stat = 0
  end function
!=======================================================================================
  integer function hesdct_kernel_clos() result(stat)
    deallocate(csou)
    deallocate(grns)
    deallocate(hess)
    deallocate(jx_modl, jy_modl, jz_modl)
    deallocate(jx_offd, jy_offd, jz_offd)
    deallocate(efns, efnr)
    deallocate(mask)
    deallocate(amask%index_mskgre_sou)
    deallocate(amask%index_mskgre_rec)
    stat = 0
  end function
!=======================================================================================
  integer function hesdct_kernel_oper() result(stat)
    type(axis) :: aws
    integer    :: iwb, iws, ith, isx, isy, ish
    integer    :: igxb, igxe, igyb, igye
    integer    :: jumprx, jumpry, jumpsx, jumpsy
    integer    :: ix_modl, iy_modl, iz_modl
    integer    :: ix_offd, iy_offd, iz_offd
    integer    :: index_x_modl, index_y_modl, index_z_modl
    integer    :: index_x_offd, index_y_offd, index_z_offd
    integer    :: imodl, nmodl
    real       :: ws, ws4
    complex    :: fs2, wgt

    nmodl = hgrid%ax_modl%n*hgrid%ay_modl%n*hgrid%az_modl%n

    jumpsx = int(dgrid%asx%d/ggrid%ax_surf%d+0.5)
    jumpsy = int(dgrid%asy%d/ggrid%ay_surf%d+0.5)
    
    jumprx = int(dgrid%agx%d/ggrid%ax_surf%d+0.5)
    jumpry = int(dgrid%agy%d/ggrid%ay_surf%d+0.5)

write(0,*)"jumpsx, jumpsy", jumpsx, jumpsy
write(0,*)"jumprx, jumpry", jumprx, jumpry

!   read the acquisition mask matrix
!   --------------------------------
    call hesdct_fileio_mask(mask, amask, ftag_%mask_, "reed")

!   tabulate the acquisition geometry information
!   ---------------------------------------------
    call hesdct_kernel_geom(amask, ggrid)

write(0,*)"amask => nsh, ntr", amask%nsh , amask%ntr
write(0,*)"dgrid => nsh, ntr", dgrid%nshg, dgrid%ntrg

!   calculate the relative index for crosscorrelation
!   -------------------------------------------------
    call hesdct_kernel_indx(jx_modl, jy_modl, jz_modl, jx_offd, jy_offd, jz_offd, hgrid, ggrid)

!   loop over frequency block
!   -------------------------
    do iwb=1, wblock%nb
       write(0,*)"iwb=", iwb, wblock%nb 
       aws%o = wblock%os(iwb); aws%d = wblock%ds; aws%n = wblock%blen(iwb)
 
!      read source signature
!      --------------------- 
       call hesdct_fileio_csou(csou, dgrid, aws%n, wblock%indx(iwb), ftag_%csou_, "reed")

!      read Green's functions
!      ----------------------
       call hesdct_fileio_grns(grns, ggrid, aws%n, wblock%indx(iwb), ftag_%grns_, "reed")
       if (.not.forwrd) grns = conjg(grns)

!      read the encoding function
!      --------------------------
       if (domain=="encsou" .or. domain=="encsim") then
          call hesdct_fileio_efns(efns, dgrid, aws%n, wblock%indx(iwb), ftag_%efns_, "reed")
!          write(0,*) maxval(real(efns)), minval(aimag(efns))
       end if 
       if (domain=="encrec" .or. domain=="encsim") then
          call hesdct_fileio_efnr(efnr, dgrid, aws%n, wblock%indx(iwb), ftag_%efnr_, "reed")
!          write(0,*) maxval(real(efns)), minval(aimag(efns))
       end if
!      loop over frequencies
!      ---------------------
       do iws=1, wblock%ns
          ws  = aws%o + (iws-1)*aws%d
          fs2 = csou(iws)*conjg(csou(iws))
          wgt = 1./fs2 

!         multiply the weight w^4, if requested
!         -------------------------------------
          if (wntwgt) then
             ws4 = ws**4
             wgt = ws4/fs2
          end if

!         loop over off-diagonals
!         -----------------------
          do iz_offd=1, hgrid%az_offd%n
          do iy_offd=1, hgrid%ay_offd%n
          do ix_offd=1, hgrid%ax_offd%n
             hess = 0.
             call hesdct_fileio_hess2(hess(:,:,:,1), hgrid, ix_offd, iy_offd, iz_offd, ftag_%hess_, "reed")

!            loop over main diagonals
!            ------------------------
             !$OMP PARALLEL DO SCHEDULE(DYNAMIC,OMPCHUNK) DEFAULT(SHARED) FIRSTPRIVATE(iws,ith,ws)    &
             !$OMP PRIVATE(imodl,ish,ix_modl,iy_modl,iz_modl,index_x_modl,index_y_modl,index_z_modl)  &
             !$OMP PRIVATE(index_x_offd,index_y_offd,index_z_offd,isx,isy,igxb,igxe,igyb,igye)        
             do imodl = 1, nmodl
                ith = omp_get_thread_num() + 1 
                ix_modl = mod((imodl-1)/1                              ,hgrid%ax_modl%n)+1
                iy_modl = mod((imodl-1)/hgrid%ax_modl%n                ,hgrid%ay_modl%n)+1
                iz_modl = mod((imodl-1)/hgrid%ax_modl%n/hgrid%ay_modl%n,hgrid%az_modl%n)+1
                index_x_modl=jx_modl(ix_modl); index_x_offd=max(1,min(index_x_modl+jx_offd(ix_offd),ggrid%ax_modl%n))
                index_y_modl=jy_modl(iy_modl); index_y_offd=max(1,min(index_y_modl+jy_offd(iy_offd),ggrid%ay_modl%n))
                index_z_modl=jz_modl(iz_modl); index_z_offd=max(1,min(index_z_modl+jz_offd(iz_offd),ggrid%az_modl%n))

                select case (domain)

!                 Hessian in the shot-profile domain
!                 ----------------------------------
                  case ("shtpro")
                       call hesdct_kernel_shtpro(hess(ix_modl,iy_modl,iz_modl,ith), grns, mask,                amask, wgt, iws, jumprx, jumpry, &
                                                 index_x_modl, index_y_modl, index_z_modl, index_x_offd, index_y_offd, index_z_offd)

!                 Hessian in the encoded source domain
!                 ------------------------------------
                  case ("encsou")
                       call hesdct_kernel_encsou(hess(ix_modl,iy_modl,iz_modl,ith), grns, mask, efns(:,:,iws), amask, wgt, iws, jumpsx, jumpsy, &
                                                 index_x_modl, index_y_modl, index_z_modl, index_x_offd, index_y_offd, index_z_offd)
                end select

             end do
             !$OMP END PARALLEL DO

             call hesdct_fileio_hess2(sum(hess,4), hgrid, ix_offd, iy_offd, iz_offd, ftag_%hess_, "rite")

          end do
          end do
          end do

       end do
    end do
    stat = 0
  end function

! ---------------------------------------------------------------
! compute one component of the Hessian in the shot-profile domain
! ---------------------------------------------------------------
  subroutine hesdct_kernel_shtpro(hescpn, grns, mask, amask, wgt, iws, jumprx, jumpry,           &
             index_x_modl, index_y_modl, index_z_modl, index_x_offd, index_y_offd, index_z_offd)
    real   , intent(inout)          :: hescpn
    complex, dimension(:,:,:,:,:,:) :: grns
    real   , dimension(:,:,:,:)     :: mask    
    type(acq_mask)                  :: amask
    complex                         :: wgt
    integer                         :: iws
    integer                         :: jumprx, jumpry
    integer                         :: index_x_modl, index_y_modl, index_z_modl
    integer                         :: index_x_offd, index_y_offd, index_z_offd
    integer                         :: ish, isx, isy, jsx, jsy
    integer                         :: jrxb, jrxe, jryb, jrye
    complex                         :: grns_sou, grns_rec

    grns_sou = 0.
    grns_rec = 0.

 
!   the relative receiver index
!   ---------------------------
    jrxb = amask%index_mskgre_rxmin
    jrxe = amask%index_mskgre_rxmax
    jryb = amask%index_mskgre_rymin
    jrye = amask%index_mskgre_rymax

!   if OBC or land acquisition geometry
!   -----------------------------------
    if (obcrec) then
       grns_rec = sum(  grns(jrxb:jrxe:jumprx,jryb:jrye:jumpry,index_x_modl,index_y_modl,index_z_modl,iws)   &
                * conjg(grns(jrxb:jrxe:jumprx,jryb:jrye:jumpry,index_x_offd,index_y_offd,index_z_offd,iws))  &
                * mask(:,:,1,1))/amask%ntr
    end if

!   loop over the shots               
!   -------------------
    do ish=1, amask%nsh
!      the absolute shot index
!      -----------------------
       isx = mod((ish-1)            , amask%asx%n) + 1
       isy = mod((ish-1)/amask%asx%n, amask%asy%n) + 1

!      the relative shot index
!      -----------------------
       jsx  = amask%index_mskgre_sou(1,ish)
       jsy  = amask%index_mskgre_sou(2,ish)

!      source-side Green's functions
!      -----------------------------
       grns_sou =       grns(jsx,jsy,index_x_modl,index_y_modl,index_z_modl,iws)              &
                * conjg(grns(jsx,jsy,index_x_offd,index_y_offd,index_z_offd,iws))
!      receiver-side Green's functions
!      -------------------------------
       if (.not.obcrec) then
          grns_rec = sum(  grns(jrxb:jrxe:jumprx,jryb:jrye:jumpry,index_x_modl,index_y_modl,index_z_modl,iws)   &
                   * conjg(grns(jrxb:jrxe:jumprx,jryb:jrye:jumpry,index_x_offd,index_y_offd,index_z_offd,iws))  &
                   * mask(:,:,isx,isy))/amask%ntr
       end if

!      generate the Hessian for one shot
!      ---------------------------------
       hescpn = hescpn + grns_sou*grns_rec*wgt/amask%nsh
    end do
    return
  end subroutine
!-------------------------------------------------------------------------------------------------------------------
! compute one component of the Hessian in the encoded source domain
!-------------------------------------------------------------------------------------------------------------------
  subroutine hesdct_kernel_encsou(hescpn, grns, mask, efns, amask, wgt, iws, jumpsx, jumpsy,         &
             index_x_modl, index_y_modl, index_z_modl, index_x_offd, index_y_offd, index_z_offd)

    real   , intent(inout)          :: hescpn
    complex, dimension(:,:,:,:,:,:) :: grns
    real   , dimension(:,:,:,:)     :: mask
    complex, dimension(:,:)         :: efns
    type(acq_mask)                  :: amask
    complex                         :: wgt
    integer                         :: iws
    integer                         :: jumpsx, jumpsy
    integer                         :: index_x_modl, index_y_modl, index_z_modl
    integer                         :: index_x_offd, index_y_offd, index_z_offd
    integer                         :: itr, irx, iry, jrx, jry
    integer                         :: jsxb, jsxe, jsyb, jsye 
    complex                         :: grns_sou1, grns_sou2, grns_rec

!   relative shot index
!   -------------------
    jsxb = amask%index_mskgre_sxmin
    jsxe = amask%index_mskgre_sxmax
    jsyb = amask%index_mskgre_symin
    jsye = amask%index_mskgre_symax

!   if OBC or land acquisition geometry
!   -----------------------------------
    if (obcrec) then
       grns_sou1 = sum(grns(jsxb:jsxe:jumpsx,jsyb:jsye:jumpsy,index_x_modl,index_y_modl,index_z_modl,iws)*efns*mask(1,1,:,:))/amask%nsh
       grns_sou2 = sum(grns(jsxb:jsxe:jumpsx,jsyb:jsye:jumpsy,index_x_offd,index_y_offd,index_z_offd,iws)*efns*mask(1,1,:,:))/amask%nsh
    end if 

    do itr=1, amask%ntr
!      absolute array indexes
!      ----------------------
       irx = mod((itr-1)            , amask%arx%n) + 1
       iry = mod((itr-1)/amask%arx%n, amask%ary%n) + 1

!      relative array indexes
!      ----------------------
       jrx = amask%index_mskgre_rec(1, itr)
       jry = amask%index_mskgre_rec(2, itr)

!      encoding source-side Green's functions
!      --------------------------------------
       if (.not.obcrec) then
          grns_sou1 = sum(grns(jsxb:jsxe:jumpsx,jsyb:jsye:jumpsy,index_x_modl,index_y_modl,index_z_modl,iws)*efns*mask(irx,iry,:,:))/amask%nsh
          grns_sou2 = sum(grns(jsxb:jsxe:jumpsx,jsyb:jsye:jumpsy,index_x_offd,index_y_offd,index_z_offd,iws)*efns*mask(irx,iry,:,:))/amask%nsh
       end if

!      receiver-side Green's functions
!      -------------------------------
       grns_rec  =       grns(jrx,jry,index_x_modl,index_y_modl,index_z_modl,iws)   &
                 * conjg(grns(jrx,jry,index_x_offd,index_y_offd,index_z_offd,iws))

       hescpn = hescpn + grns_sou1*conjg(grns_sou2)*grns_rec*wgt/amask%ntr
    end do
    return
  end subroutine

  !--------------------------------------------------------------------------------------------------
  subroutine hesdct_kernel_indx(jx_modl, jy_modl, jz_modl, jx_offd, jy_offd, jz_offd, hgrid, ggrid)
    integer, dimension(:) :: jx_modl, jy_modl, jz_modl
    integer, dimension(:) :: jx_offd, jy_offd, jz_offd
    type(hes_grid)        :: hgrid
    type(gre_grid)        :: ggrid

    integer               :: ix_modl, iy_modl, iz_modl
    integer               :: ix_offd, iy_offd, iz_offd
    real                  :: x_modl, y_modl, z_modl
    real                  :: x_offd, y_offd, z_offd 
    
!   relative index of main diagonal with respect to the Green's function
!   --------------------------------------------------------------------
    do ix_modl=1, hgrid%ax_modl%n
       x_modl = hgrid%ax_modl%o+(ix_modl-1)*hgrid%ax_modl%d
       jx_modl(ix_modl) = 1.5 + (x_modl-ggrid%ax_modl%o)/ggrid%ax_modl%d
    end do
    do iy_modl=1, hgrid%ay_modl%n
       y_modl = hgrid%ay_modl%o+(iy_modl-1)*hgrid%ay_modl%d
       jy_modl(iy_modl) = 1.5 + (y_modl-ggrid%ay_modl%o)/ggrid%ay_modl%d
    end do
    do iz_modl=1, hgrid%az_modl%n
       z_modl = hgrid%az_modl%o+(iz_modl-1)*hgrid%az_modl%d
       jz_modl(iz_modl) = 1.5 + (z_modl-ggrid%az_modl%o)/ggrid%az_modl%d
    end do

!   offset from the main diagonal to the off-diagonal
!   -------------------------------------------------
    do ix_offd=1, hgrid%ax_offd%n
       x_offd = hgrid%ax_offd%o+(ix_offd-1)*hgrid%ax_offd%d
       jx_offd(ix_offd) = nint(x_offd/ggrid%ax_modl%d)
    end do
    do iy_offd=1, hgrid%ay_offd%n
       y_offd = hgrid%ay_offd%o+(iy_offd-1)*hgrid%ay_offd%d
       jy_offd(iy_offd) = nint(y_offd/ggrid%ay_modl%d)
    end do
    do iz_offd=1, hgrid%az_offd%n
       z_offd = hgrid%az_offd%o+(iz_offd-1)*hgrid%az_offd%d
       jz_offd(iz_offd) = nint(z_offd/ggrid%az_modl%d)
    end do
    return
  end subroutine

!=====================================================================================
  subroutine hesdct_kernel_geom(amask, ggrid)
    type(acq_mask)          :: amask
    type(gre_grid)          :: ggrid

    integer                 :: ish, isx, isy
    integer                 :: itr, irx, iry
    real                    :: sx, sy, rx, ry

!   tabulating the source coordinates
!   ---------------------------------
    do ish=1, amask%nsh
       isx = mod((ish-1)            , amask%asx%n) + 1
       isy = mod((ish-1)/amask%asx%n, amask%asy%n) + 1
       sx  = amask%asx%o + (isx-1)*amask%asx%d
       sy  = amask%asy%o + (isy-1)*amask%asy%d

       amask%index_mskgre_sou(1,ish) = 1.5+(sx-ggrid%ax_surf%o)/ggrid%ax_surf%d
       amask%index_mskgre_sou(2,ish) = 1.5+(sy-ggrid%ay_surf%o)/ggrid%ay_surf%d
    end do

!   tabulating the receiver coordinates
!   -----------------------------------
    do itr=1, amask%ntr
       irx = mod((itr-1)            , amask%arx%n) + 1
       iry = mod((itr-1)/amask%arx%n, amask%ary%n) + 1
       rx  = amask%arx%o+(irx-1)*amask%arx%d
       ry  = amask%ary%o+(iry-1)*amask%ary%d

       amask%index_mskgre_rec(1,itr) = 1.5+(rx-ggrid%ax_surf%o)/ggrid%ax_surf%d
       amask%index_mskgre_rec(2,itr) = 1.5+(ry-ggrid%ay_surf%o)/ggrid%ay_surf%d
    end do

!   the minimum and maximum receiver indexes
!   ----------------------------------------
    amask%index_mskgre_rxmin=1.5+(amask%rxmin-ggrid%ax_surf%o)/ggrid%ax_surf%d
    amask%index_mskgre_rxmax=1.5+(amask%rxmax-ggrid%ax_surf%o)/ggrid%ax_surf%d
    amask%index_mskgre_rymin=1.5+(amask%rymin-ggrid%ay_surf%o)/ggrid%ay_surf%d
    amask%index_mskgre_rymax=1.5+(amask%rymax-ggrid%ay_surf%o)/ggrid%ay_surf%d

!   the minimum and maximum source indexes
!   --------------------------------------
    amask%index_mskgre_sxmin=1.5+(amask%sxmin-ggrid%ax_surf%o)/ggrid%ax_surf%d
    amask%index_mskgre_sxmax=1.5+(amask%sxmax-ggrid%ax_surf%o)/ggrid%ax_surf%d
    amask%index_mskgre_symin=1.5+(amask%symin-ggrid%ay_surf%o)/ggrid%ay_surf%d
    amask%index_mskgre_symax=1.5+(amask%symax-ggrid%ay_surf%o)/ggrid%ay_surf%d

    return
  end subroutine

end module

