module wei_type_mod
!--------------------------------------------------
  !!.. define type for axes
  type axis 
    integer :: n   ! total number
    real    :: d   ! spacing
    real    :: o   ! origin
  end type

!---------------------------------------------------
  !!.. define propagation grid
  type pro_grid
    type(axis) :: ax, ay, az
    real       :: sx, sy, sz
    real       :: gxmin, gxmax
    real       :: gymin, gymax
    real       :: pxmin, pxmax
    real       :: pymin, pymax
    real       :: jx, jy             ! determine the sampling for the propagation grid : dpx = jx * dgx
    integer    :: index_datpro_sx    ! index of sx on the propagation grid
    integer    :: index_datpro_sy    ! index of sy on the propagation grid
    integer    :: index_datpro_gxmin   ! index of the first receiver gx on the propagation grid (for regular data)
    integer    :: index_datpro_gymin   ! index of the first receiver gy on the propagation grid (for regular data)
    integer, allocatable, dimension(:,:) :: index_datpro_hrec  ! the index file for unsorted receiver
    integer, allocatable, dimension(:)   :: index_datpro_hsou  ! the index file for unsorted source
    integer    :: nxpad_beg, nxpad_end
    integer    :: nypad_beg, nypad_end
  end type

!---------------------------------------------------
  !!.. define data grid
  type dat_grid
    type(axis) :: aw
    type(axis) :: agx, agy ! for regular data
    type(axis) :: asx, asy ! for regular data
    real       :: sx, sy   ! for unsorted data
    real       :: sz, gz   ! not used yet
    integer    :: ntr      ! total traces per shot
    integer    :: nsh      ! total shot number
    logical    :: hdr_rec  ! flag indicating whether the receiver header is presented
    logical    :: hdr_sou  ! flag indicating whether the source header is presented
    real, allocatable, dimension(:,:) :: header_rec   ! header file for receiver locations, header_rec(2,ntraces)
    real, allocatable, dimension(:)   :: header_sou   ! header file for source locations, header_sou(2), x, y
  end type

!----------------------------------------------------
  type vel_grid
    type(axis) :: ax, ay, az
    integer    :: index_provel_xmin  ! first x index of the input velocity slice on the whole velocity grid
    integer    :: index_provel_ymin  ! first y index of the input velocity slice on the whole velocity grid
    integer    :: index_provel_xmax  ! last x index of the input velocity slice on the whole velocity grid
    integer    :: index_provel_ymax  ! last y index of the input velocity slice on the whole velocity grid
  end type


!----------------------------------------------------
  !!.. define imaging grid
  type img_grid
    type(axis) :: ax, ay, az
    type(axis) :: ahx, ahy
    real       :: image_xmin
    real       :: image_ymin
    real       :: image_xmax
    real       :: image_ymax
    real       :: image_dx
    real       :: image_dy
  end type

!-----------------------------------------------------
  type out_grid
    type(axis) :: ax, ay, ahx, ahy, az
    integer    :: index_outimg_xmin  ! first x index of the output image on the image grid
    integer    :: index_outimg_ymin  ! first y index of the output image on the image grid
    integer    :: index_outimg_xmax  ! last x index of the output image on the image grid
    integer    :: index_outimg_ymax  ! last y index of the output image on the image grid
    integer    :: index_outpro_xmin  ! first x index of the output image on the propagation grid
    integer    :: index_outpro_ymin  ! first y index of the output image on the propagation grid
    integer    :: index_outpro_xmax  ! last x index of the output image on the propagation grid
    integer    :: index_outpro_ymax  ! last y index of the output image on the propagation grid
  end type

!--------------------------------------------------------
  type gre_grid
    type(axis) :: ax_modl, ay_modl, az_modl
    type(axis) :: ax_surf, ay_surf, aw
    real       :: surface_xmin       ! xmin for the surface location
    real       :: surface_xmax       ! xmax for the surface location
    real       :: surface_ymin       ! ymin for the surface location
    real       :: surface_ymax       ! ymax for the surface location
    integer    :: index_surpro_xmin  ! first x index of the surface grid on the propagation grid
    integer    :: index_surpro_xmax  ! last x index of the surface grid on the propagation grid
    integer    :: index_surpro_ymin
    integer    :: index_surpro_ymax
    integer    :: index_modpro_modx  ! x index of the model point on the propagation grid (similar to shot mapping)
    integer    :: index_modpro_mody  ! y index of the model point on the propagation grid (similar to shot mapping)
    integer    :: index_modpro_modz  ! z index of the model point on the propagation grid (similar to shot mapping)
    integer    :: index_datgre_sx
    integer    :: index_datgre_sy  
    integer, dimension(:,:), pointer :: index_datgre_rec
  end type 
!---------------------------------------------------------------
  type hes_grid
    type(axis)                       :: ax_modl, ay_modl, az_modl ! defines the main diagonal of Hessian, or the points in the model space
    type(axis)                       :: ax_offd, ay_offd, az_offd ! defines the off diagonal of Hessian
    type(axis)                       :: ahx_modl, ahy_modl        ! defines the subsurface offset for the main diagonal of Hessian
    type(axis)                       :: ahx_offd, ahy_offd        ! defines the subsurface offset for the off diagonal of Hessian
    type(axis)                       :: aax_modl, aax_offd
    real                             :: sx, sy                    ! source location for current shot
    real                             :: xmin_modl, xmax_modl
    real                             :: ymin_modl, ymax_modl
    real                             :: zmin_modl, zmax_modl
    integer                          :: index_datgre_sx
    integer                          :: index_datgre_sy
    integer                          :: index_datgre_gxmin
    integer                          :: index_datgre_gymin
    integer, dimension(:,:), pointer :: index_datgre_rec     ! index of receiver locations on the Green's surface grid for each shot
    integer, dimension(:,:), pointer :: mask_rec_loc         ! receiver location mask, 1 for where receivers are located, 0 for not.
  end type
!----------------------------------------------------------------
  type block
    integer                            :: ns
    integer                            :: nb
    integer                            :: nb_res
    real                               :: ds
    real   , allocatable, dimension(:) :: os
    integer, allocatable, dimension(:) :: blen
    integer, allocatable, dimension(:) :: indx
    logical                            :: bres
  end type





end module







