import os, sys
from optparse import OptionParser

usage = 'usage: %prog [options] arg1 arg2'

parser = OptionParser(usage=usage)

parser.add_option('--bindir', dest='bindir', type='string', default='.', help='Path of the excutable')
parser.add_option('--datdir', dest='datdir', type='string', default='.', help='Path of the input data')
parser.add_option('--resdir', dest='resdir', type='string', default='.', help='Path of the results')
parser.add_option('--tmpdir', dest='tmpdir', type='string', default='.', help='Path of the temp files')

parser.add_option('--csou', dest='csou', type='string', default='CSOU.H', help='Source wavefield')
parser.add_option('--crec', dest='crec', type='string', default='CREC.H', help='Receiver wavefield')
parser.add_option('--bvel', dest='bvel', type='string', default='BVEL.H', help='Velocity model')
parser.add_option('--bimg', dest='bimg', type='string', default='BIMG.H', help='Migrated image')

parser.add_option('--splitsou', dest='splitsou', type='string', default='n', help='y/[n], whether or not to split sources')
parser.add_option('--splitrec', dest='splitrec', type='string', default='y', help='[y]/n, whether or not to split receivers')
parser.add_option('--restart', dest='restart', type='string', default='n', help='y/[n], whether or not to split the data again')

parser.add_option('--oper', dest='oper', type='string', default=' ', help='Propagator type')
parser.add_option('--pdip', dest='pdip', type='string', default=' ', help='Propagation dip (foufd and impfd)')
parser.add_option('--nvrf', dest='nvrf', type='string', default=' ', help='Number of reference velocities')
parser.add_option('--nws' , dest='nws' , type='string', default=' ', help='Number of frequencies to hold in memory')
parser.add_option('--offset', dest='offset', type='string', default=' ', help='y/[n], marine or land geometry')

parser.add_option('--nxpad_beg', dest='nxpad_beg', type='string', default=' ', help='Pad in x in the front')
parser.add_option('--nxpad_end', dest='nxpad_end', type='string', default=' ', help='Pad in x in the back')
parser.add_option('--nypad_beg', dest='nypad_beg', type='string', default=' ', help='Pad in y in the front')
parser.add_option('--nypad_end', dest='nypad_end', type='string', default=' ', help='Pad in y in the back')

parser.add_option('--ntpx', dest='ntpx', type='string', default=' ', help='Taper in x')
parser.add_option('--ntpy', dest='ntpy', type='string', default=' ', help='Taper in y')

parser.add_option('--nhx', dest='nhx', type='string', default=' ', help='Local offset in x')
parser.add_option('--ohx', dest='ohx', type='string', default=' ', help='Local offset in x')
parser.add_option('--dhx', dest='dhx', type='string', default=' ', help='Local offset in x')
parser.add_option('--nhy', dest='nhy', type='string', default=' ', help='Local offset in y')
parser.add_option('--ohy', dest='ohy', type='string', default=' ', help='Local offset in y')
parser.add_option('--dhy', dest='dhy', type='string', default=' ', help='Local offset in y')

parser.add_option('--image_xmin', dest='image_xmin', type='string', default=' ', help='Output image xmin')
parser.add_option('--image_xmax', dest='image_xmax', type='string', default=' ', help='Output image xmax')
parser.add_option('--image_ymin', dest='image_ymin', type='string', default=' ', help='Output image ymin')
parser.add_option('--image_ymax', dest='image_ymax', type='string', default=' ', help='Output image ymax')
parser.add_option('--image_zmin', dest='image_zmin', type='string', default=' ', help='Output image zmin')
parser.add_option('--image_zmax', dest='image_zmax', type='string', default=' ', help='Output image zmax')

parser.add_option('--comsou', dest='comsou', type='string', default=' ', help='y/[n], composite source or not')

parser.add_option('--memchk', dest='memchk', type='string', default=' ', help='Memory check or not')

parser.add_option('--nsx_tot', dest='nsx_tot', type='string', default='1', help='Total number of shots')
parser.add_option('--jsx_win', dest='jsx_win', type='string', default='1', help='Number of shots in a windowed shot in x')

parser.add_option('--queue', dest='queue', type='string', default='Q35', help='[Q35]/default, which queue to use')
parser.add_option('--grepinfo', dest='grepinfo', type='string', default='line0', help='Info for grep')
parser.add_option('--datapath', dest='datapath', type='string', default='/tmp/', help='Datapath for output files')
(options, args) = parser.parse_args()

BINDIR = options.bindir
DATDIR = options.datdir
RESDIR = options.resdir
TMPDIR = options.tmpdir
CSOU = options.csou
CREC = options.crec
BVEL = options.bvel
BIMG = options.bimg

root, ext = os.path.splitext(CSOU)
CSOU_TEMP = root+'.shot%d-temp'+ext
CSOU_TEMP = os.path.join(TMPDIR, CSOU_TEMP)

root, ext = os.path.splitext(CREC)
CREC_TEMP = root+'.shot%d-temp'+ext
CREC_TEMP = os.path.join(TMPDIR, CREC_TEMP)

root, ext = os.path.splitext(BIMG)
BIMG_TEMP = root+'.shot%d-temp'+ext
BIMG_TEMP = os.path.join(TMPDIR, BIMG_TEMP)

root, ext = os.path.splitext(BVEL)
BVEL_TEMP = root+'.shot%d-temp'+ext
BVEL_TEMP = os.path.join(TMPDIR, BVEL_TEMP)

root, ext = os.path.splitext(BIMG)
BIMG_SPLIT = root+'.shot%d'+ext
BIMG_SPLIT = os.path.join(RESDIR, BIMG_SPLIT)

root, ext = os.path.splitext(BIMG)
BIMG_JUNKS = root+'.junk'+ext
BIMG_JUNKS = os.path.join(RESDIR, BIMG_JUNKS)

CSOU = os.path.join(DATDIR, CSOU)
CREC = os.path.join(DATDIR, CREC)
BVEL = os.path.join(DATDIR, BVEL)
BIMG = os.path.join(RESDIR, BIMG)

PARAM_PRO = 'oper=%s pdip=%s nvrf=%s nws=%s offset=%s comsou=%s memchk=%s ' % (options.oper, options.pdip, options.nvrf, options.nws, options.offset, options.comsou, options.memchk)
PARAM_PAD = 'nxpad_beg=%s nxpad_end=%s nypad_beg=%s nypad_end=%s ' % (options.nxpad_beg, options.nxpad_end, options.nypad_beg, options.nypad_end)
PARAM_TAP = 'ntpx=%s ntpy=%s ' % (options.ntpx, options.ntpy)
PARAM_IMG = 'image_xmin=%s image_xmax=%s image_ymin=%s image_ymax=%s image_zmin=%s image_zmax=%s ' % (options.image_xmin, options.image_xmax, 
		options.image_ymin, options.image_ymax, options.image_zmin, options.image_zmax)
PARAM_OFF = 'nhx=%s ohx=%s dhx=%s nhy=%s ohy=%s dhy=%s ' % (options.nhx, options.ohx, options.dhx, options.nhy, options.ohy, options.dhy)
DATAPATH  = 'datapath=%s ' % options.datapath

NSX_TOT = int(options.nsx_tot)
JSX_WIN = int(options.jsx_win)
JSX_RES = NSX_TOT % JSX_WIN
if JSX_RES == 0:
	NSX_WIN = NSX_TOT/JSX_WIN
else:
	NSX_WIN = (NSX_TOT-JSX_RES)/JSX_WIN + 1

print 'NSX_TOT = ', NSX_TOT
print 'JSX_WIN = ', JSX_WIN
print 'JSX_RES = ', JSX_RES
print 'NSX_WIN = ', NSX_WIN

SHOT_INDEX_BEG = 1
SHOT_INDEX_END = NSX_WIN + 1

#########
# loop over shots
#########
for ishot in range(SHOT_INDEX_BEG, SHOT_INDEX_END):
	########
        # generate the script for each shot
        ########
	scriptnamebase = os.path.join(RESDIR, 'PBS_%s' % options.bimg + '_shot%d' % ishot)
        scriptname     = scriptnamebase + '.sh'
        logname        = scriptnamebase + '.log'
        fptr           = open(scriptname, 'wt')
        fptr.write('#!/bin/tcsh\n')
	fptr.write('#PBS -N %s\n' % options.grepinfo)
	fptr.write('#PBS -l nodes=1:ppn=8\n')
	fptr.write('#PBS -q %s\n' % options.queue)
	fptr.write('#PBS -V\n')
	fptr.write('#PBS -m e\n')
	fptr.write('#PBS -W x="PARTITION:sw121"\n')
	#fptr.write('#PBS -M tang@sep.stanford.edu\n')
	fptr.write('#PBS -e '+scriptnamebase+'.err\n')
	fptr.write('#PBS -o '+scriptnamebase+'.out\n')
	fptr.write('cd $PBS_O_WORKDIR\n')
	fptr.write('\n')

	###############
	fptr.write('echo "Copying the velocity"\n')
	cmd  = 'Cp %s ' % BVEL + ' %s ' % (BVEL_TEMP % ishot)
	cmd += ' datapath=%s\n ' % options.datapath
	fptr.write('%s\n' %cmd)

	###############
	fptr.write('echo "Splitting sources"\n')
	if JSX_RES==0:
		cmd = 'Window3d ' + '< ' + CSOU + ' n4=%d ' % JSX_WIN + 'f4=%d' % ((ishot-1)*JSX_WIN) + ' squeeze=n >' + CSOU_TEMP % ishot
	else:
		if ishot<SHOT_INDEX_END-1:
			cmd = 'Window3d '+ '< '+ CSOU + ' n4=%d ' % JSX_WIN + 'f4=%d' % ((ishot-1)*JSX_WIN)   + ' squeeze=n >' + CSOU_TEMP % ishot
		else:
			cmd = 'Window3d '+ '< '+ CSOU + ' n4=%d ' % JSX_RES + 'f4=%d' % ((NSX_WIN-1)*JSX_WIN) + ' squeeze=n >' + CSOU_TEMP % NSX_WIN

	cmd += ' datapath=%s\n ' % options.datapath
	fptr.write('%s\n' % cmd)

	###############
	fptr.write('echo "Splitting receivers"\n')
	if JSX_RES==0:
                cmd = 'Window3d ' + '< ' + CREC + ' n4=%d ' % JSX_WIN + 'f4=%d' % ((ishot-1)*JSX_WIN) + ' squeeze=n >' + CREC_TEMP % ishot
        else:
                if ishot<SHOT_INDEX_END-1:
                        cmd = 'Window3d '+ '< '+ CREC + ' n4=%d ' % JSX_WIN + 'f4=%d' % ((ishot-1)*JSX_WIN)   + ' squeeze=n >' + CREC_TEMP % ishot
                else:
                        cmd = 'Window3d '+ '< '+ CREC + ' n4=%d ' % JSX_RES + 'f4=%d' % ((NSX_WIN-1)*JSX_WIN) + ' squeeze=n >' + CREC_TEMP % NSX_WIN
	cmd += ' datapath=%s\n ' % options.datapath
	fptr.write('%s\n' % cmd)

	###############
	fptr.write('echo "MIGRATION"\n')
	cmd  = os.path.join(BINDIR, 'bwi-wem3d.x ')

	cmd += 'csou=%s ' % (CSOU_TEMP % ishot)
	cmd += 'crec=%s ' % (CREC_TEMP % ishot)
	cmd += 'bvel=%s ' % (BVEL_TEMP % ishot)
	cmd += 'bimg=%s ' % (BIMG_TEMP % ishot)

	cmd += 'mode=imgadj '

	cmd += PARAM_PRO
	cmd += PARAM_PAD
	cmd += PARAM_TAP
	cmd += PARAM_IMG
	cmd += PARAM_OFF
	cmd += DATAPATH
 
	fptr.write(' %s\\\n' % cmd)
	fptr.write(' >& %s\n' % logname)

	fptr.write('\n')
	fptr.write('Cp %s ' % (BIMG_TEMP % ishot) + ' %s\n' % (BIMG_SPLIT % ishot))

	fptr.write('\n')
	fptr.write('Rm %s\n' % (CSOU_TEMP % ishot))

	fptr.write('\n')
	fptr.write('Rm %s\n' % (CREC_TEMP % ishot))

	fptr.write('\n')
	fptr.write('Rm %s\n' % (BIMG_TEMP % ishot))

	fptr.write('\n')
	fptr.write('Rm %s\n' % (BVEL_TEMP % ishot))

	fptr.close()

	cmd = 'qsub '+scriptname
	os.system(cmd)

while os.popen('qstat | grep tangyx | grep -w -c %s' % options.grepinfo).read() != '0\n':
	os.system('sleep 30')

print '********************'
print 'Now collecting shots'
print '********************'
os.system('sleep 300')
cmd = 'Add ' + (BIMG_SPLIT % SHOT_INDEX_BEG) + ' scale=0 '+'> '+ BIMG_JUNKS
print cmd
os.system(cmd)
for ishot in range(SHOT_INDEX_BEG, SHOT_INDEX_END):
	cmd1 = 'Add ' + BIMG_JUNKS + ' ' + (BIMG_SPLIT % ishot) + ' > ' + BIMG
	cmd2 = 'Cp '  + BIMG + ' ' + BIMG_JUNKS
	print cmd1
	os.system(cmd1)
	print cmd2
	os.system(cmd2)

cmd = 'Rm ' + BIMG_JUNKS
print cmd
os.system(cmd)
