#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "mpi.h"
#include "string.h"
#include "seplib.h"
#include "bwi-commd-mpi.c"
#include "bwi-files-mpi.c"

#define LMAX_DIR 256
#define LMAX_FNS 256
#define LMAX_CMD 1024

int ssystem(char* cmd) {
  fprintf(stderr,"executing cmd: %s \n", cmd);
  return 0;
}

int main(int argc, char* argv[]) {
  char bindir[LMAX_DIR], datdir[LMAX_DIR], resdir[LMAX_DIR], tmpdir[LMAX_DIR], datapath[LMAX_DIR];
  char fncsou[LMAX_FNS], fncrec[LMAX_FNS], fnbvel[LMAX_FNS], fnbimg[LMAX_FNS], fnlogs[LMAX_FNS];
  char fncsou_orig[LMAX_FNS], fncrec_orig[LMAX_FNS], fnbvel_orig[LMAX_FNS], fnbimg_orig[LMAX_FNS];
  char fngrdo_orig[LMAX_FNS], fngrdp_orig[LMAX_FNS], fnsrch_orig[LMAX_FNS], fnfobj_orig[LMAX_FNS];
  char fncsou_temp[LMAX_FNS], fncrec_temp[LMAX_FNS], fnbvel_temp[LMAX_FNS], fnbimg_temp[LMAX_FNS];
  char fngrdo_temp[LMAX_FNS], fngrdp_temp[LMAX_FNS], fnsrch_temp[LMAX_FNS], fnfobj_temp[LMAX_FNS];
  char fnlogs_temp[LMAX_FNS];
  char fnlogs_orig[LMAX_FNS];
  char cpy_csou[LMAX_CMD], cpy_crec[LMAX_CMD], cpy_bvel[LMAX_CMD], cpy_bimg[LMAX_CMD];
  char cpy_grdo[LMAX_CMD], cpy_grdp[LMAX_CMD], cpy_srch[LMAX_CMD], cpy_fobj[LMAX_CMD];
  char cpy_mmov[LMAX_CMD], cpy_fmov[LMAX_CMD];
  char cmd_imagcube[LMAX_CMD];
  char par_imagcube[LMAX_CMD];
  char cmd_logs[LMAX_CMD];
  char suffix[20], shot_string[24], numstr[4];
  char nsx_string[20], dsx_string[20], osx_string[20];
  char nsy_string[20], dsy_string[20], osy_string[20];
  
  axis ahx_igrid={1,1.,0.}, ahy_igrid={1,1.,0.}, ax_igrid={1,1.,0.}, ay_igrid={1,1.,0.}, az_igrid={1,1.,0.};
  
  float *btmp=NULL, *bimg=NULL;
  
  int err, irank, nrank, nsb, reduce, nsmp_igrid, iz, ismp;
  
  MPI_Init(&argc,&argv);
  MPI_Comm_rank(MPI_COMM_WORLD, &irank);
  MPI_Comm_size(MPI_COMM_WORLD, &nrank);
  
  initpar(argc,argv);
  doc(" ");
  
  if (0==getch("bindir","s",&bindir)) strcpy(bindir,"./");
  if (0==getch("datdir","s",&datdir)) strcpy(datdir,"./");
  if (0==getch("resdir","s",&resdir)) strcpy(resdir,"./");
  if (0==getch("tmpdir","s",&tmpdir)) strcpy(tmpdir,"./");
  
  if (0==getch("suffix","s",&suffix)) strcpy(suffix,"./");
  
  if (0==getch("fncsou","s",&fncsou)) strcpy(fncsou,"CSOU.H");
  if (0==getch("fncrec","s",&fncrec)) strcpy(fncrec,"CREC.H");
  if (0==getch("fnbvel","s",&fnbvel)) strcpy(fnbvel,"BVEL.H");
  if (0==getch("fnbimg","s",&fnbimg)) strcpy(fnbimg,"BIMG.H");
  
  if (0==getch("par_imagcube","s",&par_imagcube)) strcpy(par_imagcube," ");
  
  if (0==getch("nsb"   ,"i",&nsb   )) nsb    = 1;
  if (0==getch("reduce","i",&reduce)) reduce = 1;
  
  if (nrank!=nsb){
     fprintf(stderr, "ERROR: nrank != nsb,  %d %d \n", nrank, nsb);
     MPI_FINALIZE();
  }
  
  strcpy(fnlogs, "log");

  sprintf(numstr, "%d", irank+1);
  strcpy(shot_string, suffix); strcat(shot_string, numstr);
  
  strcpy(fnlogs_orig, resdir); strcat(fnlogs_orig, fnlogs); strcat(fnlogs_orig, shot_string);
  strcpy(fnlogs_temp, tmpdir); strcat(fnlogs_temp, fnlogs); strcat(fnlogs_temp, shot_string);
  
  strcpy(fnbimg_orig, resdir); strcat(fnbimg_orig, fnbimg); 
  strcpy(fnbimg_temp, tmpdir); strcat(fnbimg_temp, fnbimg); strcat(fnbimg_temp, shot_string);
  
  strcpy(fncsou_orig, datdir); strcat(fncsou_orig, fncsou); strcat(fncsou_orig, shot_string);
  strcpy(fncsou_temp, tmpdir); strcat(fncsou_temp, fncsou); strcat(fncsou_temp, shot_string);
  
  strcpy(fncrec_orig, datdir); strcat(fncrec_orig, fncrec); strcat(fncrec_orig, shot_string);
  strcpy(fncrec_temp, tmpdir); strcat(fncrec_temp, fncrec); strcat(fncrec_temp, shot_string);
  
  strcpy(fnbvel_orig, datdir); strcat(fnbvel_orig, fnbvel);
  strcpy(fnbvel_temp, tmpdir); strcat(fnbvel_temp, fnbvel); strcat(fnbvel_temp, shot_string);
  
  ////////////////////////////
  // copy source to all nodes
  ////////////////////////////
  create_cpy_one2all(cpy_csou, tmpdir, fncsou_orig, fncsou_temp);
  create_cpy_one2all(cpy_crec, tmpdir, fncrec_orig, fncrec_temp);
  create_cpy_one2all(cpy_bvel, tmpdir, fnbvel_orig, fnbvel_temp);
  
  create_cpy_all2one(cpy_bimg, fnbimg_orig, fnbimg_temp);
  
  create_cmd_imagcube(cmd_imagcube, bindir, tmpdir, fncsou_temp, fncrec_temp, fnbvel_temp, fnbimg_temp, par_imagcube);
  
  // copy sources
  err = ssystem(cpy_csou);
  if (0!=err) {fprintf(stderr, "ERR=%d, irank=%d, %s\n", err, irank, cpy_csou); MPI_Finalize(); exit(1);}
  
  // copy datas
  err = ssystem(cpy_crec);
  if (0!=err) {fprintf(stderr, "ERR=%d, irank=%d, %s\n", err, irank, cpy_crec); MPI_Finalize(); exit(1);}
  
  // copy background velocity
  err = ssystem(cpy_bvel);
  if (0!=err) {fprintf(stderr, "ERR=%d, irank=%d, %s\n", err, irank, cpy_bvel); MPI_Finalize(); exit(1);}
  
  
  //strcat(cmd_imagcube, ">&"); strcat(cmd_imagcube, fnlogs_temp);
  
  // compute background image
  err = ssystem(cmd_imagcube);
  if (0!=err) {
     fprintf(stderr, "ERR=%d, irank=%d, %s\n", err, irank, cmd_imagcube); 
  //   strcpy(cmd_logs, "cp "); strcat(cmd_logs, fnlogs_temp); strcat(cmd_logs, " "); strcat(cmd_logs, fnlogs_orig);
  //   fprintf(stderr, "%s\n", cmd_logs);
  //   ssystem(cmd_logs); 
     MPI_Finalize(); exit(1);
  }
  
  
  MPI_Barrier(MPI_COMM_WORLD);
  
  //Collect all images and add them to a single images.
  if (reduce){
     // collect images
     auxinout(fnbimg_temp);
     hinaxis(fnbimg_temp, &ahx_igrid, 1);
     hinaxis(fnbimg_temp, &ahy_igrid, 2);
     hinaxis(fnbimg_temp, &ax_igrid , 3);
     hinaxis(fnbimg_temp, &ay_igrid , 4);
     hinaxis(fnbimg_temp, &az_igrid , 5);
     auxclose(fnbimg_temp);
  
  /*
     nsmp_igrid = ahx_igrid.n*ahy_igrid.n*ax_igrid.n*ay_igrid.n*az_igrid.n;
  
     btmp = (float *) malloc(sizeof(float)*nsmp_igrid);
     bimg = (float *) malloc(sizeof(float)*nsmp_igrid);
  
     auxinout(fnbimg_temp); sseek(fnbimg_temp, 0, 0); sreed(fnbimg_temp, btmp, 4*nsmp_igrid); auxclose(fnbimg_temp);
  
     MPI_Barrier(MPI_COMM_WORLD);
  
     MPI_Allreduce(btmp,bimg,nsmp_igrid,MPI_FLOAT,MPI_SUM,MPI_COMM_WORLD);
  
     auxinout(fnbimg_temp); sseek(fnbimg_temp, 0, 0); srite(fnbimg_temp, bimg, 4*nsmp_igrid); auxclose(fnbimg_temp);
  */
     nsmp_igrid = ahx_igrid.n*ahy_igrid.n*ax_igrid.n*ay_igrid.n;
     btmp = (float *) malloc(sizeof(float)*nsmp_igrid);
     bimg = (float *) malloc(sizeof(float)*nsmp_igrid);
  
     auxinout(fnbimg_temp); sseek(fnbimg_temp, 0, 0);
     for (iz=1; iz<=az_igrid.n; iz++){
         for (ismp=0; ismp<nsmp_igrid; ismp++){
             btmp[ismp] = 0.;
             bimg[ismp] = 0.;
         }
         fio_imag(btmp,ahx_igrid.n,ahy_igrid.n,ax_igrid.n,ay_igrid.n,az_igrid.n,iz,fnbimg_temp,0);
         MPI_Barrier(MPI_COMM_WORLD);
         MPI_Allreduce(btmp,bimg,nsmp_igrid,MPI_FLOAT,MPI_SUM,MPI_COMM_WORLD);
         fio_imag(bimg,ahx_igrid.n,ahy_igrid.n,ax_igrid.n,ay_igrid.n,az_igrid.n,iz,fnbimg_temp,1);
     }
     auxclose(fnbimg_temp);
  
     if (irank==0){
        err = ssystem(cpy_bimg); 
        if (0!=err) {fprintf(stderr, "ERR=%d, irank=%d, %s\n", err, irank, cpy_bimg); MPI_Finalize(); exit(1);}
     }
  }
  
  if (irank==0){
     fprintf(stderr, "********************************\n");
     fprintf(stderr, " FINISHED BORN MIGRATION        \n");
     fprintf(stderr, "********************************\n");
  }
  MPI_Finalize();
  exit(0);
}




