#include "mpi.h"
#include "string.h"
#include <stdio.h>
#include "seplib.h"
#include "bwi-files-mpi.h"

#define LMAX_DIR 256
#define LMAX_FNS 256
#define LMAX_CMD 1024

int main(argc,argv)
int argc;
char *argv[];  {

char bindir[LMAX_DIR], datdir[LMAX_DIR], resdir[LMAX_DIR], tmpdir[LMAX_DIR], datapath[LMAX_DIR];
char fncsou[LMAX_FNS], fncrec[LMAX_FNS], fnbvel[LMAX_FNS], fnbimg[LMAX_FNS];
char fncsou_orig[LMAX_FNS], fncrec_orig[LMAX_FNS], fnbvel_orig[LMAX_FNS], fnbimg_orig[LMAX_FNS];
char fncsou_temp[LMAX_FNS], fncrec_temp[LMAX_FNS], fnbvel_temp[LMAX_FNS], fnbimg_temp[LMAX_FNS];
char copy_csou[LMAX_CMD], copy_crec[LMAX_CMD], copy_bvel[LMAX_CMD], copy_bimg[LMAX_CMD];
char command[LMAX_CMD], params[LMAX_CMD];
char suffix[20], shot_string[24], numstr[4], bimg_[5];

axis ahx={1,1.,0.}, ahy={1,1.,0.}, ax={1,1.,0.}, ay={1,1.,0.}, az={1,1.,0.};

float *imag, *istk;

int irank, nrank, nsmp, nsb;

MPI_Init(&argc,&argv);
MPI_Comm_rank(MPI_COMM_WORLD, &irank);
MPI_Comm_size(MPI_COMM_WORLD, &nrank);

initpar(argc,argv);
doc(" ");

strcpy(bimg_, "bimg");

if (0==getch("bindir","s",&bindir)) strcpy(bindir,"./");
if (0==getch("datdir","s",&datdir)) strcpy(datdir,"./");
if (0==getch("resdir","s",&resdir)) strcpy(resdir,"./");
if (0==getch("tmpdir","s",&tmpdir)) strcpy(tmpdir,"./");

if (0==getch("suffix","s",&suffix)) strcpy(suffix,"./");

if (0==getch("fncsou","s",&fncsou)) strcpy(fncsou,"./");
if (0==getch("fncrec","s",&fncrec)) strcpy(fncrec,"./");
if (0==getch("fnbvel","s",&fnbvel)) strcpy(fnbvel,"./");
if (0==getch("fnbimg","s",&fnbimg)) strcpy(fnbimg,"./");

if (0==getch("params","s",&params)) strcpy(params,"  ");

if (0==getch("nsb"   ,"i",&nsb   )) nsb = 1;

if (nrank!=nsb){
   fprintf(stderr, "ERROR: nrank-1 < nsb,  %d %d \n", nrank, nsb);
   MPI_FINALIZE();
}

strcpy(fnbimg_orig, resdir); strcat(fnbimg_orig, fnbimg);
strcpy(fnbimg_temp, tmpdir); strcat(fnbimg_temp, fnbimg);

// distribute the data 
//if (irank!=0){ 
   sprintf(numstr, "%d", irank+1);
   strcpy(shot_string, suffix); strcat(shot_string, numstr);

   strcpy(fncsou_orig, datdir); strcat(fncsou_orig, fncsou); 
   strcpy(fncrec_orig, datdir); strcat(fncrec_orig, fncrec); strcat(fncrec_orig, shot_string);
   strcpy(fnbvel_orig, datdir); strcat(fnbvel_orig, fnbvel);

   strcpy(fncsou_temp, tmpdir); strcat(fncsou_temp, fncsou);
   strcpy(fncrec_temp, tmpdir); strcat(fncrec_temp, fncrec); strcat(fncrec_temp, shot_string);
   strcpy(fnbvel_temp, tmpdir); strcat(fnbvel_temp, fnbvel);

// copy files to local machines
// copy source
   strcpy(copy_csou, "Cp "); strcat(copy_csou, fncsou_orig); strcat(copy_csou, " "); strcat(copy_csou, fncsou_temp); 
   strcat(copy_csou, " datapath="); strcat(copy_csou, tmpdir);

// copy receivers
   strcpy(copy_crec, "Cp "); strcat(copy_crec, fncrec_orig); strcat(copy_crec, " "); strcat(copy_crec, fncrec_temp);
   strcat(copy_crec, " datapath="); strcat(copy_crec, tmpdir);

// copy velocity model
   strcpy(copy_bvel, "Cp "); strcat(copy_bvel, fnbvel_orig); strcat(copy_bvel, " "); strcat(copy_bvel, fnbvel_temp);
   strcat(copy_bvel, " datapath="); strcat(copy_bvel, tmpdir);

   fprintf(stderr, "%s\n", copy_csou); system(copy_csou);
   fprintf(stderr, "%s\n", copy_crec); system(copy_crec);
   fprintf(stderr, "%s\n", copy_bvel); system(copy_bvel);

   strcpy(command, bindir      ); strcat(command, "bwi-wem3d.x"); strcat(command, " csou="); strcat(command, fncsou_temp);
   strcat(command, " crec="    ); strcat(command, fncrec_temp)  ; strcat(command, " bvel="); strcat(command, fnbvel_temp);
   strcat(command, " bimg="    ); strcat(command, fnbimg_temp)  ; strcat(command, " "     ); strcat(command, params     );
   strcat(command, " datapath="); strcat(command, tmpdir);

   fprintf(stderr, "%s\n", command); system(command);
//}

MPI_Barrier(MPI_COMM_WORLD);

auxinout(fnbimg_temp);
hinaxis(fnbimg_temp, &ahx, 1);
hinaxis(fnbimg_temp, &ahy, 2);
hinaxis(fnbimg_temp, &ax , 3);
hinaxis(fnbimg_temp, &ay , 4);
hinaxis(fnbimg_temp, &az , 5);

nsmp = ahx.n*ahy.n*ax.n*ay.n*az.n;

fprintf(stderr, "nsmp = %d, irank = %d \n", nsmp, irank);

imag = (float *) malloc(sizeof(float)*nsmp);

if (irank==0) istk = (float *) malloc(sizeof(float)*nsmp);

sseek(fnbimg_temp, 0, 0);
sreed(fnbimg_temp, imag, 4*nsmp);

MPI_Barrier(MPI_COMM_WORLD);

MPI_Reduce(imag,istk,nsmp,MPI_FLOAT,MPI_SUM,0,MPI_COMM_WORLD);

if (irank==0){
   sseek(fnbimg_temp, 0, 0);
   srite(fnbimg_temp, istk, 4*nsmp);
   strcpy(copy_bimg, "Cp "); strcat(copy_bimg, fnbimg_temp); strcat(copy_bimg, " "); strcat(copy_bimg, fnbimg_orig);
   fprintf(stderr, "%s\n", copy_bimg); system(copy_bimg);
}

MPI_Finalize();
}




