!---------------------------------------------
! One-way wave-equation shot-profile migration
!---------------------------------------------
! Author : Yaxun Tang
! Version: 2.0
! Date   : Jan. 1, 2008
program main
  use sep
  use base_struct_mod
  use base_ioutil_mod
  use base_string_mod
  implicit none
  include 'mpif.h'

  character(len=256)  :: bindir, datdir, resdir, tmpdir, datapath
  character(len=256)  :: fncsou, fncrec, fnbvel, fnbimg, bimg_
  character(len=256)  :: fncsou_orig, fncrec_orig, fnbvel_orig, fnbimg_orig
  character(len=256)  :: fncsou_temp, fncrec_temp, fnbvel_temp, fnbimg_temp
  character(len=1024) :: copy_csou, copy_crec, copy_bvel, copy_bimg, command, param
  character(len=256)  :: suffix, shot_string
  character(len=10)   :: numstr

  real, dimension(:), allocatable :: imag, istk
  type(axis) :: ahx, ahy, ax, ay, az
  integer             :: ierr, rc, irank, nrank

  integer             :: isb, nsb
  integer, external   :: fetch

  bimg_ = "bimg"

  bindir = ""
  datdir = ""
  resdir = ""
  tmpdir = ""
  suffix = ""
  fncsou = ""
  fncrec = ""
  fnbvel = ""
  fnbimg = ""

  call MPI_INIT(ierr)
  if (ierr .ne. MPI_SUCCESS) then
     write(0,*),'Error starting MPI program. Terminating.'
     call MPI_ABORT(MPI_COMM_WORLD, rc, ierr)
  end if

  call MPI_COMM_RANK(MPI_COMM_WORLD, irank, ierr)
  call MPI_COMM_SIZE(MPI_COMM_WORLD, nrank, ierr)

write(0,*)"irank=", irank, "nrank=", nrank

  call sep_init()
  call getch_add_string("head=/dev/null")
  call getch_add_string("noheader=y")
  call set_no_putch()
  call sep_begin_prog()

! read the parameters

  call base_string_init("bindir", bindir, "./")
  call base_string_init("datdir", datdir, "./")
  call base_string_init("resdir", resdir, "./")
  call base_string_init("tmpdir", tmpdir, "./")

  call base_string_init("suffix", suffix, "-shot")

  call base_string_init("fncsou", fncsou, "CSOU.H")
  call base_string_init("fncrec", fncrec, "CREC.H")
  call base_string_init("fnbvel", fnbvel, "BVEL.H")
  call base_string_init("fnbimg", fnbimg, "BIMG.H")

  call from_param("nsb", nsb, 1)

  if (nrank-1<nsb) then
     write(0,*)"ERROR: nrank-1 < nsb ", nrank, nsb
     call MPI_FINALIZE(ierr)
     stop
  end if
 
!  call MPI_BCAST(bindir,len(bindir),MPI_CHARACTER,0,MPI_COMM_WORLD,ierr) 
!  call MPI_BCAST(datdir,len(datdir),MPI_CHARACTER,0,MPI_COMM_WORLD,ierr)
!  call MPI_BCAST(resdir,len(resdir),MPI_CHARACTER,0,MPI_COMM_WORLD,ierr)
!  call MPI_BCAST(tmpdir,len(tmpdir),MPI_CHARACTER,0,MPI_COMM_WORLD,ierr)
!  call MPI_BCAST(suffix,len(suffix),MPI_CHARACTER,0,MPI_COMM_WORLD,ierr) 
!  call MPI_BCAST(fncsou,len(fncsou),MPI_CHARACTER,0,MPI_COMM_WORLD,ierr)
!  call MPI_BCAST(fncrec,len(fncrec),MPI_CHARACTER,0,MPI_COMM_WORLD,ierr)
!  call MPI_BCAST(fnbvel,len(fnbvel),MPI_CHARACTER,0,MPI_COMM_WORLD,ierr)
!  call MPI_BCAST(fnbimg,len(fnbimg),MPI_CHARACTER,0,MPI_COMM_WORLD,ierr)
!  call MPI_BCAST(nsb   ,1          ,MPI_INTEGER  ,0,MPI_COMM_WORLD,ierr)

  fnbimg_orig = trim(resdir)//trim(fnbimg)
  fnbimg_temp = trim(tmpdir)//trim(fnbimg)

! distribute the data 
  if (irank/=0) then
     write(numstr,'(I10)') irank
     numstr = adjustl(numstr)
     shot_string = trim(suffix)//trim(numstr)

     fncsou_orig = trim(datdir)//trim(fncsou)
     fncrec_orig = trim(datdir)//trim(fncrec)//trim(shot_string)
     fnbvel_orig = trim(datdir)//trim(fnbvel)

     fncsou_temp = trim(tmpdir)//trim(fncsou)
     fncrec_temp = trim(tmpdir)//trim(fncrec)//trim(shot_string)
     fnbvel_temp = trim(tmpdir)//trim(fnbvel)

!    copy files to local machines
     copy_csou = "Cp"//" "//trim(fncsou_orig)//" "//trim(fncsou_temp)//" "//"datapath="//trim(tmpdir)
     copy_crec = "Cp"//" "//trim(fncrec_orig)//" "//trim(fncrec_temp)//" "//"datapath="//trim(tmpdir)
     copy_bvel = "Cp"//" "//trim(fnbvel_orig)//" "//trim(fnbvel_temp)//" "//"datapath="//trim(tmpdir)

     write(0,*) "irank=", irank, trim(copy_csou)
     call system(trim(copy_csou))
     write(0,*) "irank=", irank, trim(copy_crec)
     call system(trim(copy_crec))
     write(0,*) "irank=", irank, trim(copy_bvel)
     call system(trim(copy_bvel))

     param = "mode=imgadj nxpad_beg=50 nxpad_end=50 ntpx=25 oper=ssfou nws=64 image_xmin=6000 image_xmax=12000 nhx=101 ohx=-1000 report=n memchk=n"

     command = trim(bindir)//"bwi-wem3d.x"//" csou="//trim(fncsou_temp)//" crec="//trim(fncrec_temp)//" bvel="//trim(fnbvel_temp)//" bimg="//trim(fnbimg_temp) &
               //" "//param//" datapath="//trim(tmpdir)

     write(0,*) trim(command)
     call system(trim(command))

  end if

  call MPI_BARRIER(MPI_COMM_WORLD, ierr)

  call auxinout(bimg_)

  if (irank==1) then
     call hinaxis(bimg_, ahx, 1)
     call hinaxis(bimg_, ahy, 2)
     call hinaxis(bimg_, ax , 3)
     call hinaxis(bimg_, ay , 4)
     call hinaxis(bimg_, az , 5)
  end if

  call MPI_BCAST(ahx%n,1,MPI_INTEGER,1,MPI_COMM_WORLD,ierr)
  call MPI_BCAST(ahx%d,1,MPI_REAL   ,1,MPI_COMM_WORLD,ierr)
  call MPI_BCAST(ahx%o,1,MPI_REAL   ,1,MPI_COMM_WORLD,ierr)
  call MPI_BCAST(ahy%n,1,MPI_INTEGER,1,MPI_COMM_WORLD,ierr)
  call MPI_BCAST(ahy%d,1,MPI_REAL   ,1,MPI_COMM_WORLD,ierr)
  call MPI_BCAST(ahy%o,1,MPI_REAL   ,1,MPI_COMM_WORLD,ierr)
  call MPI_BCAST(ax%n ,1,MPI_INTEGER,1,MPI_COMM_WORLD,ierr)
  call MPI_BCAST(ax%d ,1,MPI_REAL   ,1,MPI_COMM_WORLD,ierr)
  call MPI_BCAST(ax%o ,1,MPI_REAL   ,1,MPI_COMM_WORLD,ierr)
  call MPI_BCAST(ay%n ,1,MPI_INTEGER,1,MPI_COMM_WORLD,ierr)
  call MPI_BCAST(ay%d ,1,MPI_REAL   ,1,MPI_COMM_WORLD,ierr)
  call MPI_BCAST(ay%o ,1,MPI_REAL   ,1,MPI_COMM_WORLD,ierr)
  call MPI_BCAST(az%n ,1,MPI_INTEGER,1,MPI_COMM_WORLD,ierr)
  call MPI_BCAST(az%d ,1,MPI_REAL   ,1,MPI_COMM_WORLD,ierr)
  call MPI_BCAST(az%o ,1,MPI_REAL   ,1,MPI_COMM_WORLD,ierr)

  allocate(imag(ahx%n*ahy%n*ax%n*ay%n*az%n))

  imag = 0.

  if (irank/=0) then
     call sep_read(imag, bimg_) 
  else
     allocate(istk(ahx%n*ahy%n*ax%n*ay%n*az%n))
     istk = 0.
     call houaxis(bimg_, ahx, 1)
     call houaxis(bimg_, ahy, 2)
     call houaxis(bimg_, ax , 3)
     call houaxis(bimg_, ay , 4)
     call houaxis(bimg_, az , 5)
  end if

  call MPI_REDUCE(imag,istk,ahx%n*ahy%n*ax%n*ay%n*az%n,MPI_REAL,MPI_SUM,0,MPI_COMM_WORLD,ierr) 

  if (irank==0) then
     write(0,*)maxval(abs(istk)) 
     call sseek(bimg_, 0, 0) 
     call sep_write(istk, bimg_)
     copy_bimg = "Cp"//" "//trim(fnbimg_temp)//" "//trim(fnbimg_orig)
     write(0,*) trim(copy_bimg)
     call system(trim(copy_bimg))
  end if

  write(0,*)"***********************"
  write(0,*)"      Finished!!!      "
  write(0,*)"***********************"


  call MPI_FINALIZE(ierr)

end program
